Package: tablet
Type: Package
Title: Tabulate Descriptive Statistics in Multiple Formats
Version: 0.9.1
Authors@R: person(given = "Tim",
                  family = "Bergsma",
                  role = c("aut", "cre"),
                  email = "bergsmat@gmail.com")
Maintainer: Tim Bergsma <bergsmat@gmail.com>
BugReports: https://github.com/bergsmat/tablet/issues
Description: Creates a table of descriptive statistics
 for factor and numeric columns in a data frame. Displays
 these by groups, if any. Highly customizable, with support
 for 'html' and 'pdf' provided by 'kableExtra'. Respects
 original column order, column labels, and factor level order.
 See ?tablet.data.frame and vignettes.
License: GPL-3
Encoding: UTF-8
Imports: dplyr (>= 1.0.2), yamlet (>= 0.10.21), rlang, tidyr,
        kableExtra (>= 0.9.0), spork (>= 0.2.7), magrittr, fs,
        reactable
RoxygenNote: 7.3.3
VignetteBuilder: knitr
Suggests: knitr, rmarkdown, boot, testthat, shiny, shinyFiles, haven,
        yaml, sortable, latexpdf, tinytex, tools, csv, xtable,
        shinyAce, R.utils
NeedsCompilation: no
Packaged: 2026-07-19 03:37:58 UTC; tim.bergsma
Author: Tim Bergsma [aut, cre]
Repository: CRAN
Date/Publication: 2026-07-19 06:00:02 UTC
Built: R 4.6.0; ; 2026-07-19 08:29:01 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(as.data.frame,gridtable)
S3method(as_calibrated,data.frame)
S3method(as_gridtable,character)
S3method(as_gridtable,data.frame)
S3method(as_kable,tablet)
S3method(as_tablet,data.frame)
S3method(as_xtable,tablet)
S3method(categoricals,data.frame)
S3method(classifiers,data.frame)
S3method(devalued,observations)
S3method(groupfull,data.frame)
S3method(groupless,data.frame)
S3method(groupwise,data.frame)
S3method(header_rows,tablet)
S3method(headerlist,tablette)
S3method(index,tablette)
S3method(io_csv,tablet)
S3method(io_tablet,character)
S3method(io_tablet,data.frame)
S3method(io_tablet,tablet)
S3method(kbl,calibrated)
S3method(kbl,default)
S3method(kbl,gridtable)
S3method(kbl,tablet)
S3method(numerics,data.frame)
S3method(observations,data.frame)
S3method(print,gridtable)
S3method(recap,knitr_kable)
S3method(splice,data.frame)
S3method(tablet,data.frame)
S3method(tablet,tablette)
S3method(tablette,groupwise)
S3method(tablette,tablet)
S3method(widgets,devalued)
export(absolutizePath)
export(as_calibrated)
export(as_gridtable)
export(as_kable)
export(as_tablet)
export(as_xtable)
export(categoricals)
export(classifiers)
export(devalued)
export(escape_latex)
export(escape_latex.default)
export(escape_latex.latex)
export(groupfull)
export(groupless)
export(groupwise)
export(header_rows)
export(headerlist)
export(index)
export(io_tablet)
export(kbl)
export(mesa)
export(numerics)
export(observations)
export(recap)
export(relativizePath)
export(splice)
export(tablet)
export(tablette)
export(widgets)
importFrom(dplyr,across)
importFrom(dplyr,add_count)
importFrom(dplyr,add_tally)
importFrom(dplyr,all_of)
importFrom(dplyr,anti_join)
importFrom(dplyr,arrange)
importFrom(dplyr,bind_rows)
importFrom(dplyr,case_when)
importFrom(dplyr,distinct)
importFrom(dplyr,everything)
importFrom(dplyr,filter)
importFrom(dplyr,full_join)
importFrom(dplyr,group_by)
importFrom(dplyr,group_vars)
importFrom(dplyr,groups)
importFrom(dplyr,left_join)
importFrom(dplyr,mutate)
importFrom(dplyr,rename)
importFrom(dplyr,select)
importFrom(dplyr,slice)
importFrom(dplyr,tally)
importFrom(dplyr,ungroup)
importFrom(fs,is_absolute_path)
importFrom(kableExtra,add_header_above)
importFrom(kableExtra,column_spec)
importFrom(kableExtra,kbl)
importFrom(kableExtra,linebreak)
importFrom(kableExtra,pack_rows)
importFrom(magrittr,"%<>%")
importFrom(magrittr,"%>%")
importFrom(reactable,reactable)
importFrom(reactable,reactableOutput)
importFrom(reactable,renderReactable)
importFrom(rlang,":=")
importFrom(rlang,as_string)
importFrom(rlang,eval_tidy)
importFrom(rlang,f_lhs)
importFrom(rlang,f_rhs)
importFrom(spork,as_html)
importFrom(spork,as_latex)
importFrom(spork,as_spork)
importFrom(tidyr,gather)
importFrom(tidyr,replace_na)
importFrom(tidyr,spread)
importFrom(yamlet,as_dvec)
importFrom(yamlet,io_csv)
importFrom(yamlet,modify)

data(mtcars)
library(haven)
library(dplyr)
library(magrittr)
library(yamlet)
library(SASxport)

x <- mtcars
x$car <- rownames(x)
x %<>% select(car, everything())

x %<>% decorate('
car:  [ Automobile ]
mpg:  [ Mileage, mi/gal ]
cyl:  [ Number of Cylinders, [ four: 4, six: 6, eight: 8 ]]
disp: [ Displacement, in^3 ]
hp:   [ Gross Horsepower, hp ]
drat: [ Rear Axle Ratio ]
wt:   [ Weight, klb ]
qsec: [ Quarter-mile Time, sec ]
vs:   [ Engine, [ V-shaped: 0, Straight: 1 ]]
am:   [ Transmission, [ Automatic: 0, Manual: 1 ]]
gear: [ Number of Forward Gears, [ 3-speed: 3, 4-speed: 4, 5-speed: 5 ]]
carb: [ Number of Carburetors, [ 1, 2, 3, 4, 6, 8 ]]
')

x %>% io_csv('mtcars.csv')
x %>% write_sas('mtcars.sas7bdat')
x %>% write.xport(file = 'mtcars.xpt')

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
#knitr::opts_chunk$set(package.startup.message = FALSE)
options(xtable.comment = FALSE)

## ----software, include = FALSE------------------------------------------------
library(tidyr)
library(dplyr)
library(magrittr)
library(kableExtra)
library(boot)
library(yamlet)
library(tablet)
# options(knitr.table.format = "latex") # not needed since kableExtra 0.9.0

## ----software2, eval = FALSE--------------------------------------------------
# library(tidyr)
# library(dplyr)
# library(magrittr)
# library(kableExtra)
# library(boot)
# library(yamlet)
# library(tablet)

## ----data---------------------------------------------------------------------
x <- melanoma
x %<>% select(-time, -year)

## ----easy---------------------------------------------------------------------
x %>%
  mutate(
    sex = factor(sex), 
    ulcer = factor(ulcer)
  ) %>%
  tablet %>%
  kbl

## -----------------------------------------------------------------------------
x <- melanoma

x %<>% decorate('
time:      [ Survival Time Since Operation, day ]
status:
 - End of Study Patient Status
 -
  - Alive: 2
  - Melanoma Death: 1
  - Unrelated Death: 3
sex:       [ Sex, [ Male: 1, Female: 0 ]]
age:       [ Age at Time of Operation, year ]
year:      [ Year of Operation, year ]
thickness: [ Tumor Thickness, mm ]
ulcer:     [ Ulceration, [ Absent: 0, Present: 1 ]]
')
x %<>% select(-time, -year)
x %<>% group_by(status)
x %<>% resolve

## ----meta---------------------------------------------------------------------
x %>% tablet %>% kbl

## ----xtable, results = 'asis'-------------------------------------------------
library(xtable)
  x %>% 
    filter(!(status == 'Alive' & sex == 'Male')) %>%
    tablet %>% as_xtable(format_value = function(x,...)x) %>% 
    print(
      booktabs = TRUE, 
      include.rownames = FALSE 
    )

## ----grouped------------------------------------------------------------------
x %<>% mutate(class = status)                          # copy the current group
x %<>% modify(class, label = 'class')                  # change its label
levels(x$status) <- c('Alive','Melanoma','Unrelated')  # tweak current group
levels(x$class)  <- c(' ',    'Death',   'Death')      # cluster groups
x %<>% group_by(class, status)                         # nest groups
x %>% tablet %>% kbl                              # render

## ----transposed---------------------------------------------------------------
x %<>% group_by(status, sex)
x %<>% select(-class)
x %>% 
  tablet %>% 
  kbl %>% 
  kable_styling(latex_options = 'scale_down')


## ----transposed2--------------------------------------------------------------
x %<>% group_by(status, ulcer)
x %>% 
  tablet %>% 
  kbl %>% 
  kable_styling(latex_options = 'scale_down')


## ----transposed3--------------------------------------------------------------
x %<>% group_by(status, ulcer, sex)
x %>% 
  tablet %>% 
  kbl %>% 
  kable_styling(latex_options = 'scale_down') # %>% landscape ?


## ----aesthetics---------------------------------------------------------------
x %<>% group_by(status)
x %>% 
  tablet(
    fac = NULL,
    lab ~ name,
    `Median (range)` ~ med + ' (' + min + ' - ' + max + ')'
  ) %>% 
  kbl

library(shiny)
library(shinyFiles)
library(fs)
library(haven)
library(tablet)
library(dplyr)
library(magrittr)
library(yamlet)
library(yaml)
library(sortable)
library(kableExtra)
library(knitr)
library(latexpdf)
library(tools)
library(csv)
library(shinyAce)
library(spork)
library(reactable)

ui <- shinyUI(
  navbarPage(
    'Mesa',
    tabPanel(
      'Input',
      sidebarLayout(
        sidebarPanel(
          width = 12,
          shinyFilesButton(
            id = 'source',
            label = 'data',
            title = 'choose data or metadata (*.yaml) file:',
            multiple = FALSE
          ),
          textOutput('filepath'),
          shinyFilesButton(
            id = 'config',
            label = 'configuration',
            title = 'choose configuration file:',
            multiple = FALSE
          ),
          textOutput('confpath'),
          uiOutput('saveconfig'),
          uiOutput('splice'),
          uiOutput('keep'),
          uiOutput('buckets')
        ),
        mainPanel(
          width = 0
        ) # end main panel
      ) # end sidebar layout
    ),
    tabPanel(
      'Data',
      sidebarLayout(
        sidebarPanel(
          width = 0
        ),
        mainPanel(
          width = 12,
          reactable::reactableOutput("data"),
        )
      )
    ),
    tabPanel(
      'Variables',
      sidebarLayout(
        sidebarPanel(
          width = 2,
          uiOutput('saveMeta'),
          textOutput('metapath')
        ),
        mainPanel(
          width = 12,
          uiOutput('meta')
        )
      )
    ),
    tabPanel(
      'Derivations',
      sidebarLayout(
        sidebarPanel(
          width = 2,
          uiOutput('saveDeri'),
          textOutput('deripath')
        ),
        mainPanel(
          width = 12,
          uiOutput('deri')
        )
      )
    ),
    tabPanel(
      'Shell',
      sidebarLayout(
        sidebarPanel(
          width = 12,
          actionButton('submit', 'Submit'),
          uiOutput('outputid'),
          uiOutput('caption'),
          uiOutput('footnotes'),
          uiOutput('lhead1'),
          uiOutput('lhead2'),
          uiOutput('rhead1'),
          uiOutput('rhead2'),
          uiOutput('cont')
        ),
        mainPanel(width = 0) #end main panel
      )
    ),
    tabPanel(
      'Preview',
      sidebarLayout(
        sidebarPanel(
          width = 2,
          uiOutput('labelhtml'),
          uiOutput('savecsv')
        ),
        mainPanel(width = 10,
                  htmlOutput('preview')
        )
      )
    ),
    tabPanel(
      'PDF',
      sidebarLayout(
        sidebarPanel(
          width = 2,
          uiOutput('repeatheader'),
          uiOutput('repeatfootnote'),
          #uiOutput('spork'),
          # uiOutput('na_string'),
          uiOutput('labeltex'),
          uiOutput('savepdf')
        ),
        mainPanel(
          width = 10,
          uiOutput('pdfview')
        )
      )
    )
  ) # end page
) # end ui

server <- shinyServer(function(input, output, session) {

  # declare the objects that control the application

  conf <- reactiveValues(
    filepath   = character(0),
    metapath   = character(0),
    deripath   = character(0),
    confpath   = character(0),
    selected   = character(0),
    filter_by  = character(0),
    keep       = list(), # a named list of filter_by levels to keep
    group_by   = character(0),
    sequential = FALSE,
    title      = 'Title',
    outputid   = 'T-00-00',
    lhead1     = 'Company',
    lhead2     = 'Project',
    rhead1     = 'Confidential',
    rhead2     = 'Draft',
    cont  = '(continued)',
    footnotes  = '(footnotes here)',
#    na_string  = 'NA',
    x          = data.frame(),
    mv         = 0,
    dv         = 0,
    editor     = NULL,
    editor2    = NULL,
    labelhtml  = 'no',
    labeltex   = 'no',
    repeathead = 'no',
    repeatfoot = 'no',
    tablet     = as.character(packageVersion('tablet'))
  )

  reset_conf <- function(){
    printer('reset_conf')
    conf$filepath   <- character(0)
    conf$metapath   <- character(0)
    conf$deripath   <- character(0)
    conf$confpath   <- character(0)
    conf$selected   <- character(0)
    conf$filter_by  <- character(0)
    conf$keep       <- list() # a named list of filter_by levels to keep
    conf$group_by   <- character(0)
    conf$sequential <- FALSE
    conf$title      <- 'Title'
    conf$outputid   <- 'T-00-00'
    conf$lhead1     <- 'Company'
    conf$lhead2     <- 'Project'
    conf$rhead1     <- 'Confidential'
    conf$rhead2     <- 'Draft'
    conf$cont       <- '(continued)'
    conf$footnotes  <- '(footnotes here)'
 #   conf$na_string  <- 'NA'
    conf$x          <- data.frame()
    conf$imputed    <- character()
    conf$mv         <- 0
    conf$dv         <- 0
    conf$editor     <- NULL
    conf$editor2    <- NULL
    labelhtml       <- 'no'
    labeltex        <- 'no'
    repeathead      <- 'no'
    repeatfoot      <- 'no'
    tablet          <- as.character(packageVersion('tablet'))
  }


  file_ok <- function(x){
    if(!length(x))return(FALSE)
    if(!file.exists(x))return(FALSE)
    return(TRUE)
  }
  # https://github.com/thomasp85/shinyFiles/issues/85

  volumes <- getVolumes()
  moreVolumes <- function()c(
    volumes(),
    examples = system.file('shiny-examples/mesa/data', package = 'tablet'),
    home = fs::path_home(),
    R = R.home()
  )
  ui_volumes <- reactive({
    printer('reactive ui_volumes')
    volumes <- moreVolumes()
    if(length(conf$filepath) & !any(is.na(conf$filepath))){
      sel_path <- dirname(conf$filepath)
      if(!sel_path %in% volumes){
        vnames <- c(basename(sel_path), names(volumes))
        volumes <- setNames(c(sel_path, volumes), vnames)
      }
    }
    if(length(conf$confpath) & !any(is.na(conf$confpath))){
      sel_path <- dirname(conf$confpath)
      if(!sel_path %in% volumes){
        vnames <- c(basename(sel_path), names(volumes))
        volumes <- setNames(c(sel_path, volumes), vnames)
      }
    }
    volumes
  })

  # set up the file choosers
  # https://stackoverflow.com/questions/53641749/how-to-use-shinyfilechoose-to-create-an-reactive-object-to-load-a-data-frame

  # file_selected <- reactive({
  #   shinyFileChoose(input, "file", roots = volumes, session = session)
  #   req(input$file)
  #   if (is.null(input$file))
  #     return(NULL)
  #   return(parseFilePaths(volumes, input$file)$datapath)
  # })

  # choose data (or metadata)
  shinyFileChoose(
    input,
    'source',
    roots = ui_volumes,
    session = session,
    filetypes = c('sas7bdat', 'csv', 'xpt', 'yaml')
  )
  observeEvent(input$source, {
    printer('observeEvent:input$source')
    req(input$source)
    if(is.null(input$source)) return(NULL)
    newsource <- parseFilePaths(ui_volumes, input$source)$datapath
    if(is.character(newsource)){
      if(length(newsource)){
        if(file.exists(newsource)){
          reset_conf()
          conf$filepath <- newsource
        }
      }
    }
  })

  # choose config

  shinyFileChoose(
    input,
    'config',
    roots = ui_volumes,
    session = session,
    filetypes = c('conf')
  )

  observeEvent(input$config,{
    printer('observeEvent:input$config')
    req(input$config)
    if(is.null(input$config)) return(NULL)
    newconfig <- parseFilePaths(ui_volumes, input$config)$datapath
    if(is.character(newconfig)){
      if(length(newconfig)){
        if(file.exists(newconfig)) {
          conf$confpath <- newconfig
        }
      }
    }
  })

  # https://stackoverflow.com/questions/39517199/how-to-specify-file-and-path-to-save-a-file-with-r-shiny-and-shinyfiles

  output$saveconfig <- renderUI({
    printer('output$saveconfig')
    shinySaveButton(
      id = 'saveconf',
      label = 'save configuration',
      title = 'save configuration as:',
      filetype = list(conf = 'conf'),
      filename = paste0(conf$outputid, '.conf')
    )

  })

  # save the current config
  observeEvent(input$saveconf, {
    printer('observeEvent:input$saveconf')
    req(input$saveconf)
    shinyFileSave(input, 'saveconf', roots = ui_volumes, session = session)
    fileinfo <- parseSavePath(ui_volumes, input$saveconf)
    if (nrow(fileinfo) > 0) {
      path <- as.character(fileinfo$datapath)

      vals <- isolate(
        reactiveValuesToList(conf)[
          !names(conf) %in% c(
            'x',
            'confpath',
            'editor',
            'editor2',
            'mv',
            'dv',
            'imputed'
          )
        ]
      )
      # dictate storage order!
      vals <- vals[c(
        'filepath',
        'metapath',
        'deripath',
        'selected',
        'group_by',
        'filter_by',
        'keep',
        'sequential',
        'outputid',
        'title',
        'lhead1',
        'lhead2',
        'rhead1',
        'rhead2',
        'footnotes',
        'repeathead',
        'repeatfoot',
        'cont',
        'labelhtml',
        'labeltex',
        'tablet'
      )]

      # note: below is the only place in the application where the configuration is written to storage.
      # filepath and metapath and deripath, like confpath, are stored internally as absolute paths.
      # but on write they are expressed relative to confpath directory,
      # and on read they are understood relative to confpath directory (and converted to absolute).

      if(length(vals$filepath))vals$filepath <- relativizePath(vals$filepath, dirname(path))
      if(length(vals$metapath))vals$metapath <- relativizePath(vals$metapath, dirname(path))
      if(length(vals$deripath))vals$deripath <- relativizePath(vals$deripath, dirname(path))
      res <- try(write_yaml(vals, path)) # only reads on save
      res <- !inherits(res, 'try-error')
      dur <- 10
      if(res) dur <- 5
      showNotification(
        duration = dur,
        type = ifelse(res, 'default', 'error'),
        ui = paste(
          ifelse(res, 'wrote', 'did not write'),
          path
        )
      )
      if(res){
        conf$confpath <- path
      }
    }
  })

  output$savecsv <- renderUI({
    printer('output$savecsv')
    shinySaveButton(
      id = 'savetable',
      label = 'save table',
      title = 'save table as:',
      filetype = list(csv = 'csv'),
      filename = paste0(conf$outputid, '.csv')
    )
  })

  # save the preview table
  observeEvent(input$savetable, {
    printer('observeEvent: input$savetable')
    req(input$savetable)
    shinyFileSave(input, 'savetable', roots = ui_volumes, session = session)
    fileinfo <- parseSavePath(ui_volumes, input$savetable)
    if (nrow(fileinfo) > 0) {
      path <- as.character(fileinfo$datapath)
      data <- isolate(summarized())
      res <- try(as.csv(data, path))
      res <- !inherits(res, 'try-error')
      dur <- 10
      if(res) dur <- 5
      showNotification(
        duration = dur,
        type = ifelse(res, 'default', 'error'),
        ui = paste(
          ifelse(res, 'wrote', 'did not write'),
          path
        )
      )
    }
  })

  output$savepdf <- renderUI({
    printer('output$savepdf')
    shinySaveButton(
      id = 'savepdf',
      label = 'save pdf',
      title = 'save pdf as:',
      filetype = list(pdf = 'pdf'),
      filename = paste0(conf$outputid, '.pdf')
    )
  })

  # save the pdf as
  observeEvent(input$savepdf, {
    printer('observeEvent:input$savepdf')
    req(input$savepdf)
    shinyFileSave(input, 'savepdf', roots = ui_volumes, session = session)
    fileinfo <- parseSavePath(ui_volumes, input$savepdf)
    if (nrow(fileinfo) > 0) {
      path <- as.character(fileinfo$datapath)
      from <- isolate(pdf_location())
      from <- file.path('www', from)
      res <- file.copy(from, path, overwrite = TRUE)
      dur <- 10
      if(res) dur <- 5
      showNotification(
        duration = dur,
        type = ifelse(res, 'default', 'error'),
        ui = paste(
          ifelse(res, 'wrote', 'did not write'),
          path
        )
      )
    }
  })

  #https://stackoverflow.com/questions/40547786/shiny-can-dynamically-generated-buttons-act-as-trigger-for-an-event

  observers <- list()

  observeEvent(input$selected,{
    printer('observeEvent: input$selected')
    conf$selected <- input$selected
  })

  observeEvent(input$filter_by,{
    printer('observeEvent:input$filter_by')
    conf$filter_by <- input$filter_by
  })

  observeEvent(input$group_by,{
    printer('observeEvent:input$group_by')
    conf$group_by <- input$group_by
  })

  observeEvent(input$splice,{
    printer('observeEvent:Input$splice')
    conf$sequential <- ifelse(input$splice == 'sequential', TRUE, FALSE)
  })

  observeEvent(input$submit,{
    printer('observeEvent:input$submit(conf$title)')
    conf$title <- input$caption
  })

  # observeEvent(input$csv,{
  #   as.csv(summarized(), paste0(conf$outputid,'.csv'))
  # })

  observeEvent(input$submit,{
    printer('observeEvent:input$submit(conf$outputid)')
    conf$outputid <- input$outputid
  })

  observeEvent(input$submit,{
    printer('observeEvent:input$submit(conf$lhead1)')
    conf$lhead1 <- input$lhead1
  })

  observeEvent(input$submit,{
    printer('observeEvent:input$submit(conf$lhead2)')
    conf$lhead2 <- input$lhead2
  })

  observeEvent(input$submit,{
    printer('observeEvent:input$submit(conf$rhead1)')
    conf$rhead1 <- input$rhead1
  })

  observeEvent(input$submit,{
    printer('observeEvent:input$submit(conf$rhead2)')
    conf$rhead2 <- input$rhead2
  })

  observeEvent(input$submit,{
    printer('observeEvent:input$submit(conf$cont)')
    conf$cont <- input$cont
  })

  observeEvent(input$submit,{
    printer('observeEvent:input$submit(conf$footnotes)')
    conf$footnotes <- input$footnotes
  })
  
  # observeEvent(input$na_string,{
  #   conf$na_string <- input$na_string
  # })
  observeEvent(input$repeathead,{
    printer('observeEvent:input$repeathead')
    conf$repeathead <- input$repeathead
  })
  
  observeEvent(input$repeatfoot,{
    printer('observeEvent:input$repeatfoot')
    conf$repeatfoot <- input$repeatfoot
  })
  
  observeEvent(input$labelhtml,{
    printer('observeEvent:input$labelhtml')
    conf$labelhtml <- input$labelhtml
  })
  
  observeEvent(input$labeltex,{
    printer('observeEvent:input$labeltex')
    printer(paste0(conf$labeltex,'<-', input$labeltex))
    conf$labeltex <- input$labeltex
  })

  observeEvent(conf$confpath,{
    printer('observeEvent:conf$confpath')
    if(!length(conf$confpath)){
      return()
    }

    if(!file.exists(conf$confpath)){
      return()
    }

    saved <- list()
    tryCatch(
      saved <- read_yaml(conf$confpath),
      error = function(e) showNotification(duration = NULL, type = 'error', as.character(e))
    )

    if(!length(saved)){
      showNotification(duration = NULL, type = 'message', 'resetting configuration')
      reset_conf()
      return()
    }

    # items not saved should be re-initialized
    if(!('filepath' %in% names(saved))) {
      showNotification(duration = NULL, type = 'error', 'configuration does not specify file path')
      reset_conf()
      return()
    }

    if(
      !file.exists(
        absolutizePath(
          saved$filepath,
          dirname(conf$confpath)
        )
      )
    ){
      showNotification(duration = NULL, type = 'error', 'configured file path not found')
      reset_conf()
      return()
    }

    # at this point:
    #   * confpath has changed
    #   * confpath is readable/parseable
    #   * configuration has a filepath
    #   * filepath exists

    # update internal configuration from saved configuration

    # note: below is the only place in the application where the configuration is read from storage.
    # filepath and metapath, like confpath, are stored internally as absolute paths.
    # but on write they are expressed relative to confpath directory,
    # and on read they are understood relative to confpath directory (and converted to absolute).

    if(!is.null(saved$filepath))conf$filepath <- absolutizePath(saved$filepath, dirname(conf$confpath))
    if(!is.null(saved$metapath))conf$metapath <- absolutizePath(saved$metapath, dirname(conf$confpath))
    if(!is.null(saved$deripath))conf$deripath <- absolutizePath(saved$deripath, dirname(conf$confpath))
    if(!is.null(saved$selected))conf$selected <- saved$selected
    if(!is.null(saved$filter_by))conf$filter_by <- saved$filter_by
    if(!is.null(saved$keep))conf$keep      <- saved$keep
    if(!is.null(saved$group_by))conf$group_by   <- saved$group_by
    if(!is.null(saved$sequential))conf$sequential<- saved$sequential
    if(!is.null(saved$title))conf$title      <- saved$title
    if(!is.null(saved$lhead1))conf$lhead1     <- saved$lhead1
    if(!is.null(saved$lhead2))conf$lhead2    <- saved$lhead2
    if(!is.null(saved$rhead1))conf$rhead1     <- saved$rhead1
    if(!is.null(saved$rhead2))conf$rhead2    <- saved$rhead2
    if(!is.null(saved$cont))conf$cont <- saved$cont
    if(!is.null(saved$footnotes))conf$footnotes <- saved$footnotes
#    if(!is.null(saved$na_string))conf$na_string <- saved$na_string
    if(!is.null(saved$outputid))conf$outputid <- saved$outputid
    if(!is.null(saved$repeathead))conf$repeathead <- saved$repeathead
    if(!is.null(saved$repeatfoot))conf$repeatfoot <- saved$repeatfoot
    if(!is.null(saved$labelhtml))conf$labelhtml <- saved$labelhtml
    if(!is.null(saved$labeltex))conf$labeltex <- saved$labeltex

    # version control
    if(!is.null(saved$tablet)){
      if(!identical(saved$tablet, conf$tablet)){
        showNotification(
          duration = NULL,
          type = 'warning',
          paste(
            'configuration was last saved by tablet version', saved$tablet,
            'but currently using', conf$tablet
          )
        )
      }
    }

    #conf$x          = data.frame()
    # if filepath has changed, data will be re-read: see observeEvent(conf$filepath)
  })

  # when conf$filepath changes, we rebuild the data
  # also need to trigger when metadata changes on disk
  # i.e. when we have saved it.

  # https://stackoverflow.com/questions/34731975/how-to-listen-for-more-than-one-event-expression-within-a-shiny-eventreactive-ha

  printer <- function(x)writeLines(as.character(x))

  observeEvent({
      conf$filepath # new data selected
      conf$mv # metadata re-written
      conf$dv # derivations re-written
      1 # prevents NULL from squelching the observation
    },
    {
    printer('observeEvent:conf$filepath')
    #browser()

    # invalidate the keep/filter observers if data changes
    observers <<- list()

    # invalidate configuration if an attempt is made to supplant data
    # conf$confpath <- character(0)
    # this does not work!

    if(!length(conf$filepath))return()
    theFile <- conf$filepath
    is_data <- grepl('\\.sas7bdat|xpt|csv$', theFile)
    is_meta <- grepl('\\.yaml$', theFile)
    is_deri <- grepl('\\.der$', theFile)
    
    datafile <- theFile
    if(is_deri) {
      datafile <- sub('der$','sas7bdat',theFile)
      if(!file.exists(datafile)) datafile <- sub('der$','xpt',theFile)
      if(!file.exists(datafile)) datafile <- sub('der$','csv',theFile)
    } # try everything

    if(is_meta) {
      datafile <- sub('yaml$','sas7bdat',theFile)
      if(!file.exists(datafile)) datafile <- sub('yaml$','xpt',theFile)
      if(!file.exists(datafile)) datafile <- sub('yaml$','csv',theFile)
    } # try everything
    
    metafile <- theFile
    if(is_data) metafile <- sub('sas7bdat|xpt|csv$', 'yaml', theFile)
    if(is_deri) metafile <- sub('der$', 'yaml', theFile)
    
    derifile <- theFile
    if(is_data) derifile <- sub('sas7bdat|xpt|csv$', 'der', theFile)
    if(is_meta) derifile <- sub('yaml$', 'der', theFile)
    
    has_data <- file.exists(datafile)
    has_meta <- file.exists(metafile)
    has_deri <- file.exists(derifile)

   # d <- data.frame() # 2022/04/13 make html() responsive to coerced columns
    d <- conf$x

    if(has_data){
      if(grepl('sas7bdat$', datafile)) d <- data.frame(read_sas(datafile))
      if(grepl('xpt$', datafile)) d <- data.frame(read.xport(datafile))
      if(grepl('csv$', datafile)) d <- data.frame(as.csv(datafile))
    }

    # at this point, best data has been defined. Define default metadata.

    m <- decorations(d)
#    e <- d %>% mutate(label = NULL, guide = NULL)
    # default derivations (blank)
    e <- d %>% 
      names %>% 
      sapply(identity, simplify = FALSE) %>% 
      as.data.frame 
    n <- decorations(e)

    # Either read the metadata or write it.
    if(has_meta){
      # try for better meta.
      tryCatch(
        m <- read_yamlet(metafile),
        error = function(e) showNotification(duration = NULL, type = 'error', as.character(e))
      )
    } else {
      write_yamlet(m, metafile)
    }

    # now we have best available metadata
    has_meta <- TRUE
    conf$metapath <- metafile # make visible
    
    if(has_deri){
      tryCatch(
        n <- read_yamlet(derifile),
        error = function(e) showNotification(duration = NULL, type = 'error', as.character(e))
      )
    } else {
      write_yamlet(n, derifile)
    }
    
    has_deri <- TRUE
    conf$deripath <- derifile

    # make data look like metadata (which may be superset)

    have <- names(d)
    need <- names(m)
    make <- setdiff(need, have)
    #browser()
    for(col in make) d[[col]] <- rep(NA_real_, nrow(d))


    # ensure positive nrow # removed at 0.5.4
    # if(nrow(d) == 0) d <- d['',,drop = FALSE]

    # drop unspecified
    d %<>% select(!!!names(m))

    # apply meta
    d <- redecorate(d, m)

    # # Promote NA to a level of the factor
    # d %<>% resolve(exclude = NULL)
    d %<>% resolve()

    # store on the session
    conf$x <- d
    conf$imputed <- sapply(select(d, !!!make), attr, 'label')

  })

  output$filepath <- renderPrint({
    #printer('observeEvent:output$filepath')
    if (!length(conf$filepath)) {
      cat('No input data selected.')
    } else {
      cat(conf$filepath)
    }
  })

  filtered <- reactive({
    printer('filtered')
    x <- conf$x
    cols <- conf$filter_by
    for(filter in cols){
      scope <- input[[paste0('mesa_filter_', filter)]]
      if(length(scope)){ # only filter if at least one choice was made!
        # save these for drawing the UI
        conf$keep[[filter]] <- scope
        index <- x[[filter]] %in% scope
        x <- x[index,,drop = FALSE]
      }
    }
    x
  })

  factorized <- reactive({
    printer('factorized')
    x <- filtered()
    x %<>% mutate_if(is.character, classified)
    suppressWarnings(x %<>% modify(title = label))
    #browser()
    hasUnits <- sapply(x, function(col)'units' %in% names(attributes(col)))
    hasUnits <- names(hasUnits[hasUnits])
    suppressWarnings(x %<>% modify(!!!hasUnits, title = paste0(label, ' (', .data$units, ')')))

    # conditionally creating scripted labels has the
    # unintended effect of making the pdf and preview displays
    # co-dependent, since the output of factorized() changes
    # when either flag changes.
    # Meanwhile, it is easy and cheap to calculate html/tex labels
    # unconditionally, but use them conditionally.
    # Accordingly, we unconditionalize the following code.

    # if(length(input$labelhtml) == 1){
    #   printer('factorized - labelhtml')
    #   if(input$labelhtml == TRUE){

    suppressWarnings(x %<>% modify(original = name))
    suppressWarnings(x %<>% modify(html = as_html(as_spork(.data$name)))) # default
    suppressWarnings(x %<>% modify(html = as_html(as_spork(.data$label))))
    hasUnits <- sapply(x, function(col)'units' %in% names(attributes(col)))
    hasUnits <- names(hasUnits[hasUnits])
        suppressWarnings(x %<>% modify(
      !!!hasUnits,
      html = concatenate(as_html(as_spork(c(.data$label, ' (', .data$units,')'))))
    ))
    #   }
    # }else{printer('factorized - no labelhtml')}
    # if(length(input$labeltex) == 1){
    #   printer('factorized - labeltex')
    #
    #   if(input$labeltex == TRUE){
        # browser()
        # we need default 'latex' tex attributes for all columns ...
    suppressWarnings(x %<>% modify(tex = as_latex(as_spork(.data$name))))
    suppressWarnings(x %<>% modify(tex = as_latex(as_spork(.data$label))))
    hasUnits <- sapply(x, function(col)'units' %in% names(attributes(col)))
    hasUnits <- names(hasUnits[hasUnits])
    suppressWarnings(x %<>% modify(
      !!!hasUnits,
      # should retain class 'latex'
      # currently pre-doubled by escape_latex.latex
      tex = concatenate( as_latex(as_spork(c(.data$label, ' (', .data$units,')'))))
    ))
    #   }
    # }else{printer('factorized - no labeltex')}
    x
  })

  selected <- reactive({
    printer('selected')
    x <- factorized()
    if(length(conf$group_by)) x %<>% group_by(!!!syms(conf$group_by))
    x %<>% select(!!!syms(conf$selected))
    x
  })

  args <- reactive({
    printer('args')
    x <- list(x = selected())
    extra <- list(
      all_levels = TRUE,
      # all = 'All',
      fun = list(
        sum ~ sum(x,  na.rm = TRUE),
        pct ~ signif(digits = 3,     sum / n * 100        ),
        ave ~ signif(digits = 3,    mean(x,  na.rm = TRUE)),
        std ~ signif(digits = 3,      sd(x,  na.rm = TRUE)),
        med ~ signif(digits = 3,  median(x,  na.rm = TRUE)),
        min ~ signif(digits = 3,     min(x,  na.rm = TRUE)),
        max ~ signif(digits = 3,     max(x,  na.rm = TRUE)),
        smn ~ sum(!is.na(x))
      ),
      num = list(
        n ~ smn,
        `Mean (SD)` ~ ave + ' (' + std + ')',
        Median ~ paste(med),
        `Min, Max` ~ min + ', ' + max
      ),
      fac = list(
        ` ` ~ ifelse(sum == 0, '0', sum + ' (' + pct + '%' + ')')
      )
    )
    bundle <- c(x, extra)
    bundle
  })

  summarized <- reactive({
    printer('summarized')
    fun <- tablet
    if(conf$sequential) fun <- splice
    args <- args()
    do.call(fun,args)
  })

  html <- reactive({
    printer('html')
   # options(knitr.kable.NA = conf$na_string)
    options(knitr.kable.NA = 0)
    # browser()
    fun <- tablet
    if(conf$sequential) fun <- splice
    args <- args()

    if(!is.null(input$labelhtml)){
      if(input$labelhtml == 'yes'){
        args$x %<>% modify(title = .data$html)
      }
    } else {
      printer('no labelhtml yet')
      return()
    }
    x <- do.call(fun, args)
    x %<>% tablette # 0.6.0 revert to old format
    # browser()
    # remove NA groups
    na <- which(names(x) == 'NA')
    for(i in rev(na))x[[na]] <- NULL

    # strikethru imputed columns for visual clarity
    codelist <- attr(x$`_tablet_name`, 'codelist')
    x$`_tablet_original` <- unlist(codelist[x$`_tablet_name`])
    # very elegant, but blows away attributes
    # x %<>% mutate(
    #   across(
    #     .cols = -starts_with('_tablet_'),
    #     .fns = ~ ifelse(`_tablet_original` %in% names(conf$imputed), '-', .x)
    #   )
    # )
    nms <- names(x)
    nontargets <- grepl('^_tablet_', nms)
    targets <- !nontargets
    #targets <- x %>% select(-starts_with('_tablet_')) %>% names
    imputed <- x$`_tablet_original` %in% names(conf$imputed)
    if(length(imputed) & length(targets)) x[imputed, targets] <- '-'
    x$`_tablet_original` <- NULL
    x %<>% tablet # 0.6.0 new format
    x %<>% as_kable(caption = conf$title)
    x %<>% kable_classic(full_width = F, html_font = "Cambria")
    x %<>% kable_styling(fixed_thead = T)
    x
  })

  tex <- reactive({
    printer('tex')
    #browser()
    old <- opts_knit$get('out.format')
    opts_knit$set(out.format = 'latex')
    # options(knitr.kable.NA = escape_latex(conf$na_string))
    options(knitr.kable.NA = 0)

    fun <- tablet
    if(conf$sequential) fun <- splice
    args <- args()
    # browser()
    if(!is.null(input$labeltex)){
      # browser()
      if(input$labeltex == 'yes'){
        printer('using spork')
        args$x %<>% modify(title = .data$tex) # should have class 'latex', unescaped
        #args$x %<>% modify(codelist = lapply(codelist, kableExtra:::escape_latex2))
      } else {
        # args$x %<>% modify(title = kableExtra:::escape_latex(title))
        # otherwise trap specials and pre-double secondary backslash
        args$x %<>% modify(title = tablet::escape_latex(title))
        #args$x %<>% modify(codelist = lapply(codelist, kableExtra:::escape_latex2))
      }
    } else {
      printer('no labeltex yet')
      return()
      # next maybe unnecessary if as_kable auto-escapes names(index) in >= 0.4.2
      # args$x %<>% modify(title = kableExtra:::escape_latex(title))

    }

    # call tablet
    x <- do.call(fun, args)
    x %<>% tablette # 0.10.21 revert to old format
    # remove NA groups
    na <- which(names(x) == 'NA')
    for(i in rev(na))x[[na]] <- NULL

    # strikethru imputed columns for visual clarity
    codelist <- attr(x$`_tablet_name`, 'codelist')
    x$`_tablet_original` <- unlist(codelist[x$`_tablet_name`])
    # very elegant, but blows away attributes
    # x %<>% mutate(
    #   across(
    #     .cols = -starts_with('_tablet_'),
    #     .fns = ~ ifelse(`_tablet_original` %in% names(conf$imputed), '-', .x)
    #   )
    # )
    nms <- names(x)
    nontargets <- grepl('^_tablet_', nms)
    targets <- !nontargets
    #targets <- x %>% select(-starts_with('_tablet_')) %>% names
    imputed <- x$`_tablet_original` %in% names(conf$imputed)
    if(length(imputed) & length(targets)) x[imputed, targets] <- '-'
    x$`_tablet_original` <- NULL
    if(!nrow(x)){
      showNotification(duration = 5, type = 'error', 'nothing selected')
      return(character(0))
    }
    #browser()
    # _tablet_name has been thoroughly pre-escaped for all cases.
    # however, it is created as factor.
    # we flag it as latex to invoke the right method in as_kable(escape_latex = tablet::escape_latex)
    x$`_tablet_name` %<>% as_latex
    x %<>% tablet # 0.10.21 new format
    # 0.7.0 multi-line title
    theTitle <- escape_latex(conf$title)
    # theTitle %<>% gsub('\\n', '\\\\newline ', .)
    # theTitle <- paste('\\parbox{0.8\\textwidth}{\\centering ', theTitle, '}')
    x %<>% as_kable(format = 'latex', caption = theTitle, longtable = TRUE)
    if(length(input$repeatheader) == 1){
      if(input$repeatheader == 'yes'){
        x %<>% kable_styling(latex_options = 'repeat_header', repeat_header_text = '')
      }
    }
    feet <- unlist(strsplit(conf$footnotes, '\n'))
    if(length(feet)){
    x %<>% footnote(general = feet ,fixed_small_size = TRUE, general_title = " ",threeparttable = TRUE)
    }
    x %<>% as.character

    # insert footnote on every page

    cont <- input$cont
    mycont <- NULL
    if(!is.null(cont)){
      if(nchar(cont) > 0){
        mycont <- c(
          paste0('\\multicolumn{1}{r}{\\emph{', cont, '}}\\\\'),
          '\\midrule'
        )
      }
    }

    insertion <- c(
      '\\endhead',
      '\\midrule',
      mycont,
      '\\insertTableNotes'
    )
    insertion <- paste(insertion, collapse = '\n')
    if(length(input$repeatfootnote) == 1){
      if(input$repeatfootnote == 'yes'){
        x %<>% sub('\\endhead', insertion, ., fixed = TRUE)
      }
    }
    x %<>% as.document(
      thispagestyle = '',
      pagestyle = '',
      preamble = c(
        '\\documentclass{article}',
        '\\usepackage[utf8]{inputenc}',
        '\\usepackage[T1]{fontenc}',
        '\\usepackage[showseconds=false]{datetime2}',
        '\\usepackage[landscape]{geometry}',
        '\\usepackage{fancyhdr}',
        '\\fancyhf{}',
        '\\renewcommand{\\headrulewidth}{0pt}',
        '\\pagestyle{fancy}',
        paste0('\\lhead{', escape_latex(conf$lhead1),' \\\\ ',escape_latex(conf$lhead2), '}'),
        '%\\chead{Table 0.0.0.xxx}',
        paste0('\\rhead{', escape_latex(conf$rhead1),' \\\\ ',escape_latex(conf$rhead2), '}'),
        #paste0('\\lfoot{\\textit{',file_path_sans_ext(conf$filepath),'}}'),
        paste0('\\lfoot{\\textit{~', sub(getwd(),'',conf$confpath, fixed = TRUE),'}}'),

        '\\rfoot{\\today{~at~\\DTMcurrenttime}}',
        '\\usepackage{booktabs}',
        '\\usepackage{longtable}',
        '\\usepackage{array}',
        '\\usepackage{multirow}',
        '\\usepackage{wrapfig}',
        '\\usepackage{float}',
        '\\usepackage{colortbl}',
        '\\usepackage{pdflscape}',
        '\\usepackage{tabu}',
        '\\usepackage{threeparttable}',
        '\\usepackage{threeparttablex}',
        '\\usepackage[normalem]{ulem}',
        '\\usepackage{xcolor}',
        '\\usepackage[labelformat=empty]{caption}',
        '\\usepackage{makecell}'
      )
    )
    opts_knit$set(out.format = old)
    x
  })

  output$buckets <- renderUI({
    printer('output$buckets')
    if(!length(conf$x))return()
    nms <- names(conf$x)
    selected <- intersect(conf$selected, nms)
    group_by <- intersect(conf$group_by, nms)
    filter_by <-intersect(conf$filter_by, nms)

    used <- union(selected, group_by)
    used <- union(used, filter_by)
    available <- sort(setdiff(nms, used)) # definitive set
    suggested <- union(input$available, available) # user's sort order
    available <- intersect(suggested, available) # defer to user where possible

    bucket_list(
      header = 'Data Item Roles',
      group_name = 'bucket_list_group',
      orientation = 'horizontal',
      add_rank_list(
        text = 'Available',
        labels = available,
        input_id = 'available'
      ),
      add_rank_list(
        text = 'Summarize',
        labels = selected,
        input_id = 'selected'
      ),
      add_rank_list(
        text = 'Group By',
        labels = group_by,
        input_id = 'group_by'
      ),
      add_rank_list(
        text = 'Filter By',
        labels = filter_by,
        input_id = 'filter_by'
      )
    )
  })

  output$splice <- renderUI({
    printer('output$splice')
    radioButtons(
      'splice',
      'Grouping Style',
      inline = TRUE,
      choices = c('nested','sequential'),
      selected = ifelse(conf$sequential,'sequential','nested')
    )

  })

  output$repeatheader <- renderUI({
    printer('output$repeatheader')
    radioButtons(
      'repeatheader',
      'repeat header on each page',
      inline = TRUE,
      choices = c('no','yes'),
      selected = conf$repeathead
    )
  })

  output$repeatfootnote <- renderUI({
    printer('output$repeatfootnote')
    radioButtons(
      'repeatfootnote',
      'repeat footnote on each page',
      inline = TRUE,
      choices = c('no','yes'),
      selected = conf$repeatfoot
    )
  })

  output$labelhtml <- renderUI({
    printer('output$labelhtml')
    radioButtons(
      'labelhtml',
      'scripted labels',
      inline = TRUE,
      choices = c('no','yes'),
      selected = conf$labelhtml
    )
  })

  output$labeltex <- renderUI({
    printer('output$labeltex')
    radioButtons(
      'labeltex',
      'scripted labels',
      inline = TRUE,
      choices = c('no','yes'),
      selected = conf$labeltex
    )
  })

  output$caption <- renderUI({
    printer('output$caption')
    textAreaInput('caption','Title', value = conf$title, resize = 'both')
  })

  output$outputid <- renderUI({
    printer('output$outputid')
    textInput('outputid','Output Identifier', value = conf$outputid)
  })

  output$lhead1 <- renderUI({
    printer('output$lhead1')
    textInput('lhead1','Left Header 1', value = conf$lhead1)
  })

  output$lhead2 <- renderUI({
    printer('output$lhead2')
    textInput('lhead2','Left Header 2', value = conf$lhead2)
  })

  output$rhead1 <- renderUI({
    printer('output$rhead1')
    textInput('rhead1','Right Header 1', value = conf$rhead1)
  })

  output$rhead2 <- renderUI({
    printer('output$rhead2')
    textInput('rhead2','Right Header 2', value = conf$rhead2)
  })

  output$cont <- renderUI({
    printer('output$cont')
    textInput('cont','Continued', value = conf$cont)
  })

  output$footnotes <- renderUI({
    printer('output$footnotes')
    textAreaInput('footnotes','Footnotes', value = conf$footnotes, resize = 'both')
  })

  # output$na_string <- renderUI({
  #   textInput('na_string','text substitute for NA', value = conf$na_string)
  # })

  output$keep <- renderUI({
    printer('output$keep')
    if(!length(input$filter_by))return()

    myFilter <- function(var, dat){
      nms <- as.character(sort(unique(dat[[var]])))
      lbl <- attr(dat[[var]], 'label')

      checkboxGroupInput(
        inline = TRUE,
        inputId = paste0('mesa_filter_',var),
        label = lbl,
        choices = nms,
        selected = conf$keep[[var]]
      )
    }

    myObserver <- function(var){
      observers[[var]] <<- observeEvent(input[[paste0('mesa_filter_',var)]], {
        conf$keep[[var]] <- input[[paste0('mesa_filter_',var)]]
      })
    }

    # pre-assign an observer if not already
    lapply(input$filter_by, myObserver)

    lapply(input$filter_by, myFilter, dat = conf$x)
  })

  output$data <- reactable::renderReactable({
    printer('output$data')
    if(!ncol(conf$x))return(structure(data.frame(` `='data goes here.', check.names = F), row.names = ' '))
    out <- conf$x
    #out %<>% resolve # already done
    #browser()
    out %<>% modify(name = paste(name, label, sep = ': '))
    hasUnits <- sapply(out, function(col)'units' %in% names(attributes(col)))
    hasUnits <- names(hasUnits[hasUnits])
    #browser()
    out %<>% modify(!!!hasUnits, name = paste0(name, ' (', .data$units, ')'))
    reactable(out)
  })

  output$preview <- renderText({
    #printer('output$preview')
    if(!length(input$selected))return('Output goes here.')
    # ensure html
    opts_knit$set(out.format = 'html')
    x <- suppressMessages(html())
    x
  })

  pdf_location <- reactive({
    printer('pdf_location')
    # browser()
    x <- suppressWarnings(tex())
    if(!length(x))return('1x1.png')
    stem <- isolate(conf$outputid) # basename(tempfile())

    # backup
    writeLines(x, con = 'www/cache.tex')

    # clean slate
    tex <- file.path('www', paste0(stem, '.tex'))
    pdf <- file.path('www', paste0(stem, '.pdf'))
    unlink(tex)
    unlink(pdf)
    
    # # some tables need to be run twice!  Not sure why!
    # # particularly for repeat headers with nesting.
    # #browser()
    # path <- try(
    #   as.pdf(
    #     x,
    #     dir = 'www',
    #     stem = stem,
    #     clean = FALSE,
    #     ignore.stdout = TRUE
    #   )
    # )
    # 
    # # ignore incomplete pdf
    # unlink(file.path('www', paste0(stem, '.pdf')))
    # 
    # path <- try(
    #   as.pdf(
    #     x,
    #     dir = 'www',
    #     stem = stem,
    #     clean = TRUE,
    #     ignore.stdout = TRUE
    #   )
    # )
    # 0.6.0: trying tinytex instead of system command in latexpdf::as.pdf.document
    # 0.6.0: must now write the tex file manually.
    writeLines(x, tex)
    path <- try(tinytex::pdflatex(tex))

    if(inherits(path, 'try-error')){
      showNotification(as.character(path), type = 'error', duration = 5)
    }

    if(!file.exists(path)) return('1x1.png')
    basename(path)

  })

  # https://stackoverflow.com/questions/19469978/displaying-a-pdf-from-a-local-drive-in-shiny

  output$pdfview <- renderUI({
    printer('output$pdfview')
    if(!ncol(conf$x))return('PDF displays here.')
    loc <- pdf_location()
    printer('directory')
    printer(getwd())
    printer('file')
    printer(loc)
    tags$iframe(
      style="height:600px; width:100%; scrolling:yes",
      src = paste0('/', loc)
    )
  })

  output$confpath <- renderPrint({
    #printer('output$confpath')
    req(conf$confpath)
    cat(conf$confpath)
    # if(!length(path)){
    #   cat('No configuration selected.')
    # } else {
    #   if(is.na(path)){
    #     cat('No configuration selected.')
    #   } else {
    #     cat(path)
    #   }
    # }
  })

  output$metapath <- renderPrint({
    #printer('output$metapath')
    if (!length(conf$metapath)) {
      cat('No data selected.')
    } else {
      cat(conf$metapath)
    }
  })
  
  output$deripath <- renderPrint({
    #printer('output$deripath')
    if (!length(conf$deripath)) {
      cat('No data selected.')
    } else {
      cat(conf$deripath)
    }
  })
  
  # https://stackoverflow.com/questions/54304518/how-to-edit-a-yml-file-in-shiny

  output$saveMeta <- renderUI({
    printer('output$saveMeta')
    actionButton("saveMeta", label = "Save")
  })

  output$saveDeri <- renderUI({
    printer('output$saveDeri')
    actionButton("saveDeri", label = "Save")
  })
  
  output$meta <- renderUI({
    printer("output$meta")
    current <- conf$editor
    if(!length(conf$metapath))return(current)
    if(is.na(conf$metapath))return(current)
    val <- NULL
    tryCatch(
      val <- readLines(conf$metapath),
      error = function(e) showNotification(
        duration = NULL,
        type = 'error',
        as.character(e)
      )
    )
    if(!is.null(val)){
      val <- aceEditor(
        outputId = "meta",
        value = val,
        mode = 'yaml',
        tabSize = 2
      )
      conf$editor <- val
    } else {
      val <- conf$editor
    }
    val
  })
  
  output$deri <- renderUI({
    printer("output$deri")
    #browser()
    current <- conf$editor2
    if(!length(conf$deripath))return(current)
    if(is.na(conf$deripath))return(current)
    val <- NULL
    tryCatch(
      val <- readLines(conf$deripath),
      error = function(e) showNotification(
        duration = NULL,
        type = 'error',
        as.character(e)
      )
    )
    if(!is.null(val)){
      val <- aceEditor(
        outputId = "deri",
        value = val,
        mode = 'yaml',
        tabSize = 2
      )
      conf$editor2 <- val
    } else {
      val <- conf$editor2
    }
    val
  })
  
  observeEvent(input$saveMeta, {
    printer('observeEvent: input$saveMeta')
    path <- isolate(conf$metapath)
    res <- try(yaml.load(input$meta))
    if(!inherits(res,'try-error')){
      res <- try(read_yamlet(input$meta))
    }
    err <- as.character(res)
    res <- !inherits(res, 'try-error')
    msg <- paste(ifelse(res, 'wrote', 'did not write'), path)
    if(!res) msg <- paste(msg, err)
    if(res){
      write(x = input$meta, file = path)
      # trigger redecoration
      conf$mv <- conf$mv + 1
    }
    dur <- 10
    if(res) dur <- 5
    showNotification(
      duration = dur,
      type = ifelse(res, 'default', 'error'),
      ui = msg
    )
  })
  
  observeEvent(input$saveDeri, {
    
    printer('observeEvent: input$saveDeri')
    path <- isolate(conf$deripath)
    res <- try(yaml.load(input$deri))
    if(!inherits(res,'try-error')){
      res <- try(read_yamlet(input$deri))
    }
    err <- as.character(res)
    res <- !inherits(res, 'try-error')
    msg <- paste(ifelse(res, 'wrote', 'did not write'), path)
    if(!res) msg <- paste(msg, err)
    if(res){
      write(x = input$deri, file = path)
      # trigger redecoration ?
      conf$dv <- conf$dv + 1
    }
    dur <- 10
    if(res) dur <- 5
    showNotification(
      duration = dur,
      type = ifelse(res, 'default', 'error'),
      ui = msg
    )
  })
})
# Create Shiny app ----
shinyApp(ui, server)

# copyright 2021 Tim Bergsma bergsmat@gmail.com

#' ---
#' title: "Test"
#' author: "Tim Bergsma"
#' date: "`r Sys.Date()`"
#' output:
#'   rmarkdown::pdf_document:
#'     toc: true
#'     keep_tex: true
#' ---

library(tablet)
library(haven)
library(yamlet)
library(magrittr)
library(dplyr)
library(kableExtra)
library(knitr)

# make adsl with imputed bmi, imputed race, and two-row footnote
x <- read_sas('adsl.sas7bdat')
m <- read_yamlet('adsl.yaml')

# https://github.com/haozhu233/kableExtra/issues/703
names(m$race$guide)[[3]] <- 'Oriental'

# fortify to mimic app.R
have <- names(x)
need <- names(m)
make <- setdiff(need, have)
for(col in make) x[[col]] <- rep(NA_integer_, nrow(x))

# ensure positive nrow
if(nrow(x) == 0) x <- x['',,drop = FALSE]

# drop unspecified
x %<>% select(!!!names(m))

# apply meta
x <- redecorate(x, m)

# # Promote NA to a level of the factor
# x %<>% resolve(exclude = NULL)
x %<>% resolve()

foot <-
'a clinicaltrial.gov
b some other comment'
options(knitr.kable.NA = 0)
#opts_knit$set(out.format = 'latex')
# debug(tablet:::widgets.devalued)
#x %>% group_by(trt01a, trt01aa) %>% select(race) %>% tablet
x$trt01a[] <- NA
x$trt01aa[] <- NA
#debug(categoricals)
#debug(numerics)
#debug(groupfull)
x <- x %>%
  filter(saffl == 'Y') %>%
  group_by(trt01a, trt01aa) %>%
  select(
  #  age, agegr, sex, weight, bmi,
    race, bmi
  ) %>%
  splice(
    all_levels = TRUE,
    fun = list(
      sum ~  sum(x,  na.rm = TRUE),
      pct ~ signif(digits = 3,     sum / n * 100        ),
      ave ~ signif(digits = 3,    mean(x,  na.rm = TRUE)),
      std ~ signif(digits = 3,      sd(x,  na.rm = TRUE)),
      med ~ signif(digits = 3,  median(x,  na.rm = TRUE)),
      min ~ signif(digits = 3,     min(x,  na.rm = TRUE)),
      max ~ signif(digits = 3,     max(x,  na.rm = TRUE)),
      smn ~ sum(!is.na(x))
    ),
    num = list(
      n ~ smn,
      `Mean (SD)` ~  ave + ' (' + std + ')',
      Median ~  paste(med),
      `Min, Max` ~  min + ', ' + max
    ),
    fac = list(
      ` ` ~ ifelse(sum == 0, '0', sum + ' (' + pct + '%' + ')')
    )
  )

# remove NA groups
na <- which(names(x) == 'NA')
for(i in rev(na))x[[na]] <- NULL

# reverse lookup on make
codelist <- attr(x$`_tablet_name`, 'codelist')
x$`_tablet_original` <- unlist(codelist[x$`_tablet_name`])
# very elegant, but blows away attributes
# x %<>% mutate(
#   across(
#     .cols = -starts_with('_tablet_'),
#     .fns = ~ ifelse(`_tablet_original` %in% names(conf$imputed), '-', .x)
#   )
# )
targets <- seq_along(x)[!(grepl('_tablet_', names(x)))]
imputed <- x$`_tablet_original` %in% make
if(length(imputed) & length(targets)) x[imputed, targets] <- '-'
x$`_tablet_original` <- NULL

x %>%
  as_kable %>%
  footnote(
    general = # escape_latex(
      c('a something','b something')
      # )
    ,
    fixed_small_size = TRUE,
    general_title = " ",
    threeparttable = TRUE
  ) %>%
  kable_styling(latex_options = 'scale_down')

globalVariables(c(
   '_tablet_level', 'n','N', '.data','_tablet_value',
   '_tablet_N','_tablet_n', '_tablet_x',
   '_tablet_devalued_N','_tablet_devalued_n',
   '_tablet_widget', '_tablet_name', '_tablet_groupless',
   '_tablet_stat',
   '_tablet_groupfull','where',
   '_tablet_class'
))

#' Identify Categories
#'
#' Identifies categories.  Generic, with method \code{\link{categoricals.data.frame}}.
#' @param x object
#' @param ... passed
#' @export
#' @return see methods
#' @keywords internal
#' @examples
#' example(classifiers)
#' categoricals(x)
#'
categoricals <- function(x, ...)UseMethod('categoricals')

#' Identify Numerics
#'
#' Identifies numerics.  Generic, with method \code{\link{numerics.data.frame}}.
#' @param x object
#' @param ... passed
#' @export
#' @return see methods
#' @keywords internal
#' @examples
#' example(classifiers)
#' numerics(x)
numerics <- function(x, ...)UseMethod('numerics')

#' Identify Observations
#'
#' Identifies observations.  Generic, with method \code{\link{observations.data.frame}}.
#' @param x object
#' @param ... passed
#' @export
#' @return see methods
#' @keywords internal
#' @examples
#' example(classifiers)
#' observations(x)
observations <- function(x, ...)UseMethod('observations')

#' Identify Classifiers
#'
#' Identifies classifiers.  Generic, with method \code{\link{classifiers.data.frame}}.
#' @param x object
#' @param ... passed
#' @export
#' @return see methods
#' @keywords internal
#' @examples
#' library(dplyr)
#' x <- data.frame(grp = 1:2, a = factor(1:2), b = 1:2, c = factor(1:2, levels = 2:1))
#' x <- group_by(x, grp)
#' classifiers(x)
classifiers <- function(x, ...)UseMethod('classifiers')

#' Identify Classifiers for Data Frame
#'
#' Identifies classifiers: literally, groups. Supplies tablet_n and tablet_N: groupwise
#' and ungrouped counts.
#' @param x data.frame
#' @param ... ignored
#' @importFrom dplyr groups ungroup tally add_count select group_by
#' @export
#' @keywords internal
#' @return data.frame, grouped if x was (see \code{\link[dplyr]{group_by}})

classifiers.data.frame <- function(x, ...){
   for(i in c('_tablet_N','_tablet_n'))if(i %in% names(x))stop('x cannot include ', i)
   grp <- groups(x)
   x <- ungroup(x)
   x <- add_count(x, name = '_tablet_N') #, wt = n())
   suppressMessages(
   x <- select(x, `_tablet_N`, everything())
   )
   for(col in grp){
      #nm <- paste0('n_', col)
      x <- group_by(x, .data[[col]], .add = TRUE)
      #x <- add_tally(x, name = nm, wt = n())
   }
   x <- add_tally(x, name = '_tablet_n') # , wt = n())
   x
}

#' Identify Categoricals for Data Frame
#'
#' Identifies categorical columns: literally, factors. Stacks factor levels and supplies value = 1.
#' @param x data.frame; names must not start with '_tablet_'
#' @param ... passed to \code{\link{classifiers}}
#' @param na.rm_fac whether to drop NA observations; passed to \code{\link[tidyr]{gather}} as na.rm
#' @param exclude_fac which factor levels to exclude; see \code{\link{factor}} (exclude)
#' @param all_levels whether to supply records for unobserved levels
#' @importFrom dplyr groups ungroup add_tally add_count select group_by mutate
#' @importFrom tidyr gather replace_na
#' @importFrom dplyr all_of across everything full_join anti_join
#' @importFrom magrittr %>% %<>%
#' @export
#' @method categoricals data.frame
#' @keywords internal
#' @return same class as x
#' @examples
#' example(classifiers)
#' categoricals(x)
#' levels(categoricals(x)$`_tablet_level`)

categoricals.data.frame <- function(
  x,
  ...,
  na.rm_fac = FALSE,
  exclude_fac = NULL,
  all_levels = FALSE
){
  y <- x
  for(i in c('_tablet_name','_tablet_level','_tablet_value'))if(i %in% names(x))stop('names x cannot include ',i)
  if(any(duplicated(names(x))))stop('names cannot be duplicated')
  x <- select(x, !!!groups(x),where(is.factor))
  var <- setdiff(names(x), sapply(groups(x), rlang::as_string))
  for(i in var){
    # prepend each factor level with column name for uniqueness
    levels(x[[i]]) <- paste(sep = '_tablet_', i, levels(x[[i]]))
  }
  lev <- unlist(lapply(select(ungroup(x), all_of(var)), levels))

  # if('tablet_numeric' %in% lev)stop('levels of factors in x cannot include \'tablet_numeric\'')
  x <- classifiers(x,...)
  x <- mutate(x, across(where(is.factor), as.character))
  x <- suppressWarnings(
    # attributes not identical across groups, they will be dropped
    # this is where _tablet_name is created
    gather(
      x,
      factor_key = TRUE,
      key = '_tablet_name',
      value = '_tablet_level',
      var,
      na.rm = na.rm_fac
    )
  )
  # possibly no var
  if(length(var) == 0){
    x <- slice(x, 0)
    x <- mutate(
      x,
      # _tablet_name created here too
      `_tablet_name` = factor(), # was character() before 0.4.3
      `_tablet_level` = factor()
    )
  }
  x <- mutate(x, `_tablet_level` = factor(`_tablet_level`, levels = lev, exclude = exclude_fac))
  x <- mutate(x, `_tablet_value` = 1L) # bogus value
  # missing observation of some category is moot if na.rm = TRUE
  # if exclude is NULL, NA is a distinct category and value should be available for tally
  # but if na not removed and not categorized, it should be passed to summary functions as na
  # ???
  # if(!na.rm){
  #    if(!is.null(exclude)){
  #       x <- mutate(x, value = ifelse(is.na(level), NA_integer_, value))
  #    }
  # }

  if(!all_levels) return(x)
  if(length(var) == 0) return(x) # at 0.5.7
  # if(!length(groups(x))) return(x) # strike at 0.5.7?
  # what groups must be populated?

  template <- function(x, ...)UseMethod('template')
  template.factor <- function(x,...){
    # create vector of all levels
    levs <- levels(x)
    x <- x[0]
    for(i in seq_along(levs))x[i] <- levs[i]
    x
  }
  groups <-  x %>% select(c(!!!groups(x),`_tablet_level`)) #, `_tablet_N`, `_tablet_n`)) %>% unique
  groups <- groups[0,,drop = FALSE]
  groups <- lapply(groups, template.factor)
  # names(groups) <- groups(x)
  for(g in names(groups)){
    groups[[g]] <- data.frame(groups[[g]])
    names(groups[[g]]) <- g
  }
  while (length(groups) > 1){
    groups[[1]] <- full_join(by = character(), groups[[1]], groups[[2]])
    groups[[2]] <- NULL
  }
 groups <- groups[[1]]
 groups$`_tablet_N` <- rep(max(0, unique(x$`_tablet_N`)), nrow(groups))
#  groups %<>% left_join(x %>% select(c(!!!groups(x)), `_tablet_n`, `_tablet_level`))
 zz <- x %>% select(c(!!!groups(x)), `_tablet_n`) %>% unique # at 0.5.7
 common <- intersect(names(groups), names(zz))
 groups %<>% left_join(zz, by = common)

 groups$`_tablet_n` %<>% replace_na(0L)

  # what factors are in play?
 factors <- y %>% select(where(is.factor)) %>% names
 factors %<>% setdiff(groups(y))
 `_tablet_name` <- character(0)
 `_tablet_level` <- character(0)
 for(i in factors){
   levs <- levels(y[[i]])
   for(lev in levs){
     `_tablet_name` %<>% append(i)
     `_tablet_level` %<>% append(paste0(i, '_tablet_', lev))
   }
 }
 stopifnot(length(`_tablet_name`) == length(`_tablet_level`))
 factors <- data.frame(`_tablet_name`,`_tablet_level`, check.names = F)
 factors$`_tablet_name` %<>% factor(levels = levels(x$`_tablet_name`))
 factors$`_tablet_level` %<>% factor(levels = levels(x$`_tablet_level`))
 suppressMessages(
 #  groups %<>% full_join(factors, by = character())
   groups %<>% full_join(factors) # at 0.5.7
 )
 suppressMessages(
 groups %<>% anti_join(x %>% select(-c(`_tablet_N`, `_tablet_n`))) # exclude combos already observed
 )
 groups$`_tablet_value` <- rep(0L, nrow(groups))
 stopifnot(setequal(names(groups), names(x)))
 x %<>% bind_rows(groups)
 x %<>% arrange(!!!groups(x), `_tablet_name`, `_tablet_level`)
 x

}

#' Identify Numerics for Data Frame
#'
#' Identifies numeric columns. Stacks values and supplies factor level 'numeric'.
#' @param x data.frame; names must not include 'name' or 'level'
#' @param ... passed to \code{\link{classifiers}}
#' @param na.rm_num whether to drop NA observations; passed to \code{\link[tidyr]{gather}} as na.rm
#' @param all_levels whether to supply records for unobserved levels
#' @importFrom dplyr groups ungroup tally add_count select group_by mutate slice
#' @importFrom tidyr gather
#' @export
#' @method numerics data.frame
#' @keywords internal
#' @return same class as x
numerics.data.frame <- function(x, ..., na.rm_num = FALSE, all_levels = FALSE){
   for(i in c('_tablet_level', '_tablet_name','_tablet_value'))if(i %in% names(x))stop('names x cannot include ',i)

   y <- x
   x <- select(x, !!!groups(x),where(is.numeric))
   var <- setdiff(names(x), sapply(groups(x),rlang::as_string))
   x <- classifiers(x,...)
   x <- mutate(x, `_tablet_level` = factor('numeric')) # bogus level
   x <- suppressWarnings(
      # attributes not identical across groups, they will be dropped
      # this is also where _tablet_name is created
      gather(
         x,
         factor_key = TRUE,
         key = '_tablet_name',
         value = '_tablet_value',
         var,
         na.rm = na.rm_num
      )
   )
   if(!length(var)){
      x <- slice(x, 0)
      x$`_tablet_name`<- factor()
      x$`_tablet_value` <- numeric(0)
   }

   if(!all_levels)return(x)
   if(!length(groups(x)))return(x) # ? at 0.5.7

   # what groups must be populated?

   template <- function(x, ...)UseMethod('template')
   template.factor <- function(x,...){
     # create vector of all levels
     levs <- levels(x)
     x <- x[0]
     for(i in seq_along(levs))x[i] <- levs[i]
     x
   }
   groups <-  x %>% select(c(!!!groups(x))) #, `_tablet_N`, `_tablet_n`)) %>% unique
   groups <- groups[0,]
   groups <- lapply(groups, template.factor)
   names(groups) <- groups(x)
   for(g in names(groups)){
     groups[[g]] <- data.frame(groups[[g]])
     names(groups[[g]]) <- g
   }
   while (length(groups) > 1){
     groups[[1]] <- full_join(by = character(), groups[[1]], groups[[2]])
     groups[[2]] <- NULL
   }
   groups <- groups[[1]]
   groups$`_tablet_N` <- max(0, unique(x$`_tablet_N`))
   suppressMessages(
   groups %<>% left_join(x %>% select(c(!!!groups(x)), `_tablet_n`))
   )
   groups$`_tablet_n` %<>% replace_na(0L)

   # what numerics are in play?
   numerics <- y %>%
     ungroup %>%
     select(-where(is.factor)) %>%
     names


   numerics %<>% setdiff(groups(y))
   `_tablet_name` <- character(0)
   `_tablet_level` <- character(0)
   for(i in numerics){
     `_tablet_name` %<>% append(i)
     `_tablet_level` %<>% append('numeric')

   }
   stopifnot(length(`_tablet_name`) == length(`_tablet_level`))
   numerics <- data.frame(`_tablet_name`,`_tablet_level`, check.names = F)
   numerics$`_tablet_name` %<>% factor(levels = levels(x$`_tablet_name`))
   numerics$`_tablet_level` %<>% factor(levels = levels(x$`_tablet_level`))
   suppressMessages(
   groups %<>% full_join(numerics, by = character())
   )
   suppressMessages(
   groups %<>% anti_join(x %>% select(-c(`_tablet_N`, `_tablet_n`))) # exclude combos already observed
   )
  # groups$`_tablet_value` <- rep(0L, nrow(groups))
   groups$`_tablet_value` <- rep(NA_real_, nrow(groups))

   stopifnot(setequal(names(groups), names(x)))
   x %<>% bind_rows(groups)
   x %<>% arrange(!!!groups(x), `_tablet_name`, `_tablet_level`)
   x
}

#' Identify Observations for Data Frame
#'
#' Identifies observations. Stacks values by level, supplying defaults as necessary.
#' @param x data.frame
#' @param ... passed to categoricals(), numerics()
#' @param exclude_name whether to drop NA levels of name; passed to \code{\link[tidyr]{gather}}
#' @importFrom dplyr groups ungroup tally add_count select group_by mutate arrange bind_rows
#' @importFrom tidyr gather
#' @export
#' @keywords internal
#' @return class 'observations' arranged by groups (presented first):
#' \item{_tablet_N}{number of records}
#' \item{_tablet_n}{number of records in group}
#' \item{_tablet_name}{observation identifier}
#' \item{_tablet_level}{factor level (or special value 'numeric' for numerics)}
#' \item{_tablet_value}{numeric value (or special value 1 for factors)}

observations.data.frame <- function(x, ..., exclude_name = NULL){
   x <- select(x, !!!groups(x), everything()) # groups first for consistency
   nms <- names(x)
   y <- categoricals(x, ...)
   z <- numerics(x, ...)
   x <- bind_rows(y, z)
   # x <- mutate(x, name = factor(exclude = exclude_name, name, levels = intersect(nms, levels(name))))
   
   
   # try to make output order match input across fac, num
   # in a mutate call, 
   # next step operates once per group
   # groups may not each have all values
   # resulting level steps are combined in group order
   # yielding unpredictable results
   # we avoid mutate
   
   levs <- intersect(nms, unique(as.character(x$`_tablet_name`)))
   x$`_tablet_name` <- factor(x$`_tablet_name`, exclude = exclude_name, levels = levs)
   # x <- mutate(x,`_tablet_name` = factor(exclude = exclude_name,`_tablet_name`,levels = intersect( nms, unique(as.character(`_tablet_name`)))))
   
   x <- group_by(x, `_tablet_name`, .add = TRUE)
   x <- group_by(x, `_tablet_level`, .add = TRUE)
   x <- arrange(x, .by_group = TRUE)
   class(x) <- union('observations', class(x))
   x
}

# Reduce a List to its Formula Members
# 
# Reduces a List to its members which are formulas with LHS.
# Captures LHS as string and returns a thus-named list.
# @importFrom rlang is_formula f_lhs as_string f_rhs
# @examples
# .named_formulas(
#   list(
#      3,
#      'x',
#      ~ 3,
#      x ~ 3,
#      y ~ NULL
#   )
# )
# 

.named_formulas <- function(x, ...){
  stopifnot(is.list(x))
  is.formula <- sapply(x, function(f)rlang::is_formula(f, scoped = TRUE, lhs = TRUE))
  if(!length(is.formula)){
    is.formula <- logical(0)
  }
  x <- x[is.formula]
  LHS <- sapply(x, rlang::f_lhs)
  LHS <- sapply(LHS, rlang::as_string)
  names(x) <- LHS
  x
}

# Identify NULL Formulas in a List
# 
# Identifies by name the named members of a list of formulas
# whose RHS are NULL
# @examples
# .null_formulas(
#   .named_formulas(
#     list(
#        3,
#        'x',
#        ~ 3,
#        x ~ 3,
#        y ~ NULL
#     )
#   )
# )
# 

.null_formulas <- function(x, ...){
  stopifnot(is.list(x))
  x <- .named_formulas(x)
  isNull <- sapply(x, function(f)rlang::is_null(rlang::f_rhs(f)))
  if(!length(isNull)){
    isNull <- logical(0)
  }
  x <- x[isNull]
  names(x)
}

#' Aggregate Values
#'
#' Aggregated values.  Generic, with method \code{\link{devalued.observations}}.
#' @param x object
#' @param ... passed
#' @export
#' @return see methods
#' @keywords internal
#' @examples
#' example(classifiers)
#' devalued(observations(x))
devalued <- function(x, ...)UseMethod('devalued')

#' Aggregate Values for Observations
#' Aggregates values for observations.
#' Accepts a list of formulas with result name on left and
#' aggregating expression on the right.
#' Behavior is undefined if any expression does not aggregate! (I.e., length != 1).
#' Expressions are evaluated in an environment where values are available as 'x',
#' grouped count is available as 'n' and ungouped count is available as 'N'.
#'
#'
#' @param x observations
#' @param ... passed formulas with matching LHS will replace corresponding element of fun.
#' @param fun default aggregate functions expressed as formulas
#' @param silent whether to suppress warnings from evaluations of 'fun'
#' @importFrom rlang f_lhs f_rhs eval_tidy :=
#' @importFrom dplyr mutate distinct rename
#' @export
#' @keywords internal
#' @return class 'devalued', presumably one record per group:
#' \item{_tablet_N}{number of records}
#' \item{_tablet_n}{number of records in group}
#' \item{_tablet_name}{observation identifier}
#' \item{_tablet_level}{factor level (or special value 'numeric' for numerics)}
#' \item{(other)}{additional column for each statistic in 'fun'}
devalued.observations <- function(
   x,
   ...,
   fun = list(
      mis ~ sum(is.na(x)),
      sum ~ sum(x, na.rm = TRUE),
      pct ~ signif(digits = 3,    sum / n * 100       ),
      ave ~ signif(digits = 3,   mean(x, na.rm = TRUE)),
      std ~ signif(digits = 3,     sd(x, na.rm = TRUE)),
      med ~ signif(digits = 3, median(x, na.rm = TRUE)),
      min ~ signif(digits = 3,    min(x, na.rm = TRUE)),
      max ~ signif(digits = 3,    max(x, na.rm = TRUE))
   ),
   silent = TRUE
){
  nms <- names(x)
  
  # remove from fun all x ~ NULL in list(...)
  ext <- .named_formulas(list(...))
  fun <- .named_formulas(fun)
  nul <- .null_formulas(ext)
  fun <- fun[!(names(fun) %in% nul)]
  ext <- ext[!(names(ext) %in% nul)]
  
  # form <- sapply(extra, function(f)inherits(f, 'formula'))
  # if(length(form))extra <- extra[form]
  # LHS <- sapply(extra, rlang::f_lhs)
  # LHS <- sapply(LHS, rlang::as_string)
  
  # guard
  if('x' %in% names(x)){
     x <- rename(x, `_tablet_x` = x)
  }
  if('N' %in% names(x)){
     x <- rename(x, `_tablet_devalued_N` = N)
  }
  if('n' %in% names(x)){
     x <- rename(x, `_tablet_devalued_n` = n)
  }

  # supply
  x <- rename(x, x = `_tablet_value`)
  x <- rename(x, N = `_tablet_N`)
  x <- rename(x, n = `_tablet_n`)

  # evaluate
  for(lhs in names(fun)){
     this <- NULL
     if(lhs %in% names(ext)){
       this <- ext[[lhs]]
     } else {
       this <- fun[[lhs]]
     }
     if(is.null(this))stop('formula cannot be NULL')
     # lhs <- rlang::as_string(rlang::f_lhs(this))
     # if(lhs %in% names(ext)) this <- ext[[match(lhs, LHS)]]
     rhs <- rlang::f_rhs(this)
     if(silent){
        x <- suppressWarnings(mutate(x, !!lhs := rlang::eval_tidy(rhs)))
     } else {
        x <- mutate(x, !!lhs := rlang::eval_tidy(rhs))
     }
  }
  # for(i in seq_along(ns)){ # cleanup
  #    that <- paste0('n_', i)
  #    x[[that]] <- NULL
  # }
  x <- select(x, -x)

  # restore
  x <- rename(x, `_tablet_N` = N)
  x <- rename(x, `_tablet_n` = n)

  if('_tablet_x' %in% names(x)){
     x <- rename(x, x = `_tablet_x`)
  }
  if('_tablet_devalued_N' %in% names(x)){
     x <- rename(x, N = `_tablet_devalued_N`)
  }
  if('_tablet_devalued_n' %in% names(x)){
     x <- rename(x, n = `_tablet_devalued_n`)
  }
  x <- distinct(x)
  class(x) <- union('devalued', class(x))
  x
}

#' Calculate Widgets
#'
#' Calculates widgets.  Generic, with method \code{\link{widgets.devalued}}.
#' @param x object
#' @param ... passed
#' @export
#' @return see methods
#' @keywords internal
#' @examples
#' example(classifiers)
#' widgets(devalued(observations(x)))
widgets <- function(x, ...)UseMethod('widgets')
# https://adv-r.hadley.nz/quasiquotation.html

#' Calculate Widgets for Devalued
#'
#' Calculates widgets for class 'devalued'.
#' In this context, a widget is a text fragment
#' that formats one or more aggregate statistics,
#' all of which must have been present as LHS of
#' some element of 'fun' in the call to devalued().
#' Furthermore, LHS must be unique across 'fac' and
#' 'num' for expected behavior.
#'
#' In the context of this call, "+" is redefined
#' to allow concatenation of text.
#' Evaluation proceeds left to right as usual.
#' @param x devalued
#' @param fac a list of formulas to generate widgets for factors
#' @param num a list of formulas to generate widgets for numerics
#' @param ... formulas with matching LHS replace defaults; x ~ NULL removes x
#' @importFrom rlang f_lhs f_rhs eval_tidy
#' @importFrom tidyr gather
#' @importFrom dplyr filter
#' @export
#' @keywords internal
#' @return class 'widgets', arranged by groups:
#' \item{_tablet_name}{observation identifier}
#' \item{_tablet_level}{factor level (or special value 'numeric' for numerics)}
#' \item{_tablet_N}{number of records}
#' \item{_tablet_n}{number of records in group}
#' \item{_tablet_stat}{the LHS of formulas in 'fac' and 'num'}
#' \item{_tablet_widget}{the RHS of formulas in 'fac' and 'num' (evaluated)}

widgets.devalued <- function(
   x,
   fac = list(
       ` ` ~ sum + ' (' + pct + '%' + ')'
    ),
   num = list(
      `Mean (SD)` ~ ave + ' (' + std + ')',
      `Median (range)` ~ med + ' (' + min + ', ' + max + ')',
      `Missing` ~ mis
   ),
   ...
){
   `+` <- function(e1, e2){
      if(is.numeric(e1) && is.numeric(e2)){
         res <- .Primitive("+")(e1, e2)
         return(res)
      }
      e1 <- as.character(e1)
      e2 <- as.character(e2)
      res <- paste0(e1, e2)
      return(res)
   }
   for(i in c('_tablet_stat', '_tablet_widget'))if(i %in% names(x))stop('names x cannot include ',i)
   
   ext <- .named_formulas(list(...))
   fac <- .named_formulas(fac)
   num <- .named_formulas(num)
   nul <- .null_formulas(ext)
   fac <- fac[!(names(fac) %in% nul)]
   num <- num[!(names(num) %in% nul)]
   ext <- ext[!(names(ext) %in% nul)]

   # form <- sapply(extra, function(f)inherits(f, 'formula'))
   # if(length(form))extra <- extra[form]
   # LHS <- sapply(extra, rlang::f_lhs)
   # LHS <- sapply(LHS, rlang::as_string)
   keep <- unlist(sapply(c(fac, num), rlang::f_lhs))
   for(lhs in names(fac)){ # evaluate
      this <- NULL
      if(lhs %in% names(ext)){
        this <- ext[[lhs]]
      } else {
        this <- fac[[lhs]]
      }
      if(is.null(this))stop('formula cannot be NULL')
      # lhs <- rlang::as_string(rlang::f_lhs(this))
      # if(lhs %in% LHS)this <- extra[[match(lhs, LHS)]]
      rhs <- rlang::f_rhs(this)
      x <- mutate(x, !!lhs := ifelse(`_tablet_level` == 'numeric', NA, rlang::eval_tidy(rhs)))
   }
   for(lhs in names(num)){ # evaluate
      this <- NULL
      if(lhs %in% names(ext)){
        this <- ext[[lhs]]
      } else {
        this <- num[[lhs]]
      }
      if(is.null(this))stop('formula cannot be NULL')
      # lhs <- rlang::as_string(rlang::f_lhs(this))
      # if(lhs %in% LHS)this <- extra[[match(lhs, LHS)]]
      rhs <- rlang::f_rhs(this)
      x <- mutate(x, !!lhs := ifelse(`_tablet_level` != 'numeric', NA, rlang::eval_tidy(rhs)))
   }
   x <- gather(x, key = '_tablet_stat', value = '_tablet_widget', !!!keep, factor_key = TRUE)
   x <- filter(x, !is.na(`_tablet_widget`))
   x <- arrange(x, .by_group = TRUE)
   x <- select(x,!!!groups(x), `_tablet_N`:`_tablet_level`, `_tablet_stat`, `_tablet_widget`)
   x <- mutate(x, `_tablet_widget` = as.character(`_tablet_widget`)) # in case logi
   class(x) <- union('widgets', class(x))
   x
}

#' Calculate Without Groups
#'
#' Calculates without groups.  Generic, with method \code{\link{groupless.data.frame}}.
#' It is meaningless to use this toolchain function on classes
#' 'observations', 'devalued', or 'widgets'.
#' @param x object
#' @param ... passed
#' @export
#' @return see methods
#' @keywords internal
groupless <- function(x, ...)UseMethod('groupless')

#' Calculate Widgets Without Groups
#'
#' Calculates widgets without groups.
#' Removes grouping variables and groups, then executes:
#' data.frame -> observations -> devalued -> widgets.
#' @param x data.frame
#' @param ... passed to observations(), devalued(), widgets()
#' @importFrom dplyr groups ungroup select
#' @export
#' @keywords internal
#' @return class 'groupless', with output like \code{\link{widgets.devalued}}
groupless.data.frame <- function(x,...){
   # remove grouping variables and remove groups
   # then execute
   grp <- groups(x)
   x <- ungroup(x)
   #x <- select(x, setdiff(names(x), grp))
   x <- select(x, setdiff(names(x), sapply(grp, rlang::as_string))) # very important when grp name contains spaces!
   x <- observations(x, ...)
   x <- devalued(x, ...)
   x <- widgets(x, ...)
   if(any(x$`_tablet_n` != x$`_tablet_N`))warning('unexpected difference between N and n')
   x$`_tablet_n` <- NULL
   class(x) <- union('groupless', class(x))
   x
}

#' Calculate With Groups
#'
#' Calculates with groups.  Generic, with method \code{\link{groupfull.data.frame}}.
#' It is meaningless to use this toolchain function on classes
#' 'observations', 'devalued', or 'widgets'.
#' @param x object
#' @param ... passed
#' @export
#' @return see methods
#' @keywords internal
groupfull <- function(x, ...)UseMethod('groupfull')


#' Calculate Widgets With Groups
#'
#' Calculates widgets with groups.
#' Executes: data.frame -> observations -> devalued -> widgets.
#' @param x data.frame
#' @param ... passed to observations(), devalued(), widgets()
#' @export
#' @keywords internal
#' @return class 'groupfull', with output like \code{\link{widgets.devalued}}
groupfull.data.frame <- function(x,...){
   # execute data.frame -> observations -> devalued -> widgets
   x <- observations(x, ...)
   x <- devalued(x, ...)
   x <- widgets(x, ...)
   class(x) <- union('groupfull', class(x))
   x
}

#' Calculate With and Without Groups
#'
#' Calculates with and without groups.  Generic, with method \code{\link{groupwise.data.frame}}.
#' It is meaningless to use this toolchain function on classes
#' 'observations', 'devalued', 'widgets', 'groupless', or 'groupfull'.
#' @param x object
#' @param ... passed
#' @importFrom dplyr left_join
#' @export
#' @return see methods
#' @keywords internal
#' @examples
#' example(classifiers)
#' groupwise(x)
groupwise <- function(x, ...)UseMethod('groupwise')

#' Calculate Widgets With and Without Groups
#'
#' Calculates widgets with and without groups.
#' Supplies 'groupfull' and 'groupless' (prefixed) columns instead of 'widgets'.
# Column attributes 'label' and 'title' (highest priority) are substituted for column name, if present.
#' @param x data.frame
#' @param fun passed to groupfull() and groupless()
#' @param fac passed to groupfull() and groupless()
#' @param num passed to groupfull() and groupless()
#' @param ... passed to groupfull() and groupless()
#' @export
#' @keywords internal
#' @return class 'groupwise', arranged by groups:
#' \item{_tablet_name}{observation identifier}
#' \item{_tablet_level}{factor level or LHS of formulas in 'num'}
#' \item{_tablet_N}{number of records}
#' \item{_tablet_n}{number of records in group}
# \item{_tablet_stat}{the LHS of formulas in 'fac' and 'num'}
#' \item{_tablet_groupfull}{the LHS of formulas in 'fac' and 'num'}
#' \item{_tablet_groupless}{the LHS of formulas in 'fac' and 'num'}

groupwise.data.frame <- function(
   x,
   fun = list(
      mis ~ sum(is.na(x)),
      sum ~ sum(x, na.rm = TRUE),
      pct ~ signif(digits = 3,    sum / n * 100    ),
      ave ~ signif(digits = 3,   mean(x, na.rm = TRUE)),
      std ~ signif(digits = 3,     sd(x, na.rm = TRUE)),
      med ~ signif(digits = 3, median(x, na.rm = TRUE)),
      min ~ signif(digits = 3,    min(x, na.rm = TRUE)),
      max ~ signif(digits = 3,    max(x, na.rm = TRUE))
   ),
   fac = list(
      ` ` ~ sum + ' (' + pct + '%' + ')'
   ),
   num = list(
      `Mean (SD)` ~ ave + ' (' + std + ')',
      `Median (range)` ~ med + ' (' + min + ', ' + max + ')',
      `Missing` ~ mis
   ),
   ...
){
   # # promote labels and titles where present
   # for(i in seq_len(ncol(x))){
   #    lab <- attr(x[[i]],'label')
   #    ttl <- attr(x[[i]], 'title')
   #    if(length(lab)) names(x)[[i]] <- lab
   #    if(length(ttl)) names(x)[[i]] <- ttl
   # }

   # execute data.frame -> observations -> devalued -> widgets
   y <- groupfull(x, fun = fun, fac = fac, num = num, ...)
   y <- rename(y, `_tablet_groupfull` = `_tablet_widget`)
   z <- groupless(x, fun = fun, fac = fac, num = num, ...)
   z <- rename(z, `_tablet_groupless` = `_tablet_widget`)
   n <- nrow(y)
   y <- suppressMessages(left_join(y, z))
   if(nrow(y) != n)warning('unexpected change in row count')
   discard <- c('groupless','groupfull','widgets','observations','devalued')
   myclass <- setdiff(class(x), discard)
   myclass <- union('groupwise', myclass)
   class(y) <- myclass
   y
}

#' Generate a Tablet
#'
#' Generates a tablet.  Generic. See \code{\link{tablet.data.frame}} and \code{\link{as_kable.tablet}}.
#' @param x object
#' @param ... passed arguments
#' @export
#' @return see methods
#' @keywords internal
#' @examples
#' example(tablet.data.frame)
tablet <- function(x, ...)UseMethod('tablet')

#' Generate a Group-wise Tablette
#'
#' Generates a group-wise tablette. Calculates statistics
#' for all factors and numerics, with and without grouping
#' variables.  Column names represent finest level of
#' grouping, distinguished by attribute 'nest' (the values of
#' higher groups). Column names include 'all' column
#' for same statistics without groups. Result columns
#' have corresponding attribute 'n'. 'lab' supplies a label attribute
#' for each column where the RHS succeeds: by default appending 'n' to result column names.
# Column attributes 'label' and 'title' (highest priority) are substituted for column name, if present.
#' @param x groupwise
#' @param ... formulas of the form lab ~ NULL will remove elements of 'lab'
#' @param all a column name for ungrouped statistics; can have length zero to suppress ungrouped column
#' @param lab a list of formulas to generate column labels; \\n is translated as <br> in html context; each formula in succession is applied to each column
#' @importFrom dplyr groups ungroup select
#' @importFrom tidyr spread
#' @importFrom rlang as_string
#' @export
#' @keywords internal
#' @return 'tablette', with columns for each combination of groups, and:
#' \item{_tablet_name}{observation identifier}
#' \item{_tablet_level}{factor level (or special value 'numeric' for numerics)}
# \item{_tablet_sort}{sorting column}
#' \item{_tablet_stat}{the LHS of formulas in 'fac' and 'num'}
#' \item{All (or value of 'all' argument)}{ungrouped results}
tablette.groupwise <- function(
   x,
   ...,
   all = 'All',
   lab = list(
      lab ~ name + '\n(N = ' + n + ')'
   )
){

  # convert certain columns to attributes
  myclass <- class(x)
  stopifnot(
     (length(all) == 0 ) |   # can be NULL or character(0)
     (is.character(all) & length(all) == 1) # otherwise atomic char
  )
  N <- unique(x$`_tablet_N`)
  if(!length(N))N <- 0L
  x$`_tablet_N` <- NULL
  grp <- groups(x)
  grp <- lapply(grp, rlang::as_string)
  strata <- setdiff(grp, c('_tablet_name','_tablet_level'))
  strata <- rev(strata)
  base <- character(0)
  if(length(strata)){
     base <- strata[[1]]
  }
  if(length(base)){
     # x[[base]] <- factor(x[[base]])
     # levs <- levels(x[[base]])
     # quan <- x$`_tablet_n`[match(levs, x[[base]])]
     # levels(x[[base]]) <- paste(levs, quan, sep = '_tablet_nlev_')

     # nested elements of base may be repeated,
     # with differing n,
     # and therefore must not collapse to mere factor
     x[[base]] <- as.character(x[[base]])
     x[[base]] <- paste(x[[base]], x$`_tablet_n`, sep = '_tablet_nlev_')
  }
  x$`_tablet_n` <- NULL
  tot <- select(
     ungroup(x),
     `_tablet_name`,
     `_tablet_level`,
     `_tablet_stat`,
     `_tablet_groupless`
  )
  tot <- unique(tot)
  if(length(all)){
     names(tot)[names(tot)=='_tablet_groupless'] <- all
  } else {
       tot$`_tablet_groupless` <- NULL
  }

  x <- ungroup(x)

  # distribute columns
  if(length(strata)){
     x$`_tablet_class` <- do.call(
        paste,
        c(
           select(x, !!!strata),
           list(sep = '_tablet_strata_')
        )
     )
     x$`_tablet_class` <- factor(x$`_tablet_class`, levels = unique(x$`_tablet_class`))
     # strsplit is used later.
     # if last class is empty string, it will be discarded.
     # guard empty string with one more instance of separator.
     levels(x$`_tablet_class`) <- paste0(levels(x$`_tablet_class`),'_tablet_strata_')
     x <- select(
        x,
        `_tablet_class`,
        `_tablet_name`,
        `_tablet_level`,
        `_tablet_stat`,
        `_tablet_groupfull`
      )
     x <- spread(x, `_tablet_class`, `_tablet_groupfull`)
  } else {
      x <- tot
  }

  # bind ungrouped stats
  suppressMessages(x <- left_join(x,tot))
  if(length(all))attr(x[[all]], 'n') <- N
  for(i in seq_len(ncol(x))){
     cl <- names(x)[[i]]
     if(grepl('_tablet_strata_', cl)){
        splits <- strsplit(cl, '_tablet_strata_')[[1]]
        names(x)[[i]] <- splits[[1]]
        attr(x[[i]], 'nest') <- splits[-1]
     }
  }
  for(i in seq_len(ncol(x))){
     cl <- names(x)[[i]]
     if(grepl('_tablet_nlev_', cl)){
        splits <- strsplit(cl, '_tablet_nlev_')[[1]]
        names(x)[[i]] <- splits[[1]]
        attr(x[[i]], 'n') <- splits[[2]]
     }
  }

  # markup result columns
  `+` <- function(e1, e2){
     if(is.numeric(e1) && is.numeric(e2)){
        res <- .Primitive("+")(e1, e2)
        return(res)
     }
     e1 <- as.character(e1)
     e2 <- as.character(e2)
     res <- paste0(e1, e2)
     return(res)
  }
  ext <- .named_formulas(list(...))
  lab <- .named_formulas(lab)
  nul <- .null_formulas(ext)
  lab <- lab[!(names(lab) %in% nul)]
  ext <- ext[!(names(ext) %in% nul)]
  for(lhs in names(lab)){ # evaluate
    this <- NULL
    if(lhs %in% names(ext)){
      this <- ext[[lhs]]
    } else {
      this <- lab[[lhs]]
    }
    if(is.null(this))stop('formula cannot be NULL')
    # lhs <- rlang::as_string(rlang::f_lhs(this))
    #  if(lhs %in% LHS)this <- extra[[match(lhs, LHS)]]
     rhs <- rlang::f_rhs(this)
     for(j in seq_len(ncol(x))){
        nm <- names(x)[[j]]
        dat <- attributes(x[[j]])
        dat <- c(dat, list(name = nm))
        out <- try(silent = TRUE, rlang::eval_tidy(rhs, data = dat))
        if(!inherits(out, 'try-error')){
           attr(x[[j]], 'label') <- out
        }
     }
  }

  # restore levels
  # for vignette transpose example:
  # Error in .subset2(chunks, self$get_current_group()) : attempt to select less than one element in integerOneIndex
  # x <- mutate(x, `_tablet_sort` = as.numeric(`_tablet_level`))
  # x <- mutate(x, `_tablet_level` = as.character(`_tablet_level`))
  # x <- mutate(x, `_tablet_level` = sub('.*_tablet_', '', `_tablet_level`))
  #x$`_tablet_sort` <- as.numeric(x$`_tablet_level`)
  x$`_tablet_level` <- as.character(x$`_tablet_level`)
  x$`_tablet_level` <- sub('.*_tablet_', '', x$`_tablet_level`)

  constitutive <- c(
     '_tablet_name',
     '_tablet_level',
    # '_tablet_sort',
     '_tablet_stat'
  )

  x <- cbind(
    x[, constitutive, drop = FALSE],
    x[,!(names(x) %in% constitutive), drop = FALSE]
  )


  # this is the only constructor for 'tablet'
  x$`_tablet_name` <- as.character(x$`_tablet_name`)
  
  x$`_tablet_level` <- as.character(x$`_tablet_level`)
  x$`_tablet_stat` <- as.character(x$`_tablet_stat`)
  x$`_tablet_level` <- ifelse(
     x$`_tablet_level` == 'numeric',
     x$`_tablet_stat`,
     x$`_tablet_level`
  )
  x$`_tablet_stat` <- NULL
  
  
  class(x) <- union('tablette', setdiff(class(x), 'groupwise'))
  #x <- tablet(x, ...) # tablet.tablette
  x
}

#' Create Header List.
#'
#' Creates header list.  Generic, with method \code{\link{headerlist.tablette}}.
#' @param x object
#' @param ... passed
#' @export
#' @return see methods
#' @keywords internal
#' @examples
#' example(classifiers)
#' headerlist(tablette(groupwise(x)))
headerlist <- function(x,...)UseMethod('headerlist')

#' Create Header List for Tablet
#'
#' Creates header list for tablet.
#' @param x tablet
#' @param ... ignored
#' @export
#' @keywords internal
#' @return list of named integer where each element describes an additional header row with names as indicated, repeated integer times
headerlist.tablette <- function(x, ...){
   numheaders <- 0
   out <- list()
   for(i in seq_len(ncol(x))){
      h <- attr(x[[i]],'nest')
      numheaders <- max(numheaders, length(h))
   }
   for(e in seq_len(numheaders)){ # for each header level
      this <- numeric(0)
      for(col in seq_len(ncol(x))){ # for each column
         h <- attr(x[[col]], 'nest')
         if(length(h) >= e){
            head <- h[[e]] # use the right level if you have it
         } else {
            head <- ' ' # otherwise use one space
         }
         if(length(this) == 0){
            this <- 1L
            names(this) <- head
         } else{
            if(rev(names(this))[[1]] == head){
               this[[length(this)]] <- this[[length(this)]] + 1
            } else {
               this[[length(this) + 1]] <- 1
               names(this)[[length(this)]] <- head
            }
         }
      }
      out[[e]] <- this
   }
   return(out)
}


#' Create Index
#'
#' Creates index.  Generic, with method \code{\link{index.tablette}}.
#' @param x object
#' @param ... passed
#' @export
#' @return see methods
#' @keywords internal
#' @examples
#' example(classifiers)
#' index(tablette(groupwise(x)))
index <- function(x,...)UseMethod('index')

#' Create Index for Tablet
#'
#' Creates index for tablet.
#' @param x tablet
#' @param ... ignored
#' @export
#' @keywords internal
#' @return named integer giving row groupings with names as indicated, repeated integer times

index.tablette <- function(x, ...) {
   out <- numeric(0)
   for (row in seq_len(nrow(x))) {
      r <- as.character(x$`_tablet_name`)[[row]]
      if (length(out) == 0) {
         out <- 1L
         names(out) <- r
      } else {
         if (names(out)[[length(out)]] == r) {
            out[[length(out)]] <- out[[length(out)]] + 1
         } else {
            out[[length(out) + 1]] <- 1
            names(out)[[length(out)]] <- r
         }
      }
   }
   return(out)
}


#' Coerce to Kable
#'
#' Coerces to kable output class.  Generic,
#' with method \code{\link{as_kable.tablet}}.
#'
#' @param x object
#' @param ... passed arguments
#' @export
#' @return see methods
#' @keywords internal
#' @examples
#' example(classifiers)
#' as_kable(tablet(tablette(groupwise(x))))
as_kable <- function(x, ...)UseMethod('as_kable')
# https://colinfay.me/writing-r-extensions/generic-functions-and-methods.html
# https://stackoverflow.com/questions/13984470/possible-to-create-latex-multicolumns-in-xtable


#' Coerce Tablet to Kable
#'
#' Renders a tablet.  Calls \code{\link[kableExtra]{kbl}} and implements
#' special features like grouped columns. 
#' See also \code{\link{tablet.data.frame}}.  
#' As of 0.9.0, calling \code{\link{kbl}} is equivalent to 
#' calling \code{\link{as_kable}}.
#'
# Column \code{_tablet_name} must inherit 'character' and
# by default (in a latex render context) its values
# will eventually be processed by \code{escape_latex}.
# Thus, elements of \code{_tablet_name} that appear to be latex
# will be handled internally by \code{\link{escape_latex.latex}}
# (which tries not to re-escape metacharacters).
#'
#'
#' @param x \code{\link{tablet}}
#' @param ... passed to \code{\link[kableExtra]{kbl}}
#' @param booktabs passed to \code{\link[kableExtra]{kbl}}
#' @param escape passed to \code{\link[kableExtra]{kbl}}; defaults FALSE to allow header linebreaks
#' @param escape_latex a function to pre-process column names and content if 'escape' is FALSE (e.g., manual escaping, latex only); default \code{\link{escape_latex}}
#' @param escape_html  a function to pre-process column names and content if 'escape' is FALSE (e.g., manual escaping, html only)
#' @param variable a column name for the variables
#' @param col.names passed to \code{\link[kableExtra]{kbl}} after any linebreaking
#' @param linebreak whether to invoke \code{\link[kableExtra]{linebreak}} for column names
#' @param align passed to \code{\link[kableExtra]{linebreak}} for column names
#' @param double_escape passed to \code{\link[kableExtra]{linebreak}} for column names
#' @param linebreaker passed to \code{\link[kableExtra]{linebreak}} for column names in latex; for html, linebreaker is replaced with <br/>
#' @param pack_rows named list passed to \code{\link[kableExtra]{pack_rows}} for finer control of variable names
#' @param secondary passed to escape_latex
#' @importFrom kableExtra kbl pack_rows add_header_above linebreak
#' @importFrom dplyr rename group_vars
#' @export
#' @return like \code{\link[kableExtra]{kbl}}
#' @examples
#' library(boot)
#' library(dplyr)
#' library(magrittr)
#' library(haven)
#' library(yamlet)
#' library(spork)
#' melanoma %>%
#'   select(-time, -year) %>%
#'   mutate(sex = factor(sex), ulcer = factor(ulcer)) %>%
#'   group_by(status) %>%
#'   tablet %>%
#'   as_kable
#'
#' x <- system.file(package = 'tablet', 'shiny-examples/mesa/data/adsl.sas7bdat')
#' x %<>% read_sas %>% data.frame
#' decorations(x) 
# note weight in pounds
# x %<>% mutate(weight = signif(digits = 3, weight * 2.2))
#'
#' # calculate BMI by assuming all males are 1.75 m, all females 1.63 cm
#' x %<>% mutate(height = ifelse(sex == 'F', 1.63, 1.75))
#' x %<>% mutate(bmi = signif(digits = 3, weight / (height^2)))
#' x %<>% filter(saffl == 'Y')
#' x %<>% select(trt01a, age, sex, weight, bmi)
#' x %<>% redecorate('
#' trt01a: [ Treatment, [ Placebo, TRT 10 mg, TRT 20 mg ]]
#' age:    [ Age, year ]
#' sex:    [ Sex, [ Female: F, Male: M ]]
#' weight: [ Body Weight, kg ]
#' bmi:    [ Index_body mass, kg/m^2 ]
#' ')
#' x %<>% resolve
#' x %<>% group_by(trt01a)
#'
#' x %>% tablet %>% as_kable
#'
#' # supply default and unit-conditional latex titles
#' x %<>% modify(title = concatenate(as_latex(as_spork(c(.data$label)))))
#' x %<>% modify(
#' age, weight, bmi,
#'   title = concatenate(
#'     sep = '',  # default ok in pdf
#'     as_latex(
#'       as_spork(
#'         c(.data$label, ' (', .data$units, ')')
#'       )
#'     )
#'   )
#' )
#' x %>% tablet %>% as_kable


as_kable.tablet <- function(
   x,
   ...,
   booktabs = TRUE,
   escape = FALSE,
   escape_latex = tablet::escape_latex,
   escape_html = function(x, ...)x,
   variable = ' ',
   col.names = NA,
   linebreak = TRUE,
   align = 'c',
   double_escape = FALSE,
   linebreaker = '\n',
   pack_rows = list(escape = escape),
   secondary = FALSE
){
  out <- kbl.tablet(
    x = x,
    ...,
    booktabs = booktabs,
    escape = escape,
    escape_latex = escape_latex,
    escape_html = escape_html,
    variable = variable,
    col.names = col.names,
    linebreak = linebreak,
    align = align,
    double_escape = double_escape,
    linebreaker = linebreaker,
    pack_rows = pack_rows,
    secondary = secondary
  )
  out
}

#' Generate a Tablet for Data Frame
#'
#' Generates a 'tablet': a summary table of
#' formatted statistics for factors (is.factor()) and
#' numerics (is.numeric()) in x, with and without grouping
#' variables (if present, see \code{\link[dplyr]{group_by}}).
#' Column names represent finest level of
#' grouping, distinguished by attribute 'nest' (the values of
#' higher other groups, if any) along with the 'all' column
#' for ungrouped statistics. Column attribute 'n' indicates
#' relevant corresponding observation count.
#' Input should not have column names beginning with '_tablet'.
#'
#' Arguments 'fun', 'fac', 'num', and 'lab' are lists
#' of two-sided formulas that are evaluated in
#' an environment where '+' expresses concatenation
#' (for character elements).
#' The values of LHS should be unique across all four lists.
#' 'fun' is a list of aggregate statistics that have access to
#' \code{N} (number of original records),
#' \code{n} (number of group members), and
#' \code{x} (the numeric observations, or 1 for each factor value).
#' Aggregate statistics generated by 'fun' are available
#' for use in 'fac' and 'num'
#' which create visualizations thereof ('widgets'). Column-specific
#' attributes are available to elements of 'lab', including
#' the special attribute \code{name} (the current column name).
#' For 'lab' only, if the RHS succeeds, it becomes the label
#' attribute of the corresponding output column. 'lab' is used
#' here principally to support annotation of *output*
#' columns; if *input* columns have attributes 'label' or 'title'
#' (highest priority) those will have been already substituted
#' for default column names at the appropriate positions in the
#' output.
#'
#' Missingness of observations (and to a lesser extent, levels of
#' grouping variables) merits special consideration.
#' Be aware that na.rm_fac and na.rm_num take their defaults
#' from na.rm.  Furthermore, na.rm_fac may interact with
#' exclude_fac, which is passed to \code{\link{factor}} as \code{exclude}.
#' The goal is to support all possible ways of expressing or ignoring
#' missingness. That said, if aggregate functions are removing
#' NA, the values of arguments beginning with 'na.rm' or 'exclude'
#' may not matter.
#'
# Dropping what follows as of 0.7.2.  escaping to be element-specific.
# Column 1 of output is character.
# Its values are typically the names of the original columns
# that were factor or numeric but not in groups(x). If any
# of these had a label attribute or (priority) a title attribute
# with class 'latex', then column 1 is assigned the
# class 'latex' as well. It makes sense therefore to be consistent
# across input columns regarding the presence or not of a 'latex'
# label or title. By default, \code{\link{as_kable.tablet}} dispatches
# class-specific methods for \code{\link{escape_latex}}.
# 
# Similarly, row 1 of output is typically character. As of version 0.6.6,
# if any of the grouping variables inherits 'latex', then the return value
# of tablet.data.frame() has an attribute 'name_class' with value 'latex'.
#'
#' @usage
#' \method{tablet}{data.frame}(
#'  x,
#'  ...,
#'  na.rm = FALSE,
#'  all = 'All',
#'  fun = list(
#'   mis ~ sum(is.na(x)),
#'   sum ~ sum(x,  na.rm = TRUE),
#'   pct ~ signif(digits = 3,     sum / n * 100        ),
#'   ave ~ signif(digits = 3,    mean(x,  na.rm = TRUE)),
#'   std ~ signif(digits = 3,      sd(x,  na.rm = TRUE)),
#'   med ~ signif(digits = 3,  median(x,  na.rm = TRUE)),
#'   min ~ signif(digits = 3,     min(x,  na.rm = TRUE)),
#'   max ~ signif(digits = 3,     max(x,  na.rm = TRUE))
#'  ),
#'  fac = list(
#'   ` ` ~ sum + ' (' + pct + '\%' + ')'
#'  ),
#'  num = list(
#'   `Mean (SD)` ~ ave + ' (' + std + ')',
#'   `Median (range)` ~ med + ' (' + min + ', ' + max + ')',
#'   `Missing` ~ mis
#'   ),
#'  lab = list(
#'   lab ~ name + '\\n(N = ' + n + ')'
#'  ),
#'  na.rm_fac = na.rm,
#'  na.rm_num = na.rm,
#'  exclude_fac = NULL,
#'  exclude_name = NULL,
#'  all_levels = FALSE
#' )
#' @param x data.frame (possibly grouped)
#' @param ... substitute formulas for elements of fun, fac, num, lab; if RHS is NULL, element is removed
#' @param na.rm whether to remove NA in general
#' @param all a column name for ungrouped statistics; can have length zero to suppress ungrouped column
#' @param fun default aggregate functions expressed as formulas
#' @param fac a list of formulas to generate widgets for factors
#' @param num a list of formulas to generate widgets for numerics
#' @param lab a list of formulas to generate label attributes for columns (see details)
#' @param na.rm_fac whether to drop NA 'factor' observations; passed to \code{\link[tidyr]{gather}} as na.rm, interacts with exclude_fac
#' @param na.rm_num whether to drop NA numeric observations; passed to \code{\link[tidyr]{gather}} as na.rm
#' @param exclude_fac which factor levels to exclude; see \code{\link{factor}} (exclude)
#' @param exclude_name whether to drop NA values of column name (for completeness); passed to \code{\link[tidyr]{gather}}
#' @param all_levels whether to supply records for unobserved levels
#' @importFrom dplyr all_of across everything full_join anti_join
#' @importFrom magrittr %>% %<>%
#' @importFrom yamlet as_dvec
#' @export
#' @return 'tablet'
#' A tablet is a special case of data.frame with grouped rows and columns.
#' \item{*}{There is always one level of row groups.}
#' \item{*}{There can be any number of column groups, including zero.}
#' \item{*}{All columns are character (as tested by \code{is.character()}).}
#' \item{*}{The first column has empty strings that represent
#' the last non-empty value. It can be class 'latex' or 'character'.}
#' \item{*}{Leading element(s) of first column are deliberately blank (one space character)
#' and correspond to header rows.  See \code{\link{header_rows}}.}
#' \item{*}{The second column represents group-specific property 
#' names. It is populated always and only where column 1 is not.}
#' \item{*}{All other columns represent group-specific property values; 
#' elements before the first non-empty group value represent nested headers.}
#' \item{*}{Header values may be repeated.}
#' \item{*}{Header values may be empty strings, representing the last non-empty value
#' to the left, or single spaces, which are deliberately blank.}
#' \item{*}{Internally, character NA is equivalent to an empty string.}
#' 
#' @seealso \code{\link{as_kable.tablet}} \code{\link{kbl.tablet}}
#' @examples
#' library(boot)
#' library(dplyr)
#' library(magrittr)
#' melanoma %>%
#'   select(-time, -year) %>%
#'   mutate(sex = factor(sex), ulcer = factor(ulcer)) %>%
#'   group_by(status) %>%
#'   tablet %>%
#'   kbl
#'   
#' ## drop 'Missing', redefine 'range'
#' melanoma %>%
#'   select(-time, -year) %>%
#'   mutate(sex = factor(sex), ulcer = factor(ulcer)) %>%
#'   group_by(status) %>%
#'   tablet(
#'     Missing ~ NULL,
#'     `Median (range)` ~ med + ' (' + min + ' to ' + max + ')'
#'   ) %>%
#'   kbl
#'   

tablet.data.frame <- function(
   x,
   ...,
   na.rm = FALSE,
   all = 'All',
   fun = list(
      mis ~ sum(is.na(x)),
      sum ~ sum(x, na.rm = TRUE),
      pct ~ signif(digits = 3,    sum / n * 100    ),
      ave ~ signif(digits = 3,   mean(x, na.rm = TRUE)),
      std ~ signif(digits = 3,     sd(x, na.rm = TRUE)),
      med ~ signif(digits = 3, median(x, na.rm = TRUE)),
      min ~ signif(digits = 3,    min(x, na.rm = TRUE)),
      max ~ signif(digits = 3,    max(x, na.rm = TRUE))
   ),
   fac = list(
      ` ` ~ sum + ' (' + pct + '%' + ')'
   ),
   num = list(
      `Mean (SD)` ~ ave + ' (' + std + ')',
      `Median (range)` ~ med + ' (' + min + ', ' + max + ')',
      `Missing` ~ mis
   ),
   lab = list(
     lab ~ name + '\n(N = ' + n + ')'
   ),
   na.rm_fac = na.rm,
   na.rm_num = na.rm,
   exclude_fac = NULL,
   exclude_name = NULL,
   all_levels = FALSE
){
   # @ 0.6.12 fun, fac, num, lab must be list
   if(!length(fun)) fun <- list()
   if(!length(fac)) fac <- list()
   if(!length(num)) num <- list()
   if(!length(lab)) lab <- list()
  
   y <- groupwise( # groupwise.data.frame, returns tablette
      x,
      ...,
      na.rm = na.rm,
      # all = all,
      fun = fun,
      fac = fac,
      num = num,
      # lab = lab,
      na.rm_fac = na.rm_fac,
      na.rm_num = na.rm_num,
      exclude_fac = exclude_fac,
      exclude_name = exclude_name,
      all_levels = all_levels
   )
   y <- tablette(y, ..., all = all, lab = lab ) # tablette.groupwise
   # y$`_tablet_level` may have latex elements from factors @ 0.6.10
   # as of 0.7.2, this is moot.  Commenting next.
   # if(any(sapply(x, function(col)inherits(col,'latex')))){
   #   class(y$`_tablet_level`) <- union('latex', class(y$`_tablet_level`))
   # }
   y$`_tablet_name` <- as.character(y$`_tablet_name`)

   codes <- unique(y$`_tablet_name`)
   decod <- codes
   for(i in seq_along(codes)){
     lbl <- attr(x[[ codes[[i]] ]],'label')
     ttl <- attr(x[[ codes[[i]] ]], 'title')
     if(length(lbl)) decod[[i]] <- lbl
     if(length(ttl)) decod[[i]] <- ttl
   }
   y$`_tablet_name` %<>% factor(levels = codes, labels = decod) %>% as.character
   codelist <- codes
   names(codelist) <- decod
   codelist %<>% as.list
   attr(y$`_tablet_name`, 'codelist') <- codelist

   # check for prime target inheriting 'latex' and coerce _tablet_name accordingly
   fac <- unlist(sapply(x, is.factor)) # unlist necessary because 0 columns returns list() instead of logical vector
   num <- unlist(sapply(x, is.numeric))
   col <- names(x)[fac | num]
   # don't need to consider grouping vars!
   # they will not contribute to _tablet_name
   col <- setdiff(col, dplyr::group_vars(x))
 # if(length(col)){
   
  # as of 0.7.2 next seems no longer relevant
  # for(this in col){
  #     prime <- this
  #     targets <- intersect(c('title','label'), names(attributes(x[[prime]])))
  #     if(length(targets)){
  #        target <- targets[[1]]
  #        if(inherits(attr(x[[prime]], target), 'latex'))
  #           class(y$`_tablet_name`) <- union('latex', class(y$`_tablet_name`))
  #     }
  # }
   
   # y$`_tablet_level` <- as.character(y$`_tablet_level`)
   # y$`_tablet_stat` <- as.character(y$`_tablet_stat`)
   # y$`_tablet_level` <- ifelse(
   #    y$`_tablet_level` == 'numeric',
   #    y$`_tablet_stat`,
   #    y$`_tablet_level`
   # )
   # y$`_tablet_stat` <- NULL
  # x$`_tablet_sort` <- NULL
   y <- tablet(y, ...)
   # propagate name class to protect pre-formatted latex elements
   # for(i in group_vars(x)){
   #   if (inherits(x[[i]], 'latex')){
   #     attr(y, 'name_class') <- 'latex'
   #   }
   # }
   y
}


#' Splice Some Things Together
#'
#' Spices some things together. Generic, with tablet-oriented method \code{\link{splice.data.frame}}.
#'
#' @param x object
#' @param ... passed arguments
#' @export
#' @return see methods
#' @keywords internal
#' @examples
#' example(splice.data.frame)
splice <- function(x, ...)UseMethod('splice')

#' Splice A Data Frame
#'
#' Splices a data.frame.  If the data.frame has groups, tablet() is
#' called for each group in succession, only the last of which requests
#' 'all'.  The results are column-bound, and duplicate columns
#' are removed.
#' @param x data.frame
#' @param all a column name for ungrouped statistics; can have length zero to suppress ungrouped column
#' @param ... passed to \code{\link{tablet}}
#' @return tablet
#' @export
#' @keywords internal
#' @examples
#' library(boot)
#' library(dplyr)
#' library(magrittr)
#' melanoma %>%
#'   select(-time, -year) %>%
#'   mutate(sex = factor(sex), ulcer = factor(ulcer)) %>%
#'   mutate(status2 = ifelse(status == 2, 2, 4)) %>%
#'   group_by(status, status2) %>%
#'   splice
splice.data.frame <- function(x, all = 'All', ...){
   grp <- groups(x)
   if(!length(grp))return(tablet(x, ...))
   x <- ungroup(x)
   part <- list()
   for(g in seq_along(grp)){
      group <- grp[[g]]
      this <- group_by(x, !!group)
      for(i in setdiff(grp, list(group))) this[i] <- NULL
      tot <- all
      if(g < length(grp)) tot <- character(0) # 'all' only for last table
      that <- tablette(tablet(this, all = tot, ...))
      part[[group]] <- that
   }
   names(part) <- NULL
   out <- do.call(cbind, part)

   # remove dup columns
   dupcol <- rep(FALSE, ncol(out))
   if(ncol(out))
   for(i in seq_along(out)[-1]){
      for( j in 1:(i-1)){
         if(identical(out[[i]], out[[j]])){
            dupcol[[i]] <- TRUE
         }
      }
   }
   out <- out[,!dupcol, drop = FALSE]
   class(out) <- class(part[[1]])
   out <- tablet(out)
   out
}

#' Escape Special Characters for Latex
#'
#' Escapes special characters in Latex context.  
#' Operates element-wise, using \code{\link{escape_latex.default}}
#' or \code{\link{escape_latex.latex}} depending
#' on whether the element appears to be latex-formatted.
#'
#' An element is considered latex-formatted if it begins
#' with "\(" or "{}".  Note that even though
#' latex-like elements likely have already had reserved
#' characters escaped, \code{\link{escape_latex.latex}}
#' will attempt to fix certain anticipated escaping issues.
#' The function \code{\link{escape_latex.default}}
#' performs conventional escaping; the result is 
#' prepended with "{}" as a signal to downstream
#' processes that the text is considered latex-ready
#' (and needn't be escaped again).
#' 
#' @export
#' @importFrom dplyr case_when
#' @keywords internal
#' @return character
#' @param x typically inherits character
#' @param ... passed arguments
#' @family escape
#' @examples
#'   escape_latex(
#'     c(
#'       '',
#'       ' ',
#'       '\t',
#'       '{}a',
#'       NA,
#'       '\\()',
#'       '3%'
#'     )
#'   )

escape_latex <- function(x, ...){
  stopifnot(is.character(x))
  latex   <- paste0(escape_latex.latex(x, ...))  # limited escape
  default <- escape_latex.default(x, ...)  # conventional escape
  default <- paste0('{}', default)         # signal escape check
  out <- case_when(
    is.na(x) ~ x,                          # don't touch NA
    grepl("^\\s*$", x) ~ x,                # ignore white space or empty string
    grepl("^(\\\\\\(|\\{\\})", x) ~ latex, # starts with \( or {}
    TRUE ~ default
  )
  out
}


#' Escape Special Characters for Latex by Default
#'
#' Escapes text characters that have special meaning in latex.
#' Adapted with gratitude from KableExtra internals.
#' I.e. inactivates material that otherwise looks like latex.
#'
#' This function is used in \code{\link{as_kable.tablet}} in
#' preparation for a call to \code{\link[kableExtra]{kbl}}.
#' At kableExtra 1.3.4, sim_double_escape() only doubles
#' primary (leading) backslashes; the \code{secondary} argument
#' by default pre-doubles later backslashes to prevent
#' certain display errors.  Behavior may change if kableExtra changes.
#'
#' @export
#' @keywords internal
#' @return character
#' @param x caracter
#' @param secondary logical: whether secondary backslashes should be pre-doubled
#' @param ... ignored
#' @family escape
#' @examples
#' escape_latex('([#$%&_{}])')
escape_latex.default <- function(x, secondary = TRUE, ...){
   x = gsub("\\\\", "\\\\textbackslash", x)
   x = gsub("([#$%&_{}])", "\\\\\\1", x)
   x = gsub("\\\\textbackslash", "\\\\textbackslash{}", x)
   x = gsub("~", "\\\\textasciitilde{}", x)
   x = gsub("\\^", "\\\\textasciicircum{}", x)
   if(secondary){
      # hotfix for kableExtra::sim_double_escape @1.3.4
      # https://github.com/haozhu233/kableExtra/issues/622
      x = gsub('\\',  '\\\\', x, fixed = TRUE)
      x = sub( '\\\\','\\',   x, fixed = TRUE) # first will be doubled later by as_kable etc.
   }
   #x <- as_latex(x)
   # @ 0.6.5, apparently this formerly set class to latex.
   # as_latex is defined by spork.
   # @ spork 0.2.7, there is no default method.
   # experimentally, we hard code the class 'latex'
   # class(x) <- union('latex', class(x))
   x
}

#' Escape Latex for Latex-formated Text
#'
#' Returns argument typically unmodified.
#' Prevents accidental double-escaping
#' of the same text.
#'
#' This function is used in \code{\link{as_kable.tablet}} in
#' preparation for a call to \code{\link[kableExtra]{kbl}}.
#' At kableExtra 1.3.4, sim_double_escape() only doubles
#' primary (leading) backslashes; the \code{secondary} argument
#' by default pre-doubles later backslashes to prevent
#' certain display errors.  Behavior may change if kableExtra changes.

#' @export
#' @keywords internal
#' @return character
#' @param x character
#' @param secondary logical: whether secondary backslashes should be pre-doubled
#' @param primary logical: whether first backslashes should be pre-doubled
#' @param ... ignored
#' @family escape
#' @examples
#' identical(
#' escape_latex('([#$%&_{}])'),
#' escape_latex(escape_latex('([#$%&_{}])'))
#' )
#'
escape_latex.latex <- function(x, secondary = TRUE, primary = FALSE, ...){
   if(secondary){
      # hotfix for kableExtra::sim_double_escape @1.3.4
      # https://github.com/haozhu233/kableExtra/issues/622
      x = gsub('\\',  '\\\\', x, fixed = TRUE)
      
      # first will be doubled later by as_kable etc.
      x = sub( '\\\\','\\',   x, fixed = TRUE) 
   }
   if(primary){
     x = sub('\\',  '\\\\', x, fixed = TRUE)
   }
  
   # x <- as_latex(x)
   # class(x) <- union('latex', class(x))
   x
}


#' Relativize a Filepath
#'
#' Relativizes a filepath.  Somewhat the opposite of \code{\link{normalizePath}}.
#'
#' x and dir are first normalized, then x is expressed relative to dir.
#' If x and dir are on different drives (i.e. C:/ D:/)
#' x is returned as an absolute path.
#'
#' @export
#' @keywords internal
#' @param x length one character: a file path
#' @param dir a reference directory
#' @param sep path separator
#' @param ... ignored arguments
#'

relativizePath <- function (x, dir = getwd(), sep = "/", ...)
{
  stopifnot(length(x) == 1)
  stopifnot(file.info(dir)$isdir)
  y <- normalizePath(x, winslash = "/")
  z <- normalizePath(dir, winslash = "/")
  if (!identical(substr(y, 1, 1), substr(z, 1, 1))) {
    return(y)
  }
  y <- strsplit(y, sep)[[1]]
  z <- strsplit(z, sep)[[1]]
  count <- 0
  while (length(y) && length(z) && y[[1]] == z[[1]]) {
    y <- y[-1]
    z <- z[-1]
  }
  z <- rep("..", length(z))
  y <- c(z, y)
  y <- do.call(file.path, as.list(y))
  y
}


#' Absolutize a Filepath
#'
#' Absolutizes a filepath.  Somewhat the opposite of \code{\link{relativizePath}}.
#'
#' x and dir are first normalized, then x is expressed relative to dir.
#' If x and dir are on different drives (i.e. C:/ D:/)
#' x is returned as an absolute path.
#'
#' @export
#' @keywords internal
#' @importFrom fs is_absolute_path
#' @param x length one character: a file path
#' @param dir a reference directory
#' @param winslash path separator on windows, passed to \code{\link{normalizePath}}
#' @param ... ignored arguments
#'

absolutizePath <- function (x, dir = getwd(), winslash = "/", ...){
  stopifnot(length(x) == 1)
  stopifnot(is.character(x))
  stopifnot(length(dir) == 1)
  stopifnot(is.character(dir))
  if(is_absolute_path(x))return(x)
  out <- normalizePath(
    winslash = '/',
    file.path(dir,x)
  )
  return(out)
}

library(testthat)
test_that('tablet package result is stable',{
  library(boot)
  library(magrittr)
  library(dplyr)
  library(yamlet)
  library(tidyr)
  library(knitr)
  library(kableExtra)
  library(tablet)
  # no groups

  # expect_equal_to_reference(
  #   file = '001.rds',
  #   data.frame() %>% tablet
  # )
  #
  # expect_equal_to_reference(
  #   file = '002.rds',
  #   data.frame(x = numeric()) %>% tablet
  # )
  #
  # expect_equal_to_reference(
  #   file = '003.rds',
  #   data.frame(x = numeric(0)) %>% tablet
  # )

  expect_equal_to_reference(
    file = '004.rds',
    data.frame(x = numeric(1)) %>% tablet %>% tablette
  )

  # expect_equal_to_reference(
  #   file = '005.rds',
  #   data.frame(x = factor()) %>% tablet
  # )

  expect_equal_to_reference(
    file = '006.rds',
    data.frame(x = factor('foo')) %>% tablet %>% tablette
  )
  # just one group
  # expect_equal_to_reference(
  #   file = '007.rds',
  #   data.frame(x = numeric()) %>% group_by(x) %>% tablet
  # )

  # expect_equal_to_reference(
  #   file = '008.rds',
  #   data.frame(x = numeric(0)) %>% group_by(x) %>% tablet
  # )

  # expect_equal_to_reference(
  #   file = '009.rds',
  #   data.frame(x = numeric(1)) %>% group_by(x) %>% tablet
  # )

  # expect_equal_to_reference(
  #   file = '010.rds',
  #   data.frame(x = factor()) %>% group_by(x) %>% tablet
  # )

  # expect_equal_to_reference(
  #   file = '011.rds',
  #   data.frame(x = factor('foo')) %>% group_by(x) %>% tablet
  # )

  # one group one value
  # expect_equal_to_reference(
  #   file = '012.rds',
  #   data.frame(x = numeric(), y = numeric()) %>% group_by(y) %>% tablet
  # )

  # expect_equal_to_reference(
  #   file = '013.rds',
  #   data.frame(x = numeric(), y = numeric()) %>% group_by(y) %>% tablet
  # )

  # expect_equal_to_reference(
  #   file = '014.rds',
  #   data.frame(x = factor(), y = numeric()) %>% group_by(y) %>% tablet
  # )

  expect_equal_to_reference(
    file = '015.rds',
    data.frame(x = numeric(1), y = numeric(1)) %>% group_by(y) %>% tablet %>% tablette
  )

  expect_equal_to_reference(
    file = '016.rds',
    data.frame(x = factor('foo'), y = numeric(1)) %>% group_by(y) %>% tablet %>% tablette
  )

  # problematic names
  expect_error(
    data.frame(`_tablet_name` = 1, check.names = F) %>% tablet
  )

  expect_error(
    data.frame(`_tablet_level` = 1, check.names = F) %>% tablet
  )

})

# test_that('tablet lists stats in the order specified by num and fac',{
#   library(boot)
#   library(dplyr)
#   library(magrittr)
#   lev <- melanoma %>%
#    select(-time, -year) %>%
#    mutate(sex = factor(sex), ulcer = factor(ulcer)) %>%
#    group_by(status) %>%
#    tablet(
#      num = list(
#        `Median (range)` ~ med + ' (' + min + ', ' + max + ')',
#        `Mean (SD)` ~ ave + ' (' + std + ')'
#      )
#    ) %>% 
#     tablette %$% 
#     `_tablet_stat` %>% levels
#   expect_identical(lev, c(" ","Median (range)", "Mean (SD)"))
# })

test_that('tablet works for decorated or undecorated solitary numerics or categoricals',{
  library(magrittr)
  library(tidyr)
  library(dplyr)
  library(yamlet)
  library(boot)
  x <- as_decorated(melanoma)
  x %<>% mutate(sex = factor(sex))
  file <- system.file(package = 'yamlet', 'extdata','quinidine.csv')
  y <- decorate(file) %>% resolve

  expect_silent(x %>% as.data.frame %>% select(sex) %>% tablet)
  expect_silent(x %>% select(sex) %>% tablet)

  expect_silent(x %>% as.data.frame %>% select(age) %>% tablet)
  expect_silent(x %>% select(age) %>% tablet)

  expect_silent(y %>% as.data.frame %>% select(Age) %>% tablet)
  expect_silent(y %>% select(Age) %>% tablet)

  (y %>% as.data.frame %>% select(Race) %>% tablet)
  (y %>% select(Race) %>% tablet)

  (y %>% as.data.frame %>% select(Race) %>% numerics)
  (y %>% select(Race) %>% numerics)


})

test_that('splice returns tablet',{
  library(boot)
  library(dplyr)
  library(magrittr)
  x <- melanoma %>%
   select(-time, -year) %>%
   mutate(sex = factor(sex), ulcer = factor(ulcer)) %>%
   mutate(status2 = ifelse(status == 2, 2, 4)) %>%
   group_by(status, status2) %>%
   splice
  expect_true(inherits(x,'tablet'))

})

# dropping at 0.7.3
# test_that('class latex on first target propagates to c1',{
#   library(dplyr)
#   library(magrittr)
#   library(haven)
#   library(yamlet)
#   library(spork)
# 
#   x <- system.file(package = 'tablet', 'shiny-examples/mesa/data/adsl.sas7bdat')
#   x %<>% read_sas %>% data.frame
#   decorations(x) # note weight in pounds
#   x %<>% mutate(weight = signif(digits = 3, weight * 2.2))
# 
#   # calculate BMI by assuming all males are 1.75 m, all females 1.63 cm
#   x %<>% mutate(height = ifelse(sex == 'F', 1.63, 1.75))
#   x %<>% mutate(bmi = signif(digits = 3, weight / (height^2)))
#   x %<>% filter(saffl == 'Y')
#   x %<>% select(trt01a, age, sex, weight, bmi)
#   x %<>% redecorate('
#   trt01a: [ Treatment, [ Placebo, TRT 10 mg, TRT 20 mg ]]
#   age:    [ Age, year ]
#   sex:    [ Sex, [ Female: F, Male: M ]]
#   weight: [ Body Weight, kg ]
#   bmi:    [ Index_body mass, kg/m^2 ]
#   ')
#   x %<>% resolve
#   x %<>% group_by(trt01a)
# 
#   x %>% tablet %>% as_kable
# 
#   # supply default and unit-conditional latex titles
#   x %<>% modify(title = concatenate(as_latex(as_spork(c(.data$label)))))
#   x %<>% modify(
#   age, weight, bmi,
#     title = concatenate(
#       sep = '',  # default ok in pdf
#       as_latex(
#         as_spork(
#           c(.data$label, ' (', .data$units, ')')
#         )
#       )
#     )
#   )
#   
#   x %>% decorations
#   y <- tablet(x)
#   z <- tablette(y)
#   expect_true(inherits(z[[1]], 'latex'))
# })

test_that('as_xtable does not sort records',{
  library(tablet)
  library(magrittr)
  library(dplyr)
  x <- data.frame(
    id = c(1,2,3,4), 
    group = c('adult', 'adult', 'ped', 'ped'),
    auc = c(50, 40, 30, 20)
  )
  
  all <- x %>% select(all = auc) %>% tablet
  adt <- x %>% filter(group == 'adult') %>% select(adults = auc) %>% tablet
  ped <- x %>% filter(group == 'ped') %>% select(pediatrics = auc) %>% tablet
  
  y <- suppressWarnings(bind_rows(all, adt, ped))
  z <- as_xtable(y)
  z <- capture.output(z)
  z <- z[!grepl('^%', z)]
  
  expect_equal_to_reference(file = '017.rds',z)
  
})
test_that('null formula removes default widget',{
  library(tablet)
  library(magrittr)
  library(dplyr)
  x <- data.frame(
    id = c(1,2,3,4), 
    group = c('adult', 'adult', 'ped', 'ped'),
    auc = c(50, 40, 30, 20)
  )
  
  all <- x %>% select(all = auc) %>% tablet(`Mean (SD)` ~ NULL, lab ~ NULL, pct ~ NULL)
  expect_true(nrow(all) == 4)
  
})

#' Convert to tablette
#' 
#' Converts to 'tablette'.  Generic,
#' with method \code{\link{tablette.tablet}}.
#' 
#' @export
#' @keywords internal
#' @family tablette
#' @param x object of dispatch
#' @param ... passed arguments

tablette <- function(x, ...)UseMethod('tablette')

#' Convert to tablette from tablet.
#' 
#' Converts to 'tablette' from 'tablet'.
#' I.e., makes compact data.frame that emulates
#' the layout expected by \code{\link{as_kable}}.
#' Preserves attributes 'n', 'label', and 'codelist' where present.
#'
#' @export
#' @keywords internal
#' @family tablet
#' @param x object of dispatch
#' @param ... passed arguments
#' @return tablette : a data.frame
#'  with columns for each combination of groups, and:
#' \item{_tablet_name}{observation identifier: character 
#' has a codelist attribute the values of which are the original column names}
#' \item{_tablet_level}{factor level, or the LHS of formulas in 'num'}
# \item{_tablet_stat}{the LHS of formulas in 'fac' and 'num'}
#' \item{All (or value of 'all' argument)}{ungrouped results}
# \item{_tablet_sort}{sorting column}

#' @examples
#' library(boot)
#' library(dplyr)
#' library(magrittr)
#' library(yamlet)
#' rm(melanoma)
#' melanoma %<>% select(-time, -year) 
#' melanoma %<>% decorate('
#' time:      [ Survival Time Since Operation, day               ]
#' status:    [ End of Study Patient Status,
#'            [ Alive: 2, Melanoma Death: 1, Unrelated Death: 3 ]]
#' sex:       [ Sex, [ Male: 1, Female: 0                       ]]
#' age:       [ Age at Time of Operation, year                   ]
#' thickness: [ Tumor Thickness, mm                              ]
#' ulcer:     [ Ulceration, [ Absent: 0, Present: 1             ]]
#' ')
#' melanoma %<>% resolve
#' melanoma %<>% group_by(status, ulcer) 
#' melanoma %<>% tablet 
#' melanoma %>% header_rows
#' melanoma %>% as_kable
#' melanoma %>% tablette
#' melanoma %>% tablette %>% tablet
#' identical(melanoma, tablet(tablette(melanoma)))
#'   
tablet.tablette <- function(x, ...){
  
  # capture supported attributes
  # latex <- inherits(x[[1]],'latex') # dropped at 0.7.3
  label <- function(x)attr(x,'label')
  count <- function(x)attr(x, 'n')
  codes <- function(x)attr(x, 'codelist')
  labels <- lapply(x, label)
  counts <- lapply(x, count)
  cdlist <- lapply(x, codes)
  # name_class <- attr(x, 'name_class')
  
  # coerce to character, preserve label, codelist, and subclass
  for(col in seq_len(ncol(x))){
   # if(!inherits(x[[col]], 'character')){
      label <- attr(x[[col]], 'label')
      codelist <- attr(x[[col]], 'codelist')
      nest <- attr(x[[col]], 'nest')
      # latex <- inherits(x[[col]], 'latex') # dropped at 0.7.3
      x[[col]] %<>% as.character
      x[[col]] %<>% as_dvec(label = label, codelist = codelist, nest = nest)
      # if(latex) class(x[[col]]) <- union('latex', class(x[[col]])) # dropped at 0.7.3
      # attr(col, 'label') <- label
      # attr(col, 'codelist') <- codelist
    #}
  }
  
  #capture nested column names
  names(x)[[1]] <- ' '
  names(x)[[2]] <- ' '
  nms <- as.list(names(x))
  for(i in seq_along(nms)){
    nest <- attr(x[[i]], 'nest')
    attr(x[[i]], 'nest') <- NULL
    for (j in rev(seq_along(nest))){
      nms[[ i ]][[ j + 1 ]] <- nest[[j]]
    }
  }
  
  #longest header
  head_len <- max(sapply(nms, length))
  
  # pad short headers
  for(i in seq_along(nms)){
    have <- nms[[i]]
    need <- head_len - length(have)
    pad <- rep(' ', need)
    nms[[i]] <- c(have, pad)
  }
  
  # bind headers
  header <- data.frame(do.call(cbind, nms))
  header <- header[rev(seq_len(nrow(header))),]
  
  # sparse headers
  for(i in seq_len(nrow(header))){
    for(j in ncol(header):2){
      if(
        header[i, j] == header[i, j - 1] &
        header[i, j] != ' ' # deliberately blank
        
      ){
        header[i, j] <- ''
      }
    }
  }

  # repeat column groups
  need <- 1 + c(1, x[[1]][-1] != x[[1]][-length(x[[1]])])
  x <- x[rep(rownames(x), times = need),]
  x$tablet_need <- NULL
  
  # sparse
  first <- c(1, x[[1]][-1] != x[[1]][-length(x[[1]])])
  for(i in seq_along(first)){
    prime <- first[[i]]
    if(!prime){
      theClass <- class(x[[1]])
      x[i, 1] <- '' # drops latex?
      class(x[[1]]) <- theClass # @0.6.10
    } else {
      for(j in 2:ncol(x)){
        theClass <- class(x[[j]])
        x[i,j] <- '' # drops latex
        class(x[[j]]) <- theClass # @0.6.10
      }
    }
  }
  
  names(header) <- paste0('c', seq_len(ncol(header)))
  stopifnot(ncol(header) == ncol(x))
  names(x) <- names(header)
  # support latex
  for(i in 1:ncol(x)){
    # class(header[[1]]) <- class(x[[1]])
    class(header[[i]]) <- class(x[[i]]) # @0.6.10
  }
  
  
  y <- bind_rows(x %>% slice(0), header, x) # also drops latex (but not dvec)
  for(i in 1:ncol(x)){
    class(y[[i]]) <- class(x[[i]]) # @0.6.10
  }
  
  rownames(y) <- NULL
  # if(latex){class(y[[1]]) <- union('latex', class(y[[1]]))} # dropped at 0.7.3
  stopifnot(length(labels) == ncol(y))
  for(col in seq_len(ncol(y))){
    attr(y[[col]], 'label') <- labels[[col]]
    attr(y[[col]], 'n')     <- counts[[col]]
    attr(y[[col]], 'codelist') <- cdlist[[col]]
    
  }
  class(y) <- setdiff(class(y),'tablette')
  y <- as_tablet(y, ...)
  # attr(y, 'name_class') <- name_class
  y
}
  
#' Identify Header Rows
#' 
#' Identifies header rows.
#' Generic, with method \code{\link{header_rows.tablet}}.
#' 
#' @export
#' @param x object of dispatch
#' @param ... passed arguments
#' @keywords internal
#' @family tablet
#' @keywords internal
header_rows <- function(x, ...)UseMethod('header_rows')

#' Identify Header Rows for tablet
#' 
#' Identifies header rows for tablet.
#' 
#' @export
#' @param x tablet
#' @param ... ignored
#' @family tablet
#' @return integer: indices for those rows representing headers
header_rows.tablet <- function(x, ...){
  nodes <- !is.na(x[[1]]) & x[[1]] != ' '
  node <- match(TRUE, nodes)
  if(is.na(node))stop('no content in column 1')
  num_head <- node - 1
  dex <- seq_len(num_head)
  dex
}



#' Convert to tablette from tablet
#' 
#' Converts to tablette from tablet.
#' Intends to be the inverse of \code{\link{tablet.tablette}}. 
#' 
#' @export
#' @keywords internal
#' @family tablette
#' @family tablet
#' @return tablet data.frame
#' @param x tablette
#' @param ... passed arguments
#' @examples
#' example(tablet.tablette)
tablette.tablet <- function(x, ...){
  stopifnot(ncol(x) >= 2)
  stopifnot(nrow(x) >= 2)
  #latex <- inherits(x[[1]], 'latex') # dropped at 0.7.3
  
  # standarize name column
  x[[1]][!is.na(x[[1]]) & x[[1]] == ''] <- NA
  
  # capture key structure
  nodes <- !is.na(x[[1]]) & x[[1]] != ' '
  headers <- header_rows(x)
  # node <- match(TRUE, nodes)
  # if(is.na(node))stop('no content in column 1')
  # num_head <- node - 1
  
  # un-sparse column groups
  for(i in 2:nrow(x)){
    if(is.na(x[i, 1])){
      x[i, 1] <- x[ i - 1, 1]
    }
  }
  
  # un-sparse headers
  if(ncol(x) > 3){
    for(i in header_rows(x)){
      for(j in 4:ncol(x)){
        if(
          is.na(x[i,j]) | x[i,j] == ''){
          x[i,j] <- x[i, j - 1]
        }
      }
    }
  }
  
  # remove node rows
  y <- x # preserve attributes
  x <- x[!nodes, , drop = FALSE]
  
  
  # isolate headers
  h <- x[headers, , drop = FALSE]
  x <- x[-headers, , drop = FALSE]
  
  # restore attributes
  for(col in seq_len(ncol(x))){
    for (at in c('class', 'codelist', 'n', 'label')){
      attr(x[[col]], at) <- attr(y[[col]], at)
    }
  }
  
  # build header nests
  h <- h[rev(rownames(h)), , drop = FALSE]
  names(x) <- h[1,]
  names(x)[1:2] <- c('_tablet_name', '_tablet_level')
  h <- h[-1, , drop = FALSE]
  if(ncol(x) >= 3){
    for(j in 3:ncol(x)){
      nest <- h[, j]
      # remove backstop padding
      backstop <- !is.na(nest) & nest == ' '
      nest <- nest[!backstop]
      if(length(nest)){
        attr(x[[j]], 'nest') <- nest
      }
    }
  }
  rownames(x) <- NULL
  # if(latex){class(x[[1]]) <- union('latex', class(x[[1]]))} # dropped at 0.7.3
  class(x) <- setdiff(class(x),'tablet')
  class(x) <- union('tablette', class(x))
  x
}


#' Coerce to tablet
#' 
#' Coerces to tablet.
#' Generic, with method \code{\link{as_tablet.data.frame}}.
#' 
#' @export
#' @keywords internal
#' @family tablet
#' @return tablet
#' @param x object of dispatch
#' @param ... passed arguments
as_tablet <- function(x, ...)UseMethod('as_tablet')

#' Coerce data.frame to tablet
#' 
#' Coerces data.frame to tablet.  Checks format and assigns the class.
#' See \code{\link{tablet.data.frame}}.
#' 
#' @export
#' @family tablet
#' @return tablet
#' @param x data.frame
#' @param ... passed arguments
as_tablet.data.frame <- function(x, ...){
  stopifnot(ncol(x) >= 2)
  stopifnot(nrow(x) >= 2)
  class(x) <- union('tablet', class(x))
  x
}

#' Coerce to Xtable
#'
#' Coerces to xtable output class.  Generic,
#' with method \code{\link{as_xtable.tablet}}.
#'
#' @param x object
#' @param ... passed arguments
#' @export
#' @return see methods
#' @keywords internal
#' @examples
#' example(classifiers)
#' as_xtable(tablet(tablette(groupwise(x))))
as_xtable <- function(x, ...)UseMethod('as_xtable')


#' Coerce Tablet to Xtable
#'
#' Renders a tablet as xtable.  Calls \code{\link[xtable]{xtable}} and implements
#' default aesthetics.
#'
#' @param x \code{\link{tablet}}
#' @param caption passed to \code{\link[xtable]{xtableList}}
#' @param label passed to \code{\link[xtable]{xtableList}}
#' @param align passed to \code{\link[xtable]{xtableList}}
#' @param digits passed to \code{\link[xtable]{xtableList}}
#' @param display passed to \code{\link[xtable]{xtableList}}
#' @param format_name function to format variable names (accepts at least x and dots)
#' @param format_stat function to format names of statistics and factor levels (accepts at least x and dots)
#' @param format_value function to format cell values (accepts at least x and dots)
#' @param ... passed to \code{\link[xtable]{xtableList}}
#' @importFrom dplyr rename
#' @export
#' @keywords internal
#' @return like \code{\link[xtable]{xtableList}}
#' @examples
#' library(boot)
#' library(dplyr)
#' library(magrittr)
#' library(xtable)
#' melanoma %>%
#'   select(-time, -year) %>%
#'   mutate(sex = factor(sex), ulcer = factor(ulcer)) %>%
#'   group_by(status) %>%
#'   tablet %>%
#'   as_xtable

as_xtable.tablet <- function(
  x,
  caption = NULL,
  label = NULL,
  align = NULL,
  digits = NULL,
  display = NULL,
  auto = FALSE,
  variable = ' ',
  format_name = function(x, ...)paste0('\\!\\!\\textbf{', x, '}'),
  format_stat = function(x, ...)x,
  format_value = function(x, ...)x,
  ...
){
  if(!requireNamespace('xtable')){
    stop('Please install and load the xtable package')
  }
  x <- tablette(x, ...)
  # x$`_tablet_sort` <- NULL
  # index <- index(x)
  x$`_tablet_name` <- match.fun(format_name)(x = x$`_tablet_name`, ...)
  split <- x$`_tablet_name`
  x$`_tablet_name` <- NULL
  split <- factor(split, levels = unique(split)) # added at 0.6.9 to preserve record order
  # x$`_tablet_stat` <- as.character(x$`_tablet_stat`)
  # x$`_tablet_level` <- ifelse(
  #   x$`_tablet_level` == 'numeric',
  #   x$`_tablet_stat`,
  #   x$`_tablet_level`
  # )
  # x$`_tablet_stat` <- NULL
  x$`_tablet_level` <- match.fun(format_stat)(x = x$`_tablet_level`, ...)
  # names(x)[names(x) == 'level'] <- ''
  headerlist <- headerlist(x)
  if(length(headerlist))warning('grouped columns not currently supported')
  for(i in seq_len(ncol(x))){
    lab <- attr(x[[i]], 'label')
    if(length(lab)){
      names(x)[[i]] <- lab
    }
  }
  stopifnot(is.character(variable), length(variable) == 1)
  names(x)[names(x) == '_tablet_level'] <- variable
  for(nm in names(x)[names(x) != variable]){
    x[[nm]] <- match.fun(format_value)(x = x[[nm]], ...)
  }
  xlist <- split(x, split)
  attr(xlist, 'subheadings') <- names(xlist)
  y <- xtable::xtableList(
    xlist,
    caption = caption,
    label = label,
    align = align,
    digits = digits,
    display = display,
    auto = auto,
    ...
  )
  y
}

#' Recap Something
#' 
#' Recaps something.
#' Generic, with method \code{\link{recap.knitr_kable}}.
#' 
#' @export
#' @keywords internal
#' @family recap
#' @param x object of dispatch
#' @param ... passed arguments
recap <- function(x, ...)UseMethod('recap')

#' Recap knitr_kable.
#' 
#' Recaps knitr_kable.
#' Specifically, it replaces the first non-tabled caption
#' with multicolumn text.  The intent is to
#' prevent repeat bookmarks when generating pdf.
#' 
#' @export
#' @family recap
#' @param x object of dispatch
#' @param cols number of columns to span; guesses \code{ncol(x)} by default
#' @param pos position of text: 'l','c' (default), or 'r'
#' @param ... ignored


recap.knitr_kable <- function(x, cols = NULL, pos = 'c', ...){
  if(is.null(cols)) cols <- length(attributes(x)$kable_meta$align_vector)
  stopifnot(length(cols) == 1)
  if(cols == 0) cols <- 1
  stopifnot(length(pos) == 1)
  stopifnot(pos %in% c('l','c','r'))
  target <- '\\caption[]{'
  replacement <- paste0('\\multicolumn{',cols, '}{', pos, '}{')
  y <- sub(target, replacement, x, fixed = TRUE)
  y
}

globalVariables(c('c1','c2','label'))
  
#' Import and Export Tablet
#'
#' Imports or exports tablet as comma-separated variable.
#' Generic, with methods that extend \code{\link[yamlet]{io_csv}}.

#' @param x object
#' @param ... passed arguments
#' @export
#' @importFrom yamlet io_csv
#' @return See methods.
#' @family io
#' @examples
#' library(yamlet)
#' library(boot)
#' library(dplyr)
#' library(magrittr)
#' library(tablet)
#' library(kableExtra)
#' melanoma %<>% redecorate('
#' time: [ Time, day ]
#' status: [ Patient Status, [ Died^1: 1, Alive^2: 2, Unrelated: 3 ]]
#' sex: [ Sex^3, [ Male: 1, Female^4: 0 ]]
#' age: [ Age, year ]
#' year: [ Year of Operation ]
#' ulcer: [ Ulceration, [ Present: 1, Absent: 0 ]]
#' thickness: [ Tumor Thickness, mm ]
#' ')

#' tbl <- melanoma %>%
#'   select(-time, -year) %>%
#'   group_by(status, ulcer) %>%
#'   enscript %>%
#'   tablet(Missing~NULL)
#' path <- tbl %>% io_tablet(tempfile(fileext = '.csv'))
#' # alternatively:
#' path <- tbl %>% io_csv(tempfile(fileext = '.csv'))
#' tab <- path %>% io_tablet
#' 
#' tbl %>% 
#' kbl(booktabs = TRUE) %>%
#' kable_styling(latex_options = 'scale_down')
#' 
#' tab %>% 
#' kbl(booktabs = TRUE) %>%
#' kable_styling(latex_options = 'scale_down')


io_tablet <- function(x, ...)UseMethod('io_tablet')

#' Import Tablet
#'
#' Imports a tablet.
#' A wrapper for \code{\link[yamlet]{io_csv.character}} that also
#' reads associated yamlet metadata, if present, and applies it
#' as attributes.
#'
#' @param x character file path
#' @param na.strings passed to \code{\link[yamlet]{io_csv.character}}
#' @param strip.white passed to \code{\link[yamlet]{io_csv.character}}
#' @param ... passed to \code{\link[yamlet]{io_csv.character}}
#' @export
#' @keywords internal
#' @family io
#' @family interface
#' @return data.frame
#' @examples
#' example(io_tablet)
io_tablet.character <- function(
    x, 
    na.strings = 'NA', 
    strip.white = FALSE, 
    ...
  ){
    y <- as_tablet(
      io_csv(
        x, 
        na.strings = na.strings,  
        strip.white = strip.white,
        ...
      )
    )
    y <- modify(y, -c(c1, c2), label = gsub('\\\\n', '\n', label))
    y
  }

#' Export Dataframe as Tablet
#'
#' Exports a tablet as comma-separated variable,
#' as well as a yamlet version of its decorations.
#' A wrapper for \code{\link[yamlet]{io_csv.data.frame}}.
#'
#' @param x tablet
#' @param file passed to \code{\link[yamlet]{io_csv}}
#' @param ...  passed to \code{\link[yamlet]{io_csv}}
#' @export
#' @keywords internal
#' @family io
#' @family interface
#' @return invisible(file)
#' @examples
#' example(io_tablet)
io_tablet.data.frame <- function(x, file = '', ...){
  y <- modify(x, -c(c1, c2), label = gsub('\n', '\\\\n', label))
  # even though we likely got here from NextMethod(),
  # y still is class 'tablet' and will invoke io_csv.tablet
  # which contains a warning probably not needed here.
  # we strip.
  class(y) <- setdiff(class(y), 'tablet')
  io_csv(y, file = file, ...)
}

#' Export Tablet
#'
#' Exports a tablet as comma-separated variable,
#' as well as a yamlet version of its decorations.
#' Forwards to next method.
#'
#' @param x tablet
#' @param file passed to \code{\link[yamlet]{io_csv}}
#' @param ...  passed to \code{\link[yamlet]{io_csv}}
#' @export
#' @keywords internal
#' @family io
#' @family interface
#' @return invisible(file)
#' @examples
#' example(io_tablet)
io_tablet.tablet <- function(x, file = '', ...)NextMethod()


#' Export Tablet as CSV
#'
#' Exports a tablet as comma-separated variable,
#' by means of \code{io_csv}.  Reminds the user
#' to prefer \code{io_tablet} for lossless recovery.
#'
#' @param x tablet
#' @param file passed to \code{\link{io_tablet}}
#' @param ...  passed to \code{\link{io_tablet}}
#' @export
#' @importFrom yamlet io_csv modify
#' @keywords internal
#' @family io
#' @family interface
#' @return invisible(file)
#' @examples
#' example(io_tablet)
#' 
io_csv.tablet <- function(x, file = '', ...){
  if(file != '') message('import with io_tablet(.../',basename(file), ')')
  NextMethod()
}

#' Drag-and-drop Descriptive Statistics
#'
#' Generate a table of descriptive statistics
#' by selecting columns from a file.
#' Currently supported formats include *.xpt, *.sas7bdat, and *.csv.
#' Launch the application using \code{mesa()} and use the interface
#' to select a data file, such as 'mtcars.xpt' under 'examples/')
#' (or select configuration file 'mtcars.conf' under 'examples/').
#' Then classify the
#' columns of interest to generate the corresponding
#' displays.
#'
#' Currently,
#'
#' * xpt files are read using the defaults for \code{\link[foreign]{read.xport}},
#'
#' * sas7bdat files are read using the defaults for \code{\link[haven]{read_sas}}, and
#'
#' * csv files are read using the defaults for \code{\link[csv]{as.csv}}.
#'
#' If a file in the same directory has a corresponding base name but a \code{.yaml}
#' extension, it is treated as metadata and an attempt is made to apply it
#' to the internal version of the data. This file will not be over-written,
#' but it WILL be constructed if missing.  You can hand-edit it to supply
#' metadata.  See \code{?yamlet} for format; see the Variables tab for an easy interface.
#'
#' This is a metadata-driven application.  Columns in the data that are *not*
#' in the metadata will be ignored, and columns in the metadata that are *not*
#' in the data will be constructed (maybe *all* of them).
#'
#' The \code{\link[datasets]{mtcars}} datasets in the 'examples' volume is from \pkg{datasets}.
#'
#' @param launch.browser passed to \code{\link[shiny]{runApp}}
#' @param display.mode passed to \code{\link[shiny]{runApp}}
#' @param ... passed to \code{\link[shiny]{runApp}}
#' @importFrom spork as_html as_latex as_spork
#' @importFrom reactable reactableOutput renderReactable reactable
#' @export
#' @return used for side effects: launches shiny application

mesa <- function(launch.browser = TRUE, display.mode = 'normal', ...) {
  dependencies <- c(
    'shiny',
    'shinyFiles',
    #'shinyWidgets',
    'fs',
    'haven',
    'tablet',
    'dplyr',
    'magrittr',
    'yamlet',
    'yaml',
    'sortable',
    'kableExtra',
    'knitr',
    'latexpdf',
    'tools',
    'csv',
    'foreign',
    'tinytex'
  )
  have <- sapply(dependencies, require, character.only = TRUE, quietly = TRUE, warn.conflicts = FALSE)
  need <- dependencies[!have]
  if(length(need)){
    message('please install dependencies: ', paste(need, collapse = ', '))
    return(NULL)
  }
  appDir <- system.file("shiny-examples", "mesa", package = "tablet")
  if (appDir == "") {
    stop("Could not find example directory. Try re-installing `tablet`.", call. = FALSE)
  }

  shiny::runApp(
    appDir,
    launch.browser = launch.browser,
    display.mode = "normal",
    ...
  )
}
# https://deanattali.com/2015/04/21/r-package-shiny-app/

#' Coerce to Gridtable
#'
#' Coerces to "gridtable", i.e. the Pandoc construct "grid table" Generic,
#' with method \code{\link{as_gridtable.character}}.
#'
#' @param x object
#' @param ... passed arguments
#' @export
#' @return see methods
#' @family gridtable
#' @keywords internal
#' @examples
#' # as_gridtable() lets you create and render small data tables on-the-fly.
#' 
#' library(magrittr)
#' library(tablet)
#' library(yamlet)
#' library(kableExtra)
#' '+--+--+--+
#' |study | subjects | doses |
#' +==+==+==+
#' | 001 | 20 | 5 mg, 10 mg |
#' +--+--+--+
#' | 002 | 40 | 5 mg, 10 mg |
#' |     |    | 15 mg, 20 mg, 30 mg |
#' +--+--+--+' %>% 
#' as_gridtable
#' 
#' # We see that as_gridtable cleans up our Pandoc grid table markdown.
#' # We copy it here and render it with kableExtra::kbl().
#' 
#' '
#' +-------+----------+---------------------+
#' | study | subjects | doses               |
#' +=======+==========+=====================+
#' | 001   | 20       | 5 mg, 10 mg         |
#' +-------+----------+---------------------+
#' | 002   | 40       | 5 mg, 10 mg         |
#' |       |          | 15 mg, 20 mg, 30 mg |
#' +-------+----------+---------------------+
#' ' %>%
#' as_gridtable %>%
#' as.data.frame(scale = 0.9) %>% # or: options(tablet.gridtable.scale = 0.9)
#' decorations
#' 
#' # We see that a gridtable when converted to data.frame
#' # has scalable column width attributes.
#' # Widths are controlled by longest line of text, including column name.
#' # kbl will try to honor widths.
#' # kbl() will call as.data.frame() implicitly if you pass a gridtable:
#' 
#' '
#' +-------+----------+---------------------+
#' | study | subjects | doses               |
#' +=======+==========+=====================+
#' | 001   | 20       | 5 mg, 10 mg         |
#' +-------+----------+---------------------+
#' | 002   | 40       | 5 mg, 10 mg         |
#' |       |          | 15 mg, 20 mg, 30 mg |
#' +-------+----------+---------------------+
#' ' %>%
#' as_gridtable %>%
#' kbl(
#'   label = "studies",
#'   booktabs = TRUE,
#'   caption = "Study Characteristics" 
#' ) %>%
#'   kable_styling(font_size = 9) %>%
#'   footnote(
#'     threeparttable = TRUE,
#'     general_title = "",
#'     general = "For more details, see study protocols."
#'   )
#'   
#' # We can start with a data.frame:
#' 
#' data.frame(
#'   study = c('001','002'), 
#'   subjects = c(20, 40)
#' ) %>%
#' as_gridtable
#' 
#' # We can integrate decorations:
#' 
#' '
#' +-------+----------+
#' | study | subjects |
#' +=======+==========+
#' | 001   | 20       |
#' +-------+----------+
#' | 002   | 40       |
#' +-------+----------+
#' ' %>%
#' as_gridtable %>%
#' as.data.frame %>%
#' decorate('
#'   study: [ Study, [ ST001: "001", ST002: "002" ]]
#'   subjects: [ N Subjects ]
#' ') %>%
#' resolve %>%
#' kbl(
#'   label = "studies",
#'   booktabs = TRUE,
#'   caption = "Study Characteristics"
#' ) %>%
#' kable_styling()

as_gridtable <- function(x, ...)UseMethod('as_gridtable')

#' Coerce to Calibrated
#'
#' Coerces to calibrated output class. Generic, with flagship method
#' \code{\link{as_calibrated.data.frame}}. The intent is to attach
#' calibrated column widths for downstream table rendering.
#'
#' @param x object
#' @param ... passed arguments
#' @export
#' @return see methods
#' @family gridtable
#' @keywords internal
as_calibrated <- function(x, ...)UseMethod("as_calibrated")

#' Render with kbl
#'
#' Generic wrapper for \code{\link[kableExtra]{kbl}}. The default method is the
#' existing \code{kableExtra::kbl} definition.
#'
#' @param x object
#' @param ... passed arguments
#' @export
#' @return see methods
#' @family gridtable
#' @seealso \code{\link[kableExtra]{kbl}}
kbl <- function(x, ...)UseMethod("kbl")

#' Default kbl Method
#'
#' Calls the existing \code{\link[kableExtra]{kbl}} implementation.
#'
#' @param x Object to render.
#' @param format,digits,row.names,col.names,align,caption,label Arguments
#'   passed to \code{\link[kableExtra]{kbl}}.
#' @param format.args,escape,table.attr,booktabs,longtable,tabular,valign
#'   Arguments passed to \code{\link[kableExtra]{kbl}}.
#' @param position,centering,vline,toprule,bottomrule,midrule,linesep
#'   Arguments passed to \code{\link[kableExtra]{kbl}}.
#' @param caption.short,table.envir Arguments passed to
#'   \code{\link[kableExtra]{kbl}}.
#' @param ... Additional arguments passed to \code{\link[kableExtra]{kbl}}.
#' @export
#' @return A kable object.
#' @family gridtable
#' @seealso \code{\link[kableExtra]{kbl}}
kbl.default <- kableExtra::kbl

#' Validate and Normalize a Simple Pandoc Grid Table
#'
#' Validate a character string containing a simple Pandoc grid table and
#' rebuild it by parsing structural `+` and `|` delimiters, discarding the
#' original spacing, and returning an aligned grid table.
#'
#' The accepted grid-table form is intentionally restricted: every boundary row
#' must contain the same number of `+` delimiters; every text row is split on
#' every `|` and must therefore contain the same number of `|` delimiters; each
#' boundary span must contain at least one `-` or `=`; the first boundary row and
#' all body boundary rows must use `-`; the second boundary row must be the only
#' `=` boundary row; and exactly one header text row must appear between the
#' first and second boundary rows; header cells must be non-empty and
#' unique after trimming. Literal `|` characters inside cell text are interpreted
#' as delimiters, not cell content.
#'
#' @param x Character scalar containing one simple Pandoc grid table.
#' @param ... Currently unused.
#'
#' @return Character scalar with class \code{gridtable} containing a canonical Pandoc grid table.
#' @family gridtable
#' @export

as_gridtable.character <- function(x, ...) {
  if (!is.character(x) || length(x) != 1L || is.na(x)) {
    stop("`x` must be a non-missing character scalar.", call. = FALSE)
  }
  lines <- strsplit(x, "\\r?\\n", perl = TRUE)[[1]]
  
  while (length(lines) > 0L && !nzchar(trimws(lines[1L]))) {
    lines <- lines[-1L]
  }
  while (length(lines) > 0L && !nzchar(trimws(lines[length(lines)]))) {
    lines <- lines[-length(lines)]
  }
  
  lines <- sub("\\s+$", "", lines)
  
  if (length(lines) < 5L) {
    stop("Grid table must contain at least five non-blank lines.", call. = FALSE)
  }
  
  split_delimited <- function(value, delimiter) {
    positions <- gregexpr(delimiter, value, fixed = TRUE)[[1]]
    
    if (positions[1L] == -1L) {
      return(value)
    }
    
    substring(value, c(1L, positions + 1L), c(positions - 1L, nchar(value)))
  }
  
  parse_boundary <- function(line) {
    line <- trimws(line)
    
    if (!startsWith(line, "+") || !endsWith(line, "+")) {
      return(NULL)
    }
    
    inner <- substring(line, 2L, nchar(line) - 1L)
    spans <- split_delimited(inner, "+")
    
    if (!length(spans) || any(!nzchar(spans))) {
      stop("Each boundary span must contain at least one '-' or '='.", call. = FALSE)
    }
    if (any(!grepl("^[-=]+$", spans))) {
      stop("Boundary spans may contain only '-' or '=' characters.", call. = FALSE)
    }
    if (any(grepl("-", spans) & grepl("=", spans))) {
      stop("A boundary span may not mix '-' and '=' characters.", call. = FALSE)
    }
    
    kind <- if (all(grepl("^-+$", spans))) {
      "dash"
    } else if (all(grepl("^=+$", spans))) {
      "equals"
    } else {
      stop("A boundary row may not mix '-' and '=' spans.", call. = FALSE)
    }
    
    list(kind = kind, ncol = length(spans))
  }
  
  parse_text <- function(line, ncol) {
    line <- trimws(line)
    
    if (!startsWith(line, "|") || !endsWith(line, "|")) {
      stop("Each text row must start and end with '|'.", call. = FALSE)
    }
    
    inner <- substring(line, 2L, nchar(line) - 1L)
    cells <- split_delimited(inner, "|")
    
    if (length(cells) != ncol) {
      stop("Each text row must contain the same number of '|' delimiters.", call. = FALSE)
    }
    
    trimws(cells)
  }
  
  boundary <- lapply(lines, parse_boundary)
  is_boundary <- vapply(boundary, Negate(is.null), logical(1L))
  boundary_index <- which(is_boundary)
  
  if (!length(boundary_index) || boundary_index[1L] != 1L ||
      boundary_index[length(boundary_index)] != length(lines)) {
    stop("Grid table must start and end with a boundary row.", call. = FALSE)
  }
  if (length(boundary_index) < 3L) {
    stop("Grid table must contain a top border, header separator, and bottom border.", call. = FALSE)
  }
  
  boundary <- boundary[is_boundary]
  ncol <- boundary[[1L]]$ncol
  
  if (any(vapply(boundary, function(item) item$ncol != ncol, logical(1L)))) {
    stop("Every boundary row must define the same number of columns.", call. = FALSE)
  }
  if (boundary[[1L]]$kind != "dash") {
    stop("The first boundary row must use '-' spans.", call. = FALSE)
  }
  if (boundary[[2L]]$kind != "equals") {
    stop("The second boundary row must be the only '=' header separator.", call. = FALSE)
  }
  if (any(vapply(boundary[-2L], function(item) item$kind != "dash", logical(1L)))) {
    stop("Only the second boundary row may use '=' spans.", call. = FALSE)
  }
  
  if ((boundary_index[2L] - boundary_index[1L]) != 2L) {
    stop("Exactly one header text row must appear before the '=' separator.", call. = FALSE)
  }
  
  text_groups <- vector("list", length(boundary_index) - 1L)
  
  for (i in seq_along(text_groups)) {
    from <- boundary_index[i] + 1L
    to <- boundary_index[i + 1L] - 1L
    
    if (from > to) {
      stop("Each pair of boundary rows must contain at least one text row.", call. = FALSE)
    }
    
    text_groups[[i]] <- do.call(
      rbind,
      lapply(lines[from:to], parse_text, ncol = ncol)
    )
  }
  
  header_group <- text_groups[[1L]]
  header_names <- header_group[1L, ]
  if (any(!nzchar(header_names))) {
    stop("Grid table column names must be defined.", call. = FALSE)
  }
  if (any(duplicated(header_names))) {
    duplicate_names <- unique(header_names[duplicated(header_names)])
    stop(
      sprintf("Grid table column names must be unique; duplicated: %s.", paste(shQuote(duplicate_names), collapse = ", ")),
      call. = FALSE
    )
  }
  
  body_groups <- text_groups[-1L]
  col_widths <- rep(1L, ncol)
  
  for (j in seq_len(ncol)) {
    cells <- c(header_group[, j], unlist(lapply(body_groups, function(group) group[, j]), use.names = FALSE))
    col_widths[j] <- max(1L, nchar(cells, type = "width"))
  }
  
  pad_cell <- function(cell, width) {
    paste0(cell, strrep(" ", max(0L, width - nchar(cell, type = "width"))))
  }
  boundary_line <- function(char) {
    paste0("+", paste(strrep(char, col_widths + 2L), collapse = "+"), "+")
  }
  text_line <- function(cells) {
    paste0(
      "|",
      paste(paste0(" ", mapply(pad_cell, cells, col_widths, USE.NAMES = FALSE), " "), collapse = "|"),
      "|"
    )
  }
  
  normalized <- c(
    boundary_line("-"),
    text_line(header_group[1L, ]),
    boundary_line("=")
  )
  
  for (group in body_groups) {
    normalized <- c(
      normalized,
      apply(group, 1L, text_line),
      boundary_line("-")
    )
  }
  
  out <- paste(normalized, collapse = "\n")
  class(out) <- 'gridtable'
  return(out)
}

#' Convert a Data Frame to gridtable
#'
#' Build a simple Pandoc grid table from a data frame. Column names become the
#' grid-table header row, each data-frame row becomes one body row, and column
#' widths are calibrated from the widest formatted value in each column.
#'
#' @param x A data frame.
#' @param ... Currently unused.
#'
#' @return Character scalar with class \code{gridtable} containing a canonical
#'   Pandoc grid table.
#' @family gridtable
#' @seealso \code{\link{as_gridtable}}, \code{\link{as_gridtable.character}}
#' @method as_gridtable data.frame
#' @export
as_gridtable.data.frame <- function(x, ...) {
  cells <- base::as.data.frame(
    lapply(x, function(column) format(column, trim = TRUE, justify = "none")),
    stringsAsFactors = FALSE,
    optional = TRUE
  )
  names(cells) <- names(x)
  
  cells[] <- lapply(cells, function(column) {
    column[is.na(column)] <- "NA"
    column
  })
  
  if (!ncol(cells)) {
    stop("Grid tables require at least one column.", call. = FALSE)
  }
  if (!nrow(cells)) {
    stop("Grid tables require at least one data row.", call. = FALSE)
  }
  if (any(!nzchar(names(cells)))) {
    stop("Grid table column names must be defined.", call. = FALSE)
  }
  if (any(duplicated(names(cells)))) {
    duplicate_names <- unique(names(cells)[duplicated(names(cells))])
    stop(
      sprintf("Grid table column names must be unique; duplicated: %s.", paste(shQuote(duplicate_names), collapse = ", ")),
      call. = FALSE
    )
  }
  
  values <- rbind(names(cells), as.matrix(cells))
  if (any(grepl("[|\r\n]", values, perl = TRUE))) {
    stop("Grid table cells may not contain '|' or line breaks.", call. = FALSE)
  }
  
  col_widths <- pmax(1L, apply(values, 2L, function(column) {
    max(nchar(column, type = "width"))
  }))
  
  pad_cell <- function(cell, width) {
    paste0(cell, strrep(" ", max(0L, width - nchar(cell, type = "width"))))
  }
  boundary_line <- function(char) {
    paste0("+", paste(strrep(char, col_widths + 2L), collapse = "+"), "+")
  }
  text_line <- function(cells) {
    paste0(
      "|",
      paste(paste0(" ", mapply(pad_cell, cells, col_widths, USE.NAMES = FALSE), " "), collapse = "|"),
      "|"
    )
  }
  
  normalized <- c(
    boundary_line("-"),
    text_line(names(cells)),
    boundary_line("=")
  )
  
  for (row_index in seq_len(nrow(cells))) {
    normalized <- c(
      normalized,
      text_line(unname(unlist(cells[row_index, ], use.names = FALSE))),
      boundary_line("-")
    )
  }
  
  out <- paste(normalized, collapse = "\n")
  class(out) <- "gridtable"
  out
}

#' Print a gridtable
#'
#' Print a gridtable as plain grid-table text.
#'
#' @param x Character scalar with class \code{gridtable}.
#' @param ... Currently unused.
#'
#' @return Invisibly returns \code{x}.
#' @family gridtable
#' @export
print.gridtable <- function(x, ...) {
  writeLines(as.character(x))
  invisible(x)
}

#' Coerce a gridtable to a Calibrated Data Frame
#'
#' Parse a validated gridtable object directly to a calibrated data frame.
#' Columns have \code{width} attributes suggested by the grid table spans.
#' Continuation lines within a grid-table cell are collapsed with single spaces.
#' Cell markup is otherwise left untouched for downstream renderers.
#'
#' @param x Character scalar with class \code{gridtable}.
#' @param row.names Optional row names for the returned data frame.
#' @param optional Currently unused; accepted for compatibility with
#'   \code{as.data.frame}.
#' @param ... Currently unused.
#' @param scale Numeric scalar multiplier applied to grid-table dash counts
#'   when deriving column-level \code{width} attributes.
#'
#' @return A data frame with class \code{c("calibrated", "data.frame")}.
#'   Each column has a \code{width} attribute derived from the final grid
#'   table dash count multiplied by \code{scale} and encoded as an em width.
#' @family gridtable
#' @seealso \code{\link{as_gridtable.character}}, \code{\link{as_calibrated.data.frame}}
#' @method as.data.frame gridtable
#' @export
as.data.frame.gridtable <- function(
    x,
    row.names = NULL,
    optional = FALSE,
    ...,
    scale = getOption('tablet.gridtable.scale', 1)
){
  if (!is.numeric(scale) || length(scale) != 1L || is.na(scale) || !is.finite(scale) || scale <= 0) {
    stop("`scale` must be a positive finite numeric scalar.", call. = FALSE)
  }
  
  split_delimited <- function(value, delimiter) {
    positions <- gregexpr(delimiter, value, fixed = TRUE)[[1]]
    
    if (positions[1L] == -1L) {
      return(value)
    }
    
    substring(value, c(1L, positions + 1L), c(positions - 1L, nchar(value)))
  }
  
  parse_boundary <- function(line) {
    line <- trimws(line)
    if (!startsWith(line, "+") || !endsWith(line, "+")) {
      return(NULL)
    }
    inner <- substring(line, 2L, nchar(line) - 1L)
    split_delimited(inner, "+")
  }
  
  parse_text <- function(line, ncol) {
    line <- trimws(line)
    if (!startsWith(line, "|") || !endsWith(line, "|")) {
      stop("Each text row must start and end with '|'.", call. = FALSE)
    }
    inner <- substring(line, 2L, nchar(line) - 1L)
    cells <- split_delimited(inner, "|")
    if (length(cells) != ncol) {
      stop("Each text row must contain the same number of '|' delimiters.", call. = FALSE)
    }
    trimws(cells)
  }
  
  collapse_cell <- function(values) {
    values <- trimws(values)
    values <- values[nzchar(values)]
    if (!length(values)) {
      return("")
    }
    trimws(gsub("[[:space:]]+", " ", paste(values, collapse = " ")))
  }
  
  table_text <- as.character(as_gridtable(as.character(x)))
  lines <- strsplit(table_text, "\\r?\\n", perl = TRUE)[[1]]
  lines <- lines[nzchar(trimws(lines))]
  
  boundary <- lapply(lines, parse_boundary)
  is_boundary <- vapply(boundary, Negate(is.null), logical(1L))
  boundary_index <- which(is_boundary)
  boundary <- boundary[is_boundary]
  ncol <- length(boundary[[1L]])
  
  column_widths <- paste0(
    format(nchar(boundary[[1L]], type = "width") * 0.44 * scale, trim = TRUE, scientific = FALSE),
    "em"
  )
  
  text_groups <- vector("list", length(boundary_index) - 1L)
  for (i in seq_along(text_groups)) {
    from <- boundary_index[i] + 1L
    to <- boundary_index[i + 1L] - 1L
    text_groups[[i]] <- do.call(
      rbind,
      lapply(lines[from:to], parse_text, ncol = ncol)
    )
  }
  
  header_names <- text_groups[[1L]][1L, ]
  body_groups <- text_groups[-1L]
  values <- do.call(
    rbind,
    lapply(body_groups, function(group) {
      vapply(seq_len(ncol), function(column_index) {
        collapse_cell(group[, column_index])
      }, character(1L))
    })
  )
  
  out <- base::as.data.frame(values, stringsAsFactors = FALSE, optional = optional)
  names(out) <- header_names
  if (!is.null(row.names)) {
    row.names(out) <- row.names
  }
  
  as_calibrated(out, column_widths)
}

#' Calibrate Data Frame Column Widths
#'
#' Assign column-level \code{width} attributes to a data frame and return
#' a calibrated data frame for downstream table rendering. Widths are
#' recycled as necessary; a warning is issued when the number of supplied
#' widths differs from the number of columns.
#'
#' @param x A data frame.
#' @param width Character or numeric vector of column widths. Values are
#'   recycled as necessary.
#' @param ... Currently unused.
#'
#' @return A data frame with class \code{calibrated} and column-level
#'   \code{width} attributes.
#' @family gridtable
#' @seealso \code{\link{as_calibrated}}, \code{\link[kableExtra]{column_spec}}
#' @export
as_calibrated.data.frame <- function(x, width, ...) {
  if (missing(width)) {
    stop("`width` must be supplied.", call. = FALSE)
  }
  if (!is.atomic(width) || is.null(width) || is.matrix(width) || is.array(width)) {
    stop("`width` must be an atomic vector.", call. = FALSE)
  }
  
  width <- as.character(width)
  n_columns <- ncol(x)
  
  if (!length(width) && n_columns > 0L) {
    stop("`width` must contain at least one value.", call. = FALSE)
  }
  if (anyNA(width) || any(!nzchar(width))) {
    stop("`width` values must be non-missing and non-empty.", call. = FALSE)
  }
  if (length(width) < n_columns) {
    warning("Fewer widths than columns; recycling `width`.", call. = FALSE)
  } else if (length(width) > n_columns) {
    warning("More widths than columns; ignoring extra `width` values.", call. = FALSE)
  }
  
  if (n_columns > 0L) {
    width <- rep(width, length.out = n_columns)
    
    for (column_index in seq_len(n_columns)) {
      attr(x[[column_index]], "width") <- width[[column_index]]
    }
  }
  
  class(x) <- union("calibrated", class(x))
  x
}

#' Render a Calibrated Data Frame with kbl
#'
#' Render a calibrated data frame with \code{kbl} and optionally apply
#' column-level \code{width} attributes with \code{\link[kableExtra]{column_spec}}.
#'
#' @param x A data frame with class \code{calibrated}.
#' @param ... Additional arguments passed to \code{kbl.default}.
#' @param use_widths Logical scalar indicating whether column-level width
#'   attributes should be applied with \code{\link[kableExtra]{column_spec}}. Defaults to TRUE.
#'
#' @return A kable object.
#' @family gridtable
#' @seealso \code{\link{kbl}}, \code{\link{kbl.default}}, \code{\link[kableExtra]{column_spec}}
#' @importFrom kableExtra column_spec
#' @export
kbl.calibrated <- function(x, ..., use_widths = TRUE) {
  if (!is.logical(use_widths) || length(use_widths) != 1L || is.na(use_widths)) {
    stop('use_widths must be TRUE or FALSE.', call. = FALSE)
  }
  
  widths <- vapply(x, function(column) {
    width <- attr(column, 'width', exact = TRUE)
    
    if (is.null(width)) {
      return(NA_character_)
    }
    if (length(width) != 1L || is.na(width) || !nzchar(as.character(width))) {
      stop('Column width attributes must be non-missing scalar values.', call. = FALSE)
    }
    
    as.character(width)
  }, character(1L))
  
  out <- kbl.default(x, ...)
  
  if (!use_widths) {
    return(out)
  }
  
  for (column_index in which(!is.na(widths))) {
    out <- kableExtra::column_spec(
      out,
      column = column_index,
      width = widths[[column_index]]
    )
  }
  
  out
}


#' Render a gridtable with kbl
#'
#' Coerce a gridtable to a calibrated data frame with \code{as.data.frame}
#' before redispatching to \code{kbl}. This lets gridtable-derived column
#' widths be consumed by \code{kbl.calibrated}.
#'
#' @param x Character scalar with class \code{gridtable}.
#' @param ... Additional arguments passed to \code{kbl}.
#'
#' @return A kable object.
#' @family gridtable
#' @seealso \code{\link{kbl}}, \code{\link{kbl.calibrated}}, \code{\link{as.data.frame.gridtable}}
#' @export
kbl.gridtable <- function(x, ...) {
  kbl(as.data.frame(x), ...)
}


#' Render tablet with kbl()
#'
#' Renders a tablet.  Calls \code{\link[kableExtra]{kbl}} and implements
#' special features like grouped columns. 
#' See also \code{\link{tablet.data.frame}}.  
#' As of 0.9.0, calling \code{\link{kbl}} is equivalent to 
#' calling \code{\link{as_kable}}.
#'
# Column \code{_tablet_name} must inherit 'character' and
# by default (in a latex render context) its values
# will eventually be processed by \code{escape_latex}.
# Thus, elements of \code{_tablet_name} that appear to be latex
# will be handled internally by \code{\link{escape_latex.latex}}
# (which tries not to re-escape metacharacters).
#'
#'
#' @param x \code{\link{tablet}}
#' @param ... passed to \code{\link[kableExtra]{kbl}}
#' @param booktabs passed to \code{\link[kableExtra]{kbl}}
#' @param escape passed to \code{\link[kableExtra]{kbl}}; defaults FALSE to allow header linebreaks
#' @param escape_latex a function to pre-process column names and content if 'escape' is FALSE (e.g., manual escaping, latex only); default \code{\link{escape_latex}}
#' @param escape_html  a function to pre-process column names and content if 'escape' is FALSE (e.g., manual escaping, html only)
#' @param variable a column name for the variables
#' @param col.names passed to \code{\link[kableExtra]{kbl}} after any linebreaking
#' @param linebreak whether to invoke \code{\link[kableExtra]{linebreak}} for column names
#' @param align passed to \code{\link[kableExtra]{linebreak}} for column names
#' @param double_escape passed to \code{\link[kableExtra]{linebreak}} for column names
#' @param linebreaker passed to \code{\link[kableExtra]{linebreak}} for column names in latex; for html, linebreaker is replaced with <br/>
#' @param pack_rows named list passed to \code{\link[kableExtra]{pack_rows}} for finer control of variable names
#' @param secondary passed to escape_latex
#' @importFrom kableExtra kbl pack_rows add_header_above linebreak
#' @importFrom dplyr rename group_vars
#' @export
#' @return like \code{\link[kableExtra]{kbl}}
#' @examples
#' library(boot)
#' library(dplyr)
#' library(magrittr)
#' library(haven)
#' library(yamlet)
#' library(spork)
#' melanoma %>%
#'   select(-time, -year) %>%
#'   mutate(sex = factor(sex), ulcer = factor(ulcer)) %>%
#'   group_by(status) %>%
#'   tablet %>%
#'   kbl
#'
#' x <- system.file(package = 'tablet', 'shiny-examples/mesa/data/adsl.sas7bdat')
#' x %<>% read_sas %>% data.frame
#' decorations(x) 
# note weight in pounds
# x %<>% mutate(weight = signif(digits = 3, weight * 2.2))
#'
#' # calculate BMI by assuming all males are 1.75 m, all females 1.63 cm
#' x %<>% mutate(height = ifelse(sex == 'F', 1.63, 1.75))
#' x %<>% mutate(bmi = signif(digits = 3, weight / (height^2)))
#' x %<>% filter(saffl == 'Y')
#' x %<>% select(trt01a, age, sex, weight, bmi)
#' x %<>% redecorate('
#' trt01a: [ Treatment, [ Placebo, TRT 10 mg, TRT 20 mg ]]
#' age:    [ Age, year ]
#' sex:    [ Sex, [ Female: F, Male: M ]]
#' weight: [ Body Weight, kg ]
#' bmi:    [ Index_body mass, kg/m^2 ]
#' ')
#' x %<>% resolve
#' x %<>% group_by(trt01a)
#'
#' x %>% tablet %>% kbl
#'
#' # supply default and unit-conditional latex titles
#' x %<>% modify(title = concatenate(as_latex(as_spork(c(.data$label)))))
#' x %<>% modify(
#' age, weight, bmi,
#'   title = concatenate(
#'     sep = '',  # default ok in pdf
#'     as_latex(
#'       as_spork(
#'         c(.data$label, ' (', .data$units, ')')
#'       )
#'     )
#'   )
#' )
#' x %>% tablet %>% kbl


kbl.tablet <- function(
    x,
    ...,
    booktabs = TRUE,
    escape = FALSE,
    escape_latex = tablet::escape_latex,
    escape_html = function(x, ...)x,
    variable = ' ',
    col.names = NA,
    linebreak = TRUE,
    align = 'c',
    double_escape = FALSE,
    linebreaker = '\n',
    pack_rows = list(escape = escape),
    secondary = FALSE
){
  x <- tablette(x, ...)
  # if(is.na(escape)){
  #    if (knitr::is_latex_output()){
  #      escape <- FALSE
  #    } else {
  #       escape <- FALSE # for <br>
  #    }
  # }
  stopifnot(is.logical(escape), length(escape) == 1)
  # x$`_tablet_sort` <- NULL
  index <- index(x)
  # draws on _tablet_name, which should be character
  nmsi <- names(index) # isolate to assign class
  #stopifnot(is.character(x$`_tablet_name`))
  # class(nmsi) <- class(x$`_tablet_name`) # class propagation irrelevant with 0.7.2
  x$`_tablet_name` <- NULL # done
  if(!escape){
    if (knitr::is_latex_output()) {
      # invokes type-specific function,
      # possibly escaping or ignoring latex metacharacters
      # revisit if kableExtra changes
      nmsi <- escape_latex(nmsi, secondary = TRUE, primary = TRUE) 
      if (linebreak){
        nmsi <- linebreak(
          nmsi,
          align = 'l',
          double_escape = TRUE,
          linebreaker = linebreaker
        )
      }
    } else {
      nmsi <- escape_html(nmsi)
    }
  }
  # nmsi now as informed as possible ... assign back
  names(index) <- nmsi
  #x$`_tablet_level` <- as.character(x$`_tablet_level`)
  # x$`_tablet_stat` <- as.character(x$`_tablet_stat`)
  # x$`_tablet_level` <- ifelse(
  #    x$`_tablet_level` == 'numeric',
  #    x$`_tablet_stat`,
  #    x$`_tablet_level`
  # )
  # x$`_tablet_stat` <- NULL
  # names(x)[names(x) == 'level'] <- ''
  headerlist <- headerlist(x)
  for(i in seq_len(ncol(x))){
    lab <- attr(x[[i]], 'label')
    if(length(lab)){
      names(x)[[i]] <- lab
    }
  }
  #x <- rename(x, !!variable := `_tablet_level`)
  stopifnot(is.character(variable), length(variable) == 1)
  names(x)[names(x) == '_tablet_level'] <- variable
  
  
  # escape is false by default to allow internal discretion
  # here we handle the escaping of column names
  if(!escape){
    if (knitr::is_latex_output()) {
      # @ 0.6.10: apparently secondary should be FALSE, now default. 
      x[] <- lapply(x, escape_latex, secondary = secondary, ...)
      these <- names(x)
      # if('latex' %in% attr(x,'name_class')){
      #   class(these) <- c('latex','character')
      # }
      these <- escape_latex(these, secondary = FALSE, ...)
      names(x) <- these
    } else {
      x[] <- lapply(x, escape_html, ...)
      names(x) <- escape_html(names(x), ...)
    }
  }
  
  # names of each element in headerlist derive from the content
  # of grouping variables, and are expected to have 
  # exactly the same escaping needs as names(x)
  if(!escape){
    for(i in seq_along(headerlist)){
      if (knitr::is_latex_output()) {
        these <- names(headerlist[[i]])
        # if('latex' %in% attr(x,'name_class')){
        #   class(these) <- c('latex','character')
        # }
        these <- escape_latex(these, secondary = FALSE, ...)
        names(headerlist[[i]]) <- these
      } else {
        names(headerlist[[i]]) <- escape_html(names(headerlist[[i]]), ...)
      }
    }
  }
  
  if(is.na(col.names))col.names <- names(x)
  if (linebreak){
    if(knitr::is_latex_output()) {
      col.names <- linebreak(
        col.names,
        align = align,
        double_escape = double_escape,
        linebreaker = linebreaker
      )
    } else {
      col.names <- gsub(linebreaker,'<br/>', col.names)
    }
  }
  y <- kableExtra::kbl(
    x,
    booktabs = booktabs,
    escape = escape,
    col.names = col.names,
    ...
  )
  for(i in seq_along(headerlist)){
    this <- headerlist[[i]]
    
    if(linebreak){
      if(knitr::is_latex_output()){
        names(this) <- linebreak(
          names(this), 
          align = align, 
          double_escape = double_escape, 
          linebreaker = linebreaker
        )
      } else {
        names(this) <- gsub(linebreaker, '<br/>', names(this))
      }
    }
    
    # kableExtra:::pdfTable_add_header_above()
    # uses str_replace(string, pattern, replacement)
    # which silently deletes orphan backslashes,
    # i.e. backslashes not followed by an integer
    # therefore, names(this) must have all backslash doubled
    # @ 0.6.7 2024-03-22
    # next action only for latex?
    if(knitr::is_latex_output()){
      names(this) <- gsub('\\','\\\\', names(this), fixed = TRUE)
    }
    y <- add_header_above(y, this, escape = escape)
  }
  
  # at 0.5.7, skip this if length(index) == 0
  # attempt to prevent error in mesa() when
  # tabulating virtual category agegr1 in isolation:
  # Error in `$<-.data.frame`(`*tmp*`, "start", value = 1) : replacement has 1 row, data has 0
  
  #if(length(index) > 0){
  y <- do.call(
    kableExtra::pack_rows,
    c(
      list(y, index = index), # @0.4.9 removing ', escape = escape
      pack_rows
    )
  )
  # }
  y
}

gridtable_example <- function() {
  paste(
    "+------+------------+----------------------+" ,
    "| Key  | 001        | 002                  |" ,
    "+======+============+======================+" ,
    "| Dose | 5 mg       | 5 mg, 10 mg          |" ,
    "+------+------------+----------------------+" ,
    sep = "\n"
  )
}

expect_gridtable_runtime <- function() {}

column_widths <- function(x) {
  unname(vapply(x, attr, character(1), which = "width", exact = TRUE))
}

content_data_frame <- function(x) {
  out <- base::as.data.frame(
    lapply(seq_along(x), function(column_index) as.character(x[[column_index]])),
    stringsAsFactors = FALSE,
    optional = TRUE
  )
  names(out) <- names(x)
  row.names(out) <- NULL
  out
}

test_that("grid tables are normalized and validated", {
  grid <- as_gridtable(gridtable_example())

  expect_true(inherits(grid, "gridtable"))
  expect_true(is.character(grid))
  expect_match(as.character(grid), "^\\+[-]+\\+[-]+\\+[-]+\\+")

  duplicate_names <- paste(
    "+---+---+" ,
    "| A | A |" ,
    "+===+===+" ,
    "| x | y |" ,
    "+---+---+" ,
    sep = "\n"
  )

  expect_error(as_gridtable(duplicate_names), "unique")
  expect_error(as_gridtable("not a grid table"), "at least five")
})

test_that("data frames convert to grid tables", {
  x <- data.frame(study = 1:2, subjects = c(30, 40))

  grid <- as_gridtable(x)
  expected <- paste(
    "+-------+----------+",
    "| study | subjects |",
    "+=======+==========+",
    "| 1     | 30       |",
    "+-------+----------+",
    "| 2     | 40       |",
    "+-------+----------+",
    sep = "\n"
  )

  expect_true(inherits(grid, "gridtable"))
  expect_identical(as.character(grid), expected)
  expect_identical(as_gridtable(as.character(grid)), grid)
})

test_that("normalized grid tables preserve content through data-frame round trip", {
  expect_gridtable_runtime()

  grid <- as_gridtable(gridtable_example())
  roundtrip <- as_gridtable(as.data.frame(grid))

  expect_true(inherits(roundtrip, "gridtable"))
  expect_identical(as.character(roundtrip), as.character(grid))
})

test_that("simple data frames preserve content through gridtable round trip", {
  expect_gridtable_runtime()

  x <- data.frame(study = 1:2, subjects = c(30, 40))
  roundtrip <- as.data.frame(as_gridtable(x))

  expect_identical(class(roundtrip), c("calibrated", "data.frame"))
  expect_identical(content_data_frame(roundtrip), content_data_frame(x))
})

test_that("gridtable coercion preserves LaTeX markup", {
  grid <- as_gridtable(paste(
    "+--+--+",
    "| Parameter | Description |",
    "+==+==+",
    "| C$_{min,ss,i}$ (ng/mL) | Minimum |",
    "+--+--+",
    "| $AUC_{0-\\tau,ss,i}$ (ng $\\cdot$ h/mL) | Area |",
    "+--+--+",
    sep = "\n"
  ))
  expected <- c(
    "C$_{min,ss,i}$ (ng/mL)",
    "$AUC_{0-\\tau,ss,i}$ (ng $\\cdot$ h/mL)"
  )

  out <- as.data.frame(grid)
  latex <- paste(kbl(grid, format = "latex", booktabs = TRUE, escape = FALSE), collapse = "\n")

  expect_identical(as.vector(out$Parameter), expected)
  expect_true(grepl(expected[[1L]], latex, fixed = TRUE))
  expect_true(grepl(expected[[2L]], latex, fixed = TRUE))
})

test_that("grid tables coerce to calibrated data frames", {
  expect_gridtable_runtime()

  out <- as.data.frame(as_gridtable(gridtable_example()))

  expect_identical(class(out), c("calibrated", "data.frame"))
  expect_equal(names(out), c("Key", "001", "002"))
  expect_identical(as.vector(out[["Key"]]), "Dose")
  expect_identical(as.vector(out[["001"]]), "5 mg")
  expect_identical(as.vector(out[["002"]]), "5 mg, 10 mg")
  expect_equal(column_widths(out), c("2.64em", "2.64em", "5.72em"))
})

test_that("gridtable width scaling follows the package option", {
  expect_gridtable_runtime()

  grid <- as_gridtable(gridtable_example())
  base_widths <- as.numeric(sub("em$", "", column_widths(as.data.frame(grid))))

  old_options <- options(tablet.gridtable.scale = 2)
  on.exit(options(old_options), add = TRUE)
  scaled_widths <- as.numeric(sub("em$", "", column_widths(as.data.frame(grid))))

  expect_equal(scaled_widths, base_widths * 2)
})

test_that("as_calibrated assigns and recycles widths", {
  x <- data.frame(a = 1, b = 2, c = 3)

  exact <- as_calibrated(x, c("1em", "2em", "3em"))
  expect_identical(class(exact), c("calibrated", "data.frame"))
  expect_equal(column_widths(exact), c("1em", "2em", "3em"))

  expect_warning(few <- as_calibrated(x, "4em"), "Fewer widths")
  expect_equal(column_widths(few), c("4em", "4em", "4em"))

  expect_warning(many <- as_calibrated(x, c("1em", "2em", "3em", "4em")), "More widths")
  expect_equal(column_widths(many), c("1em", "2em", "3em"))

  expect_error(as_calibrated(x, NA_character_), "non-missing")
  expect_error(as_calibrated(x, character()), "at least one")
})

test_that("kbl consumes calibrated widths", {
  skip_if_not_installed("kableExtra")

  calibrated <- as_calibrated(data.frame(a = 1, b = 2), c("3em", "4em"))

  with_widths <- paste(kbl(calibrated, format = "latex", booktabs = TRUE), collapse = "\n")
  without_widths <- paste(kbl(calibrated, format = "latex", booktabs = TRUE, use_widths = FALSE), collapse = "\n")

  expect_match(with_widths, "p\\{3em\\}.*p\\{4em\\}")
  expect_false(grepl("p\\{3em\\}", without_widths))
})

test_that("kbl.gridtable calibrates before rendering", {
  expect_gridtable_runtime()
  skip_if_not_installed("kableExtra")

  grid <- as_gridtable(gridtable_example())

  with_widths <- paste(kbl(grid, format = "latex", booktabs = TRUE), collapse = "\n")
  without_widths <- paste(kbl(grid, format = "latex", booktabs = TRUE, use_widths = FALSE), collapse = "\n")

  expect_true(grepl("p\\{", with_widths))
  expect_false(grepl("p\\{", without_widths))
})

test_that("sloppy grid-table input is normalized", {
  sloppy <- paste(
    "+-+--+",
    "| study | subjects |",
    "+=+==+",
    "|1| 30|",
    "+-+--+",
    "| 2 |40 |",
    "+-+--+",
    sep = "\n"
  )
  expected <- paste(
    "+-------+----------+",
    "| study | subjects |",
    "+=======+==========+",
    "| 1     | 30       |",
    "+-------+----------+",
    "| 2     | 40       |",
    "+-------+----------+",
    sep = "\n"
  )
  
  grid <- as_gridtable(sloppy)
  
  expect_true(inherits(grid, "gridtable"))
  expect_identical(as.character(grid), expected)
})

test_that("gridtable widths respond to displayed column content", {
  expect_gridtable_runtime()
  
  grid <- as_gridtable(paste(
    "+-+-+",
    "| A | B |",
    "+=+=+",
    "| short | much longer entry |",
    "+-+-+",
    sep = "\n"
  ))
  
  out <- as.data.frame(grid)
  widths <- as.numeric(sub("em$", "", column_widths(out)))
  
  expect_gt(widths[[2]], widths[[1]])
})

---
title: "An Introduction to Tablet for PDF"
author: "Tim Bergsma"
date: "`r Sys.Date()`"
urlcolor: blue
output: 
  rmarkdown::pdf_document:
    toc: true
    keep_tex: true
vignette: >
  %\VignetteIndexEntry{An Introduction to Tablet for PDF}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  chunk_output_type: console
header-includes:
  - \usepackage{booktabs}
  - \usepackage{longtable}
  - \usepackage{array}
  - \usepackage{multirow}
  - \usepackage{wrapfig}
  - \usepackage{float}
  - \usepackage{colortbl}
  - \usepackage{pdflscape}
  - \usepackage{tabu}
  - \usepackage{threeparttable}
  - \usepackage{threeparttablex}
  - \usepackage[normalem]{ulem}
  - \usepackage{makecell}
  - \usepackage{xcolor}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
#knitr::opts_chunk$set(package.startup.message = FALSE)
options(xtable.comment = FALSE)
```

# Motivation

Occasionally it is useful to generate a table of 
summary statistics for rows of a dataset, where
such rows represent sampling units and and columns
may be categorical or continuous.
The excellent R package [table1](https://CRAN.R-project.org/package=table1)
does exactly this, and was the inspiration for 
```tablet```.  ```table1``` however is optimized
for html; ```tablet``` tries to provide
a format-neutral implementation and relies
on [kableExtra](https://CRAN.R-project.org/package=kableExtra) to handle the rendering.
Support for pdf (latex) is of particular interest,
and is illustrated here.  See the companion
vignette for a proof-of-concept html implementation.

# Software
To support our examples, we load some other packages
and in particular locate the melanoma dataset from
[boot](https://CRAN.R-project.org/package=boot).
By the way, in the yaml header for the Rmd source
file, we've added the header-includes as
described on p. 4 of the
```kableExtra``` [documentation](https://haozhu233.github.io/kableExtra/awesome_table_in_pdf.pdf).


```{r software, include = FALSE}
library(tidyr)
library(dplyr)
library(magrittr)
library(kableExtra)
library(boot)
library(yamlet)
library(tablet)
# options(knitr.table.format = "latex") # not needed since kableExtra 0.9.0
```

```{r software2, eval = FALSE}
library(tidyr)
library(dplyr)
library(magrittr)
library(kableExtra)
library(boot)
library(yamlet)
library(tablet)
```


```{r, data}
x <- melanoma
x %<>% select(-time, -year)
```

# Simple Case
For starters, we'll just coerce two variables to factor
to show that they are categorical, and then pass the whole
thing to tablet(). Then we forward to as_kable() for rendering
(calls kableExtra::kbl and adds some magic).

```{r, easy}
x %>%
  mutate(
    sex = factor(sex), 
    ulcer = factor(ulcer)
  ) %>%
  tablet %>%
  as_kable
```

# With Metadata

Now we redefine the dataset, supplying metadata almost verbatim from ```?melanoma```.
This is fairly easy using package ```yamlet```. Note that we reverse
the authors' factor order of 1, 0 for ulcer and move status 'Alive' to 
first position.

```{r}
x <- melanoma

x %<>% decorate('
time:      [ Survival Time Since Operation, day ]
status:
 - End of Study Patient Status
 -
  - Alive: 2
  - Melanoma Death: 1
  - Unrelated Death: 3
sex:       [ Sex, [ Male: 1, Female: 0 ]]
age:       [ Age at Time of Operation, year ]
year:      [ Year of Operation, year ]
thickness: [ Tumor Thickness, mm ]
ulcer:     [ Ulceration, [ Absent: 0, Present: 1 ]]
')
x %<>% select(-time, -year)
x %<>% group_by(status)
x %<>% resolve
```

* group_by(status) causes statistics to be summarized
in columns by group.  

* resolve() disambiguates 
labels, units, and factor levels (actually creating 
factors where appropriate, such as for sex and ulcer).

Now we pass x to tablet() and as_kable() for a more
informative result.

```{r meta}
x %>% tablet %>% as_kable
```

Notice that:

* the order of variables down the left side is exactly their order in the dataset;
* the order of factor levels is exactly that in x;
* the order of groups across the top is exactly the levels (if any) of the grouping variable(s), and
* labels and titles (highest priority) are substituted for item names.

If you don't particularly care for some aspect of the
presentation, you can jump in between tablet() and as_kable()
to fix things up.  For example, if you don't want the "All"
column you can just say 

* ```x %>% tablet %>% select(-All) %>% as_kable```.  

If you **only** want the the "All" column,
you can just remove the group(s): 

* ```x %>% ungroup %>% select(-1) %>% tablet %>% as_kable```.

By the way, you can also pass ```all = NULL``` to suppress the 'All' column.

# As Xtable

Some support is provided for 'xtable'.  Currently, grouped columns
(see next section) are not supported.

```{r xtable, results = 'asis'}
library(xtable)
  x %>% 
    filter(!(status == 'Alive' & sex == 'Male')) %>%
    tablet %>% as_xtable(format_value = function(x,...)x) %>% 
    print(
      booktabs = TRUE, 
      include.rownames = FALSE 
    )
```


# Grouped Columns

In tablet(), most columns are the consequences of a grouping
variable.  Not surprisingly, grouped columns are just 
a consequence of nested grouping variables.  To illustrate,
we follow the [table1 vignette](https://CRAN.R-project.org/package=table1) by adding a grouping variable that
groups the two kinds of death.

```{r, grouped}
x %<>% mutate(class = status)                          # copy the current group
x %<>% modify(class, label = 'class')                  # change its label
levels(x$status) <- c('Alive','Melanoma','Unrelated')  # tweak current group
levels(x$class)  <- c(' ',    'Death',   'Death')      # cluster groups
x %<>% group_by(class, status)                         # nest groups
x %>% tablet %>% as_kable                              # render
```

# Transposed Groups

Categorical observations (in principle) and grouping variables
are all factors, and are thus transposable.
To illustrate, we drop the column group above and instead
nest sex within status ...

```{r, transposed}
x %<>% group_by(status, sex)
x %<>% select(-class)
x %>% 
  tablet %>% 
  as_kable %>% 
  kable_styling(latex_options = 'scale_down')

```

... or nest ulceration within status ...

```{r, transposed2}
x %<>% group_by(status, ulcer)
x %>% 
  tablet %>% 
  as_kable %>% 
  kable_styling(latex_options = 'scale_down')

```

... or where it makes sense, use multiple levels of nesting.

```{r, transposed3}
x %<>% group_by(status, ulcer, sex)
x %>% 
  tablet %>% 
  as_kable %>% 
  kable_styling(latex_options = 'scale_down') # %>% landscape ?

```

# Aesthetics

```tablet``` tries to give rather exhaustive control
over formatting. Much can be achieved by replacing 
elements of 'fun', 'fac', 'num', and 'lab' (see ```?tablet.data.frame```).  For finer control,
you can replace these entirely.  In this example,
we will ...

* ignore categoricals (other than groups) by replacing 'fac' with something of length zero,

* drop the 'N =' header material by substituting in 'lab', and

* switch to '(min - max)' instead of '(min, max)'.

```{r, aesthetics}
x %<>% group_by(status)
x %>% 
  tablet(
    fac = NULL,
    lab ~ name,
    `Median (range)` ~ med + ' (' + min + ' - ' + max + ')'
  ) %>% 
  as_kable

```


# Note

The default presentation includes  "N = " under the header,
but also has percent characters in the table.  Considerable
gymnastics are required to make this work!  If you change
the defaults you may want to consider the arguments to
```?as_kable.tablet```.  

* 'linebreaker' anticipates that 'lab' has inserted newlines before "N =". 
* 'linebreak' invokes the eponymous kableExtra function internally.
* 'escape' is turned off to protect resulting latex markup.
* 'escape_latex' is supplied to restore the usual escaping of percent characters.

# Conclusion

```tablet``` gives a flexible way of summarizing tables
of observations.  It reacts to numeric columns, factors, and 
grouping variables.  Display order derives from 
the order of columns and factor levels in the data.
Result columns can be grouped arbitrarily deep by
supplying extra groups. 
Column labels and titles are respected.
Rendering is largely the 
responsibility of ```kableExtra``` and can be extended.
Further customization is possible by manipulating
data after calling tablet() but before calling as_kable().
Powerful results are possible with very little code.

---
title: "An Introduction to Tablet for HTML"
author: "Tim Bergsma"
date: "`r Sys.Date()`"
urlcolor: blue
output: 
  rmarkdown::html_document:
    toc: true
    keep_md: true
vignette: >
  %\VignetteIndexEntry{An Introduction to Tablet for HTML}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  chunk_output_type: console
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
#knitr::opts_chunk$set(package.startup.message = FALSE)
```

# Motivation

Occasionally it is useful to generate a table of 
summary statistics for rows of a dataset, where
such rows represent sampling units and and columns
may be categorical or continuous.  The excellent R package [table1](https://CRAN.R-project.org/package=table1)
does exactly this, and was the inspiration for 
```tablet```.  ```table1``` however is optimized
for html; ```tablet``` tries to provide
a format-neutral implementation and relies
on [kableExtra](https://CRAN.R-project.org/package=kableExtra) to handle the rendering.
Support for pdf (latex) is of particular interest,
and is illustrated in the companion vignette. 
Here we convert that presentation to html for completeness.
If you only need html output, you may prefer the interface,
flexibility, and styling options of ```table1```.


# Software
To support our examples, we load some other packages
and in particular locate the melanoma dataset from
[boot](https://CRAN.R-project.org/package=boot).


```{r software, include = FALSE}
library(tidyr)
library(dplyr)
library(magrittr)
library(kableExtra)
library(boot)
library(yamlet)
library(tablet)
# options(knitr.table.format = "latex") # not needed since kableExtra 0.9.0
```

```{r software2, eval = FALSE}
library(tidyr)
library(dplyr)
library(magrittr)
library(kableExtra)
library(boot)
library(yamlet)
library(tablet)
```


```{r, data}
x <- melanoma
x %<>% select(-time, -year)
```

# Simple Case
For starters, we'll just coerce two variables to factor
to show that they are categorical, and then pass the whole
thing to tablet(). Then we forward to kbl() for rendering
(calls kableExtra::kbl and adds some magic).

```{r, easy}
x %>%
  mutate(
    sex = factor(sex), 
    ulcer = factor(ulcer)
  ) %>%
  tablet %>%
  kbl
```

# With Metadata

Now we redefine the dataset, supplying metadata almost verbatim from ```?melanoma```.
This is fairly easy using package ```yamlet```. Note that we reverse
the authors' factor order of 1, 0 for ulcer and move status 'Alive' to 
first position.

```{r}
x <- melanoma

x %<>% decorate('
time:      [ Survival Time Since Operation, day ]
status:
 - End of Study Patient Status
 -
  - Alive: 2
  - Melanoma Death: 1
  - Unrelated Death: 3
sex:       [ Sex, [ Male: 1, Female: 0 ]]
age:       [ Age at Time of Operation, year ]
year:      [ Year of Operation, year ]
thickness: [ Tumor Thickness, mm ]
ulcer:     [ Ulceration, [ Absent: 0, Present: 1 ]]
')
x %<>% select(-time, -year)
x %<>% group_by(status)
x %<>% resolve

```

* group_by(status) causes statistics to be summarized
in columns by group.  

* resolve() disambiguates 
labels, units, and factor levels (actually creating 
factors where appropriate, such as for sex and ulcer).

Now we pass x to tablet() and kbl() for a more
informative result.

```{r meta}
x %>% tablet %>% kbl
```

Notice that:

* the order of variables down the left side is exactly their order in the dataset;
* the order of factor levels is exactly that in x;
* the order of groups across the top is exactly the levels (if any) of the grouping variable(s), and
* labels and titles (highest priority) are substituted for item names.

If you don't particularly care for some aspect of the
presentation, you can jump in between tablet() and kbl()
to fix things up.  For example, if you don't want the "All"
column you can just say 

* ```x %>% tablet %>% select(-All) %>% kbl```.  

If you **only** want the the "All" column,
you can just remove the group(s): 

* ```x %>% ungroup %>% select(-1) %>% tablet %>% kbl```.

By the way, you can also pass ```all = NULL``` to suppress the 'All' column.

# Grouped Columns

In tablet(), most columns are the consequences of a grouping
variable.  Not surprisingly, grouped columns are just 
a consequence of nested grouping variables.  To illustrate,
we follow the [table1 vignette](https://CRAN.R-project.org/package=table1) by adding a grouping variable that
groups the two kinds of death.

```{r, grouped}
x %<>% mutate(class = status)                          # copy the current group
x %<>% modify(class, label = 'class')                  # change its label
levels(x$status) <- c('Alive','Melanoma','Unrelated')  # tweak current group
levels(x$class)  <- c(' ',    'Death',   'Death')      # cluster groups
x %<>% group_by(class, status)                         # nest groups
x %>% tablet %>% kbl                              # render
```

# Transposed Groups

Categorical observations (in principle) and grouping variables
are all factors, and are thus transposable.
To illustrate, we drop the column group above and instead
nest sex within status ...

```{r, transposed}
x %<>% group_by(status, sex)
x %<>% select(-class)
x %>% 
  tablet %>% 
  kbl

```

... or nest ulceration within status ...

```{r, transposed2}
x %<>% group_by(status, ulcer)
x %>% 
  tablet %>% 
  kbl

```

... or where it makes sense, use multiple levels of nesting.

```{r, transposed3}
x %<>% group_by(status, ulcer, sex)
x %>% 
  tablet %>% 
  kbl

```

# Aesthetics

```tablet``` tries to give rather exhaustive control
over formatting. Much can be achieved by replacing 
elements of 'fun', 'fac', 'num', and 'lab' (see ```?tablet.data.frame```).  For finer control,
you can replace these entirely.  In this example,
we will ...

* ignore categoricals (other than groups) by replacing 'fac' with something of length zero,

* drop the 'N =' header material by substituting in 'lab', and

* switch to '(min - max)' instead of '(min, max)'.

```{r, aesthetics}
x %<>% group_by(status)
x %>% 
  tablet(
    fac = NULL,
    lab ~ name,
    `Median (range)` ~ med + ' (' + min + ' - ' + max + ')'
  ) %>% 
  kbl

```


# Note

The default presentation includes  "N = " under the header,
but also has percent characters in the table.  Considerable
gymnastics are required to make this work in latex.  For 
html, we simply replace the newline character (supplied by 'lab')
with a 'br' tag internally.

# Conclusion

```tablet``` gives a flexible way of summarizing tables
of observations.  It reacts to numeric columns, factors, and 
grouping variables.  Display order derives from 
the order of columns and factor levels in the data.
Result columns can be grouped arbitrarily deep by
supplying extra groups. 
Column labels and titles are respected.
Rendering is largely the 
responsibility of ```kableExtra``` and can be extended.
Further customization is possible by manipulating
data after calling tablet() but before calling kbl().
Powerful results are possible with very little code.

---
title: "An Introduction to Tablet for PDF"
author: "Tim Bergsma"
date: "`r Sys.Date()`"
urlcolor: blue
output: 
  rmarkdown::pdf_document:
    toc: true
    keep_tex: true
vignette: >
  %\VignetteIndexEntry{An Introduction to Tablet for PDF}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  chunk_output_type: console
header-includes:
  - \usepackage{booktabs}
  - \usepackage{longtable}
  - \usepackage{array}
  - \usepackage{multirow}
  - \usepackage{wrapfig}
  - \usepackage{float}
  - \usepackage{colortbl}
  - \usepackage{pdflscape}
  - \usepackage{tabu}
  - \usepackage{threeparttable}
  - \usepackage{threeparttablex}
  - \usepackage[normalem]{ulem}
  - \usepackage{makecell}
  - \usepackage{xcolor}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
#knitr::opts_chunk$set(package.startup.message = FALSE)
options(xtable.comment = FALSE)
```

# Motivation

Occasionally it is useful to generate a table of 
summary statistics for rows of a dataset, where
such rows represent sampling units and and columns
may be categorical or continuous.
The excellent R package [table1](https://CRAN.R-project.org/package=table1)
does exactly this, and was the inspiration for 
```tablet```.  ```table1``` however is optimized
for html; ```tablet``` tries to provide
a format-neutral implementation and relies
on [kableExtra](https://CRAN.R-project.org/package=kableExtra) to handle the rendering.
Support for pdf (latex) is of particular interest,
and is illustrated here.  See the companion
vignette for a proof-of-concept html implementation.

# Software
To support our examples, we load some other packages
and in particular locate the melanoma dataset from
[boot](https://CRAN.R-project.org/package=boot).
By the way, in the yaml header for the Rmd source
file, we've added the header-includes as
described on p. 4 of the
```kableExtra``` [documentation](https://haozhu233.github.io/kableExtra/awesome_table_in_pdf.pdf).


```{r software, include = FALSE}
library(tidyr)
library(dplyr)
library(magrittr)
library(kableExtra)
library(boot)
library(yamlet)
library(tablet)
# options(knitr.table.format = "latex") # not needed since kableExtra 0.9.0
```

```{r software2, eval = FALSE}
library(tidyr)
library(dplyr)
library(magrittr)
library(kableExtra)
library(boot)
library(yamlet)
library(tablet)
```


```{r, data}
x <- melanoma
x %<>% select(-time, -year)
```

# Simple Case
For starters, we'll just coerce two variables to factor
to show that they are categorical, and then pass the whole
thing to tablet(). Then we forward to kbl() for rendering
(calls kableExtra::kbl and adds some magic).

```{r, easy}
x %>%
  mutate(
    sex = factor(sex), 
    ulcer = factor(ulcer)
  ) %>%
  tablet %>%
  kbl
```

# With Metadata

Now we redefine the dataset, supplying metadata almost verbatim from ```?melanoma```.
This is fairly easy using package ```yamlet```. Note that we reverse
the authors' factor order of 1, 0 for ulcer and move status 'Alive' to 
first position.

```{r}
x <- melanoma

x %<>% decorate('
time:      [ Survival Time Since Operation, day ]
status:
 - End of Study Patient Status
 -
  - Alive: 2
  - Melanoma Death: 1
  - Unrelated Death: 3
sex:       [ Sex, [ Male: 1, Female: 0 ]]
age:       [ Age at Time of Operation, year ]
year:      [ Year of Operation, year ]
thickness: [ Tumor Thickness, mm ]
ulcer:     [ Ulceration, [ Absent: 0, Present: 1 ]]
')
x %<>% select(-time, -year)
x %<>% group_by(status)
x %<>% resolve
```

* group_by(status) causes statistics to be summarized
in columns by group.  

* resolve() disambiguates 
labels, units, and factor levels (actually creating 
factors where appropriate, such as for sex and ulcer).

Now we pass x to tablet() and kbl() for a more
informative result.

```{r meta}
x %>% tablet %>% kbl
```

Notice that:

* the order of variables down the left side is exactly their order in the dataset;
* the order of factor levels is exactly that in x;
* the order of groups across the top is exactly the levels (if any) of the grouping variable(s), and
* labels and titles (highest priority) are substituted for item names.

If you don't particularly care for some aspect of the
presentation, you can jump in between tablet() and kbl()
to fix things up.  For example, if you don't want the "All"
column you can just say 

* ```x %>% tablet %>% select(-All) %>% kbl```.  

If you **only** want the the "All" column,
you can just remove the group(s): 

* ```x %>% ungroup %>% select(-1) %>% tablet %>% kbl```.

By the way, you can also pass ```all = NULL``` to suppress the 'All' column.

# As Xtable

Some support is provided for 'xtable'.  Currently, grouped columns
(see next section) are not supported.

```{r xtable, results = 'asis'}
library(xtable)
  x %>% 
    filter(!(status == 'Alive' & sex == 'Male')) %>%
    tablet %>% as_xtable(format_value = function(x,...)x) %>% 
    print(
      booktabs = TRUE, 
      include.rownames = FALSE 
    )
```


# Grouped Columns

In tablet(), most columns are the consequences of a grouping
variable.  Not surprisingly, grouped columns are just 
a consequence of nested grouping variables.  To illustrate,
we follow the [table1 vignette](https://CRAN.R-project.org/package=table1) by adding a grouping variable that
groups the two kinds of death.

```{r, grouped}
x %<>% mutate(class = status)                          # copy the current group
x %<>% modify(class, label = 'class')                  # change its label
levels(x$status) <- c('Alive','Melanoma','Unrelated')  # tweak current group
levels(x$class)  <- c(' ',    'Death',   'Death')      # cluster groups
x %<>% group_by(class, status)                         # nest groups
x %>% tablet %>% kbl                              # render
```

# Transposed Groups

Categorical observations (in principle) and grouping variables
are all factors, and are thus transposable.
To illustrate, we drop the column group above and instead
nest sex within status ...

```{r, transposed}
x %<>% group_by(status, sex)
x %<>% select(-class)
x %>% 
  tablet %>% 
  kbl %>% 
  kable_styling(latex_options = 'scale_down')

```

... or nest ulceration within status ...

```{r, transposed2}
x %<>% group_by(status, ulcer)
x %>% 
  tablet %>% 
  kbl %>% 
  kable_styling(latex_options = 'scale_down')

```

... or where it makes sense, use multiple levels of nesting.

```{r, transposed3}
x %<>% group_by(status, ulcer, sex)
x %>% 
  tablet %>% 
  kbl %>% 
  kable_styling(latex_options = 'scale_down') # %>% landscape ?

```

# Aesthetics

```tablet``` tries to give rather exhaustive control
over formatting. Much can be achieved by replacing 
elements of 'fun', 'fac', 'num', and 'lab' (see ```?tablet.data.frame```).  For finer control,
you can replace these entirely.  In this example,
we will ...

* ignore categoricals (other than groups) by replacing 'fac' with something of length zero,

* drop the 'N =' header material by substituting in 'lab', and

* switch to '(min - max)' instead of '(min, max)'.

```{r, aesthetics}
x %<>% group_by(status)
x %>% 
  tablet(
    fac = NULL,
    lab ~ name,
    `Median (range)` ~ med + ' (' + min + ' - ' + max + ')'
  ) %>% 
  kbl

```


# Note

The default presentation includes  "N = " under the header,
but also has percent characters in the table.  Considerable
gymnastics are required to make this work!  If you change
the defaults you may want to consider the arguments to
```?as_kable.tablet```.  

* 'linebreaker' anticipates that 'lab' has inserted newlines before "N =". 
* 'linebreak' invokes the eponymous kableExtra function internally.
* 'escape' is turned off to protect resulting latex markup.
* 'escape_latex' is supplied to restore the usual escaping of percent characters.

# Conclusion

```tablet``` gives a flexible way of summarizing tables
of observations.  It reacts to numeric columns, factors, and 
grouping variables.  Display order derives from 
the order of columns and factor levels in the data.
Result columns can be grouped arbitrarily deep by
supplying extra groups. 
Column labels and titles are respected.
Rendering is largely the 
responsibility of ```kableExtra``` and can be extended.
Further customization is possible by manipulating
data after calling tablet() but before calling kbl().
Powerful results are possible with very little code.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tablet.R
\name{devalued}
\alias{devalued}
\title{Aggregate Values}
\usage{
devalued(x, ...)
}
\arguments{
\item{x}{object}

\item{...}{passed}
}
\value{
see methods
}
\description{
Aggregated values.  Generic, with method \code{\link{devalued.observations}}.
}
\examples{
example(classifiers)
devalued(observations(x))
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tablet.R
\name{as_kable.tablet}
\alias{as_kable.tablet}
\title{Coerce Tablet to Kable}
\usage{
\method{as_kable}{tablet}(
  x,
  ...,
  booktabs = TRUE,
  escape = FALSE,
  escape_latex = tablet::escape_latex,
  escape_html = function(x, ...) x,
  variable = " ",
  col.names = NA,
  linebreak = TRUE,
  align = "c",
  double_escape = FALSE,
  linebreaker = "\\n",
  pack_rows = list(escape = escape),
  secondary = FALSE
)
}
\arguments{
\item{x}{\code{\link{tablet}}}

\item{...}{passed to \code{\link[kableExtra]{kbl}}}

\item{booktabs}{passed to \code{\link[kableExtra]{kbl}}}

\item{escape}{passed to \code{\link[kableExtra]{kbl}}; defaults FALSE to allow header linebreaks}

\item{escape_latex}{a function to pre-process column names and content if 'escape' is FALSE (e.g., manual escaping, latex only); default \code{\link{escape_latex}}}

\item{escape_html}{a function to pre-process column names and content if 'escape' is FALSE (e.g., manual escaping, html only)}

\item{variable}{a column name for the variables}

\item{col.names}{passed to \code{\link[kableExtra]{kbl}} after any linebreaking}

\item{linebreak}{whether to invoke \code{\link[kableExtra]{linebreak}} for column names}

\item{align}{passed to \code{\link[kableExtra]{linebreak}} for column names}

\item{double_escape}{passed to \code{\link[kableExtra]{linebreak}} for column names}

\item{linebreaker}{passed to \code{\link[kableExtra]{linebreak}} for column names in latex; for html, linebreaker is replaced with <br/>}

\item{pack_rows}{named list passed to \code{\link[kableExtra]{pack_rows}} for finer control of variable names}

\item{secondary}{passed to escape_latex}
}
\value{
like \code{\link[kableExtra]{kbl}}
}
\description{
Renders a tablet.  Calls \code{\link[kableExtra]{kbl}} and implements
special features like grouped columns. 
See also \code{\link{tablet.data.frame}}.  
As of 0.9.0, calling \code{\link{kbl}} is equivalent to 
calling \code{\link{as_kable}}.
}
\examples{
library(boot)
library(dplyr)
library(magrittr)
library(haven)
library(yamlet)
library(spork)
melanoma \%>\%
  select(-time, -year) \%>\%
  mutate(sex = factor(sex), ulcer = factor(ulcer)) \%>\%
  group_by(status) \%>\%
  tablet \%>\%
  as_kable

x <- system.file(package = 'tablet', 'shiny-examples/mesa/data/adsl.sas7bdat')
x \%<>\% read_sas \%>\% data.frame
decorations(x) 

# calculate BMI by assuming all males are 1.75 m, all females 1.63 cm
x \%<>\% mutate(height = ifelse(sex == 'F', 1.63, 1.75))
x \%<>\% mutate(bmi = signif(digits = 3, weight / (height^2)))
x \%<>\% filter(saffl == 'Y')
x \%<>\% select(trt01a, age, sex, weight, bmi)
x \%<>\% redecorate('
trt01a: [ Treatment, [ Placebo, TRT 10 mg, TRT 20 mg ]]
age:    [ Age, year ]
sex:    [ Sex, [ Female: F, Male: M ]]
weight: [ Body Weight, kg ]
bmi:    [ Index_body mass, kg/m^2 ]
')
x \%<>\% resolve
x \%<>\% group_by(trt01a)

x \%>\% tablet \%>\% as_kable

# supply default and unit-conditional latex titles
x \%<>\% modify(title = concatenate(as_latex(as_spork(c(.data$label)))))
x \%<>\% modify(
age, weight, bmi,
  title = concatenate(
    sep = '',  # default ok in pdf
    as_latex(
      as_spork(
        c(.data$label, ' (', .data$units, ')')
      )
    )
  )
)
x \%>\% tablet \%>\% as_kable
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tablet.R
\name{splice.data.frame}
\alias{splice.data.frame}
\title{Splice A Data Frame}
\usage{
\method{splice}{data.frame}(x, all = "All", ...)
}
\arguments{
\item{x}{data.frame}

\item{all}{a column name for ungrouped statistics; can have length zero to suppress ungrouped column}

\item{...}{passed to \code{\link{tablet}}}
}
\value{
tablet
}
\description{
Splices a data.frame.  If the data.frame has groups, tablet() is
called for each group in succession, only the last of which requests
'all'.  The results are column-bound, and duplicate columns
are removed.
}
\examples{
library(boot)
library(dplyr)
library(magrittr)
melanoma \%>\%
  select(-time, -year) \%>\%
  mutate(sex = factor(sex), ulcer = factor(ulcer)) \%>\%
  mutate(status2 = ifelse(status == 2, 2, 4)) \%>\%
  group_by(status, status2) \%>\%
  splice
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tablet.R
\name{observations}
\alias{observations}
\title{Identify Observations}
\usage{
observations(x, ...)
}
\arguments{
\item{x}{object}

\item{...}{passed}
}
\value{
see methods
}
\description{
Identifies observations.  Generic, with method \code{\link{observations.data.frame}}.
}
\examples{
example(classifiers)
observations(x)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tablet.R
\name{absolutizePath}
\alias{absolutizePath}
\title{Absolutize a Filepath}
\usage{
absolutizePath(x, dir = getwd(), winslash = "/", ...)
}
\arguments{
\item{x}{length one character: a file path}

\item{dir}{a reference directory}

\item{winslash}{path separator on windows, passed to \code{\link{normalizePath}}}

\item{...}{ignored arguments}
}
\description{
Absolutizes a filepath.  Somewhat the opposite of \code{\link{relativizePath}}.
}
\details{
x and dir are first normalized, then x is expressed relative to dir.
If x and dir are on different drives (i.e. C:/ D:/)
x is returned as an absolute path.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io_tablet.R
\name{io_csv.tablet}
\alias{io_csv.tablet}
\title{Export Tablet as CSV}
\usage{
\method{io_csv}{tablet}(x, file = "", ...)
}
\arguments{
\item{x}{tablet}

\item{file}{passed to \code{\link{io_tablet}}}

\item{...}{passed to \code{\link{io_tablet}}}
}
\value{
invisible(file)
}
\description{
Exports a tablet as comma-separated variable,
by means of \code{io_csv}.  Reminds the user
to prefer \code{io_tablet} for lossless recovery.
}
\examples{
example(io_tablet)

}
\seealso{
Other io: 
\code{\link{io_tablet}()},
\code{\link{io_tablet.character}()},
\code{\link{io_tablet.data.frame}()},
\code{\link{io_tablet.tablet}()}

Other interface: 
\code{\link{io_tablet.character}()},
\code{\link{io_tablet.data.frame}()},
\code{\link{io_tablet.tablet}()}
}
\concept{interface}
\concept{io}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tablet.R
\name{headerlist.tablette}
\alias{headerlist.tablette}
\title{Create Header List for Tablet}
\usage{
\method{headerlist}{tablette}(x, ...)
}
\arguments{
\item{x}{tablet}

\item{...}{ignored}
}
\value{
list of named integer where each element describes an additional header row with names as indicated, repeated integer times
}
\description{
Creates header list for tablet.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io_tablet.R
\name{io_tablet.character}
\alias{io_tablet.character}
\title{Import Tablet}
\usage{
\method{io_tablet}{character}(x, na.strings = "NA", strip.white = FALSE, ...)
}
\arguments{
\item{x}{character file path}

\item{na.strings}{passed to \code{\link[yamlet]{io_csv.character}}}

\item{strip.white}{passed to \code{\link[yamlet]{io_csv.character}}}

\item{...}{passed to \code{\link[yamlet]{io_csv.character}}}
}
\value{
data.frame
}
\description{
Imports a tablet.
A wrapper for \code{\link[yamlet]{io_csv.character}} that also
reads associated yamlet metadata, if present, and applies it
as attributes.
}
\examples{
example(io_tablet)
}
\seealso{
Other io: 
\code{\link{io_csv.tablet}()},
\code{\link{io_tablet}()},
\code{\link{io_tablet.data.frame}()},
\code{\link{io_tablet.tablet}()}

Other interface: 
\code{\link{io_csv.tablet}()},
\code{\link{io_tablet.data.frame}()},
\code{\link{io_tablet.tablet}()}
}
\concept{interface}
\concept{io}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tablet.R
\name{classifiers.data.frame}
\alias{classifiers.data.frame}
\title{Identify Classifiers for Data Frame}
\usage{
\method{classifiers}{data.frame}(x, ...)
}
\arguments{
\item{x}{data.frame}

\item{...}{ignored}
}
\value{
data.frame, grouped if x was (see \code{\link[dplyr]{group_by}})
}
\description{
Identifies classifiers: literally, groups. Supplies tablet_n and tablet_N: groupwise
and ungrouped counts.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gridtable.R
\name{as_calibrated.data.frame}
\alias{as_calibrated.data.frame}
\title{Calibrate Data Frame Column Widths}
\usage{
\method{as_calibrated}{data.frame}(x, width, ...)
}
\arguments{
\item{x}{A data frame.}

\item{width}{Character or numeric vector of column widths. Values are
recycled as necessary.}

\item{...}{Currently unused.}
}
\value{
A data frame with class \code{calibrated} and column-level
  \code{width} attributes.
}
\description{
Assign column-level \code{width} attributes to a data frame and return
a calibrated data frame for downstream table rendering. Widths are
recycled as necessary; a warning is issued when the number of supplied
widths differs from the number of columns.
}
\seealso{
\code{\link{as_calibrated}}, \code{\link[kableExtra]{column_spec}}

Other gridtable: 
\code{\link{as.data.frame.gridtable}()},
\code{\link{as_calibrated}()},
\code{\link{as_gridtable}()},
\code{\link{as_gridtable.character}()},
\code{\link{as_gridtable.data.frame}()},
\code{\link{kbl}()},
\code{\link{kbl.calibrated}()},
\code{\link{kbl.default}()},
\code{\link{kbl.gridtable}()},
\code{\link{print.gridtable}()}
}
\concept{gridtable}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tablette.R
\name{tablet.tablette}
\alias{tablet.tablette}
\title{Convert to tablette from tablet.}
\usage{
\method{tablet}{tablette}(x, ...)
}
\arguments{
\item{x}{object of dispatch}

\item{...}{passed arguments}
}
\value{
tablette : a data.frame
 with columns for each combination of groups, and:
\item{_tablet_name}{observation identifier: character 
has a codelist attribute the values of which are the original column names}
\item{_tablet_level}{factor level, or the LHS of formulas in 'num'}
\item{All (or value of 'all' argument)}{ungrouped results}
}
\description{
Converts to 'tablette' from 'tablet'.
I.e., makes compact data.frame that emulates
the layout expected by \code{\link{as_kable}}.
Preserves attributes 'n', 'label', and 'codelist' where present.
}
\examples{
library(boot)
library(dplyr)
library(magrittr)
library(yamlet)
rm(melanoma)
melanoma \%<>\% select(-time, -year) 
melanoma \%<>\% decorate('
time:      [ Survival Time Since Operation, day               ]
status:    [ End of Study Patient Status,
           [ Alive: 2, Melanoma Death: 1, Unrelated Death: 3 ]]
sex:       [ Sex, [ Male: 1, Female: 0                       ]]
age:       [ Age at Time of Operation, year                   ]
thickness: [ Tumor Thickness, mm                              ]
ulcer:     [ Ulceration, [ Absent: 0, Present: 1             ]]
')
melanoma \%<>\% resolve
melanoma \%<>\% group_by(status, ulcer) 
melanoma \%<>\% tablet 
melanoma \%>\% header_rows
melanoma \%>\% as_kable
melanoma \%>\% tablette
melanoma \%>\% tablette \%>\% tablet
identical(melanoma, tablet(tablette(melanoma)))
  
}
\seealso{
Other tablet: 
\code{\link{as_tablet}()},
\code{\link{as_tablet.data.frame}()},
\code{\link{header_rows}()},
\code{\link{header_rows.tablet}()},
\code{\link{tablette.tablet}()}
}
\concept{tablet}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gridtable.R
\name{print.gridtable}
\alias{print.gridtable}
\title{Print a gridtable}
\usage{
\method{print}{gridtable}(x, ...)
}
\arguments{
\item{x}{Character scalar with class \code{gridtable}.}

\item{...}{Currently unused.}
}
\value{
Invisibly returns \code{x}.
}
\description{
Print a gridtable as plain grid-table text.
}
\seealso{
Other gridtable: 
\code{\link{as.data.frame.gridtable}()},
\code{\link{as_calibrated}()},
\code{\link{as_calibrated.data.frame}()},
\code{\link{as_gridtable}()},
\code{\link{as_gridtable.character}()},
\code{\link{as_gridtable.data.frame}()},
\code{\link{kbl}()},
\code{\link{kbl.calibrated}()},
\code{\link{kbl.default}()},
\code{\link{kbl.gridtable}()}
}
\concept{gridtable}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tablet.R
\name{groupless}
\alias{groupless}
\title{Calculate Without Groups}
\usage{
groupless(x, ...)
}
\arguments{
\item{x}{object}

\item{...}{passed}
}
\value{
see methods
}
\description{
Calculates without groups.  Generic, with method \code{\link{groupless.data.frame}}.
It is meaningless to use this toolchain function on classes
'observations', 'devalued', or 'widgets'.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tablet.R
\name{escape_latex}
\alias{escape_latex}
\title{Escape Special Characters for Latex}
\usage{
escape_latex(x, ...)
}
\arguments{
\item{x}{typically inherits character}

\item{...}{passed arguments}
}
\value{
character
}
\description{
Escapes special characters in Latex context.  
Operates element-wise, using \code{\link{escape_latex.default}}
or \code{\link{escape_latex.latex}} depending
on whether the element appears to be latex-formatted.
}
\details{
An element is considered latex-formatted if it begins
with "\(" or "{}".  Note that even though
latex-like elements likely have already had reserved
characters escaped, \code{\link{escape_latex.latex}}
will attempt to fix certain anticipated escaping issues.
The function \code{\link{escape_latex.default}}
performs conventional escaping; the result is 
prepended with "{}" as a signal to downstream
processes that the text is considered latex-ready
(and needn't be escaped again).
}
\examples{
  escape_latex(
    c(
      '',
      ' ',
      '\t',
      '{}a',
      NA,
      '\\\\()',
      '3\%'
    )
  )
}
\seealso{
Other escape: 
\code{\link{escape_latex.default}()},
\code{\link{escape_latex.latex}()}
}
\concept{escape}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tablette.R
\name{tablette}
\alias{tablette}
\title{Convert to tablette}
\usage{
tablette(x, ...)
}
\arguments{
\item{x}{object of dispatch}

\item{...}{passed arguments}
}
\description{
Converts to 'tablette'.  Generic,
with method \code{\link{tablette.tablet}}.
}
\seealso{
Other tablette: 
\code{\link{tablette.tablet}()}
}
\concept{tablette}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tablet.R
\name{tablette.groupwise}
\alias{tablette.groupwise}
\title{Generate a Group-wise Tablette}
\usage{
\method{tablette}{groupwise}(x, ..., all = "All", lab = list(lab ~ name + "\\n(N = " + n + ")"))
}
\arguments{
\item{x}{groupwise}

\item{...}{formulas of the form lab ~ NULL will remove elements of 'lab'}

\item{all}{a column name for ungrouped statistics; can have length zero to suppress ungrouped column}

\item{lab}{a list of formulas to generate column labels; \\n is translated as <br> in html context; each formula in succession is applied to each column}
}
\value{
'tablette', with columns for each combination of groups, and:
\item{_tablet_name}{observation identifier}
\item{_tablet_level}{factor level (or special value 'numeric' for numerics)}
\item{_tablet_stat}{the LHS of formulas in 'fac' and 'num'}
\item{All (or value of 'all' argument)}{ungrouped results}
}
\description{
Generates a group-wise tablette. Calculates statistics
for all factors and numerics, with and without grouping
variables.  Column names represent finest level of
grouping, distinguished by attribute 'nest' (the values of
higher groups). Column names include 'all' column
for same statistics without groups. Result columns
have corresponding attribute 'n'. 'lab' supplies a label attribute
for each column where the RHS succeeds: by default appending 'n' to result column names.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tablet.R
\name{groupwise.data.frame}
\alias{groupwise.data.frame}
\title{Calculate Widgets With and Without Groups}
\usage{
\method{groupwise}{data.frame}(
  x,
  fun = list(mis ~ sum(is.na(x)), sum ~ sum(x, na.rm = TRUE), pct ~ signif(digits = 3,
    sum/n * 100), ave ~ signif(digits = 3, mean(x, na.rm = TRUE)), std ~ signif(digits =
    3, sd(x, na.rm = TRUE)), med ~ signif(digits = 3, median(x, na.rm = TRUE)), min ~
    signif(digits = 3, min(x, na.rm = TRUE)), max ~ signif(digits = 3, max(x, na.rm =
    TRUE))),
  fac = list(` ` ~ sum + " (" + pct + "\%" + ")"),
  num = list(`Mean (SD)` ~ ave + " (" + std + ")", `Median (range)` ~ med + " (" + min +
    ", " + max + ")", Missing ~ mis),
  ...
)
}
\arguments{
\item{x}{data.frame}

\item{fun}{passed to groupfull() and groupless()}

\item{fac}{passed to groupfull() and groupless()}

\item{num}{passed to groupfull() and groupless()}

\item{...}{passed to groupfull() and groupless()}
}
\value{
class 'groupwise', arranged by groups:
\item{_tablet_name}{observation identifier}
\item{_tablet_level}{factor level or LHS of formulas in 'num'}
\item{_tablet_N}{number of records}
\item{_tablet_n}{number of records in group}
\item{_tablet_groupfull}{the LHS of formulas in 'fac' and 'num'}
\item{_tablet_groupless}{the LHS of formulas in 'fac' and 'num'}
}
\description{
Calculates widgets with and without groups.
Supplies 'groupfull' and 'groupless' (prefixed) columns instead of 'widgets'.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tablette.R
\name{header_rows}
\alias{header_rows}
\title{Identify Header Rows}
\usage{
header_rows(x, ...)
}
\arguments{
\item{x}{object of dispatch}

\item{...}{passed arguments}
}
\description{
Identifies header rows.
Generic, with method \code{\link{header_rows.tablet}}.
}
\seealso{
Other tablet: 
\code{\link{as_tablet}()},
\code{\link{as_tablet.data.frame}()},
\code{\link{header_rows.tablet}()},
\code{\link{tablet.tablette}()},
\code{\link{tablette.tablet}()}
}
\concept{tablet}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tablet.R
\name{observations.data.frame}
\alias{observations.data.frame}
\title{Identify Observations for Data Frame}
\usage{
\method{observations}{data.frame}(x, ..., exclude_name = NULL)
}
\arguments{
\item{x}{data.frame}

\item{...}{passed to categoricals(), numerics()}

\item{exclude_name}{whether to drop NA levels of name; passed to \code{\link[tidyr]{gather}}}
}
\value{
class 'observations' arranged by groups (presented first):
\item{_tablet_N}{number of records}
\item{_tablet_n}{number of records in group}
\item{_tablet_name}{observation identifier}
\item{_tablet_level}{factor level (or special value 'numeric' for numerics)}
\item{_tablet_value}{numeric value (or special value 1 for factors)}
}
\description{
Identifies observations. Stacks values by level, supplying defaults as necessary.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tablet.R
\name{classifiers}
\alias{classifiers}
\title{Identify Classifiers}
\usage{
classifiers(x, ...)
}
\arguments{
\item{x}{object}

\item{...}{passed}
}
\value{
see methods
}
\description{
Identifies classifiers.  Generic, with method \code{\link{classifiers.data.frame}}.
}
\examples{
library(dplyr)
x <- data.frame(grp = 1:2, a = factor(1:2), b = 1:2, c = factor(1:2, levels = 2:1))
x <- group_by(x, grp)
classifiers(x)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tablet.R
\name{escape_latex.latex}
\alias{escape_latex.latex}
\title{Escape Latex for Latex-formated Text}
\usage{
escape_latex.latex(x, secondary = TRUE, primary = FALSE, ...)
}
\arguments{
\item{x}{character}

\item{secondary}{logical: whether secondary backslashes should be pre-doubled}

\item{primary}{logical: whether first backslashes should be pre-doubled}

\item{...}{ignored}
}
\value{
character
}
\description{
Returns argument typically unmodified.
Prevents accidental double-escaping
of the same text.
}
\details{
This function is used in \code{\link{as_kable.tablet}} in
preparation for a call to \code{\link[kableExtra]{kbl}}.
At kableExtra 1.3.4, sim_double_escape() only doubles
primary (leading) backslashes; the \code{secondary} argument
by default pre-doubles later backslashes to prevent
certain display errors.  Behavior may change if kableExtra changes.
}
\examples{
identical(
escape_latex('([#$\%&_{}])'),
escape_latex(escape_latex('([#$\%&_{}])'))
)

}
\seealso{
Other escape: 
\code{\link{escape_latex}()},
\code{\link{escape_latex.default}()}
}
\concept{escape}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tablet.R
\name{devalued.observations}
\alias{devalued.observations}
\title{Aggregate Values for Observations
Aggregates values for observations.
Accepts a list of formulas with result name on left and
aggregating expression on the right.
Behavior is undefined if any expression does not aggregate! (I.e., length != 1).
Expressions are evaluated in an environment where values are available as 'x',
grouped count is available as 'n' and ungouped count is available as 'N'.}
\usage{
\method{devalued}{observations}(
  x,
  ...,
  fun = list(mis ~ sum(is.na(x)), sum ~ sum(x, na.rm = TRUE), pct ~ signif(digits = 3,
    sum/n * 100), ave ~ signif(digits = 3, mean(x, na.rm = TRUE)), std ~ signif(digits =
    3, sd(x, na.rm = TRUE)), med ~ signif(digits = 3, median(x, na.rm = TRUE)), min ~
    signif(digits = 3, min(x, na.rm = TRUE)), max ~ signif(digits = 3, max(x, na.rm =
    TRUE))),
  silent = TRUE
)
}
\arguments{
\item{x}{observations}

\item{...}{passed formulas with matching LHS will replace corresponding element of fun.}

\item{fun}{default aggregate functions expressed as formulas}

\item{silent}{whether to suppress warnings from evaluations of 'fun'}
}
\value{
class 'devalued', presumably one record per group:
\item{_tablet_N}{number of records}
\item{_tablet_n}{number of records in group}
\item{_tablet_name}{observation identifier}
\item{_tablet_level}{factor level (or special value 'numeric' for numerics)}
\item{(other)}{additional column for each statistic in 'fun'}
}
\description{
Aggregate Values for Observations
Aggregates values for observations.
Accepts a list of formulas with result name on left and
aggregating expression on the right.
Behavior is undefined if any expression does not aggregate! (I.e., length != 1).
Expressions are evaluated in an environment where values are available as 'x',
grouped count is available as 'n' and ungouped count is available as 'N'.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gridtable.R
\name{as.data.frame.gridtable}
\alias{as.data.frame.gridtable}
\title{Coerce a gridtable to a Calibrated Data Frame}
\usage{
\method{as.data.frame}{gridtable}(
  x,
  row.names = NULL,
  optional = FALSE,
  ...,
  scale = getOption("tablet.gridtable.scale", 1)
)
}
\arguments{
\item{x}{Character scalar with class \code{gridtable}.}

\item{row.names}{Optional row names for the returned data frame.}

\item{optional}{Currently unused; accepted for compatibility with
\code{as.data.frame}.}

\item{...}{Currently unused.}

\item{scale}{Numeric scalar multiplier applied to grid-table dash counts
when deriving column-level \code{width} attributes.}
}
\value{
A data frame with class \code{c("calibrated", "data.frame")}.
  Each column has a \code{width} attribute derived from the final grid
  table dash count multiplied by \code{scale} and encoded as an em width.
}
\description{
Parse a validated gridtable object directly to a calibrated data frame.
Columns have \code{width} attributes suggested by the grid table spans.
Continuation lines within a grid-table cell are collapsed with single spaces.
Cell markup is otherwise left untouched for downstream renderers.
}
\seealso{
\code{\link{as_gridtable.character}}, \code{\link{as_calibrated.data.frame}}

Other gridtable: 
\code{\link{as_calibrated}()},
\code{\link{as_calibrated.data.frame}()},
\code{\link{as_gridtable}()},
\code{\link{as_gridtable.character}()},
\code{\link{as_gridtable.data.frame}()},
\code{\link{kbl}()},
\code{\link{kbl.calibrated}()},
\code{\link{kbl.default}()},
\code{\link{kbl.gridtable}()},
\code{\link{print.gridtable}()}
}
\concept{gridtable}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gridtable.R
\name{as_gridtable.data.frame}
\alias{as_gridtable.data.frame}
\title{Convert a Data Frame to gridtable}
\usage{
\method{as_gridtable}{data.frame}(x, ...)
}
\arguments{
\item{x}{A data frame.}

\item{...}{Currently unused.}
}
\value{
Character scalar with class \code{gridtable} containing a canonical
  Pandoc grid table.
}
\description{
Build a simple Pandoc grid table from a data frame. Column names become the
grid-table header row, each data-frame row becomes one body row, and column
widths are calibrated from the widest formatted value in each column.
}
\seealso{
\code{\link{as_gridtable}}, \code{\link{as_gridtable.character}}

Other gridtable: 
\code{\link{as.data.frame.gridtable}()},
\code{\link{as_calibrated}()},
\code{\link{as_calibrated.data.frame}()},
\code{\link{as_gridtable}()},
\code{\link{as_gridtable.character}()},
\code{\link{kbl}()},
\code{\link{kbl.calibrated}()},
\code{\link{kbl.default}()},
\code{\link{kbl.gridtable}()},
\code{\link{print.gridtable}()}
}
\concept{gridtable}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tablet.R
\name{index}
\alias{index}
\title{Create Index}
\usage{
index(x, ...)
}
\arguments{
\item{x}{object}

\item{...}{passed}
}
\value{
see methods
}
\description{
Creates index.  Generic, with method \code{\link{index.tablette}}.
}
\examples{
example(classifiers)
index(tablette(groupwise(x)))
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tablet.R
\name{groupwise}
\alias{groupwise}
\title{Calculate With and Without Groups}
\usage{
groupwise(x, ...)
}
\arguments{
\item{x}{object}

\item{...}{passed}
}
\value{
see methods
}
\description{
Calculates with and without groups.  Generic, with method \code{\link{groupwise.data.frame}}.
It is meaningless to use this toolchain function on classes
'observations', 'devalued', 'widgets', 'groupless', or 'groupfull'.
}
\examples{
example(classifiers)
groupwise(x)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tablet.R
\name{as_kable}
\alias{as_kable}
\title{Coerce to Kable}
\usage{
as_kable(x, ...)
}
\arguments{
\item{x}{object}

\item{...}{passed arguments}
}
\value{
see methods
}
\description{
Coerces to kable output class.  Generic,
with method \code{\link{as_kable.tablet}}.
}
\examples{
example(classifiers)
as_kable(tablet(tablette(groupwise(x))))
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xtable.R
\name{as_xtable.tablet}
\alias{as_xtable.tablet}
\title{Coerce Tablet to Xtable}
\usage{
\method{as_xtable}{tablet}(
  x,
  caption = NULL,
  label = NULL,
  align = NULL,
  digits = NULL,
  display = NULL,
  auto = FALSE,
  variable = " ",
  format_name = function(x, ...) paste0("\\\\!\\\\!\\\\textbf{", x, "}"),
  format_stat = function(x, ...) x,
  format_value = function(x, ...) x,
  ...
)
}
\arguments{
\item{x}{\code{\link{tablet}}}

\item{caption}{passed to \code{\link[xtable]{xtableList}}}

\item{label}{passed to \code{\link[xtable]{xtableList}}}

\item{align}{passed to \code{\link[xtable]{xtableList}}}

\item{digits}{passed to \code{\link[xtable]{xtableList}}}

\item{display}{passed to \code{\link[xtable]{xtableList}}}

\item{format_name}{function to format variable names (accepts at least x and dots)}

\item{format_stat}{function to format names of statistics and factor levels (accepts at least x and dots)}

\item{format_value}{function to format cell values (accepts at least x and dots)}

\item{...}{passed to \code{\link[xtable]{xtableList}}}
}
\value{
like \code{\link[xtable]{xtableList}}
}
\description{
Renders a tablet as xtable.  Calls \code{\link[xtable]{xtable}} and implements
default aesthetics.
}
\examples{
library(boot)
library(dplyr)
library(magrittr)
library(xtable)
melanoma \%>\%
  select(-time, -year) \%>\%
  mutate(sex = factor(sex), ulcer = factor(ulcer)) \%>\%
  group_by(status) \%>\%
  tablet \%>\%
  as_xtable
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tablet.R
\name{widgets}
\alias{widgets}
\title{Calculate Widgets}
\usage{
widgets(x, ...)
}
\arguments{
\item{x}{object}

\item{...}{passed}
}
\value{
see methods
}
\description{
Calculates widgets.  Generic, with method \code{\link{widgets.devalued}}.
}
\examples{
example(classifiers)
widgets(devalued(observations(x)))
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tablet.R
\name{groupfull}
\alias{groupfull}
\title{Calculate With Groups}
\usage{
groupfull(x, ...)
}
\arguments{
\item{x}{object}

\item{...}{passed}
}
\value{
see methods
}
\description{
Calculates with groups.  Generic, with method \code{\link{groupfull.data.frame}}.
It is meaningless to use this toolchain function on classes
'observations', 'devalued', or 'widgets'.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gridtable.R
\name{kbl.calibrated}
\alias{kbl.calibrated}
\title{Render a Calibrated Data Frame with kbl}
\usage{
\method{kbl}{calibrated}(x, ..., use_widths = TRUE)
}
\arguments{
\item{x}{A data frame with class \code{calibrated}.}

\item{...}{Additional arguments passed to \code{kbl.default}.}

\item{use_widths}{Logical scalar indicating whether column-level width
attributes should be applied with \code{\link[kableExtra]{column_spec}}. Defaults to TRUE.}
}
\value{
A kable object.
}
\description{
Render a calibrated data frame with \code{kbl} and optionally apply
column-level \code{width} attributes with \code{\link[kableExtra]{column_spec}}.
}
\seealso{
\code{\link{kbl}}, \code{\link{kbl.default}}, \code{\link[kableExtra]{column_spec}}

Other gridtable: 
\code{\link{as.data.frame.gridtable}()},
\code{\link{as_calibrated}()},
\code{\link{as_calibrated.data.frame}()},
\code{\link{as_gridtable}()},
\code{\link{as_gridtable.character}()},
\code{\link{as_gridtable.data.frame}()},
\code{\link{kbl}()},
\code{\link{kbl.default}()},
\code{\link{kbl.gridtable}()},
\code{\link{print.gridtable}()}
}
\concept{gridtable}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tablette.R
\name{as_tablet.data.frame}
\alias{as_tablet.data.frame}
\title{Coerce data.frame to tablet}
\usage{
\method{as_tablet}{data.frame}(x, ...)
}
\arguments{
\item{x}{data.frame}

\item{...}{passed arguments}
}
\value{
tablet
}
\description{
Coerces data.frame to tablet.  Checks format and assigns the class.
See \code{\link{tablet.data.frame}}.
}
\seealso{
Other tablet: 
\code{\link{as_tablet}()},
\code{\link{header_rows}()},
\code{\link{header_rows.tablet}()},
\code{\link{tablet.tablette}()},
\code{\link{tablette.tablet}()}
}
\concept{tablet}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tablet.R
\name{groupless.data.frame}
\alias{groupless.data.frame}
\title{Calculate Widgets Without Groups}
\usage{
\method{groupless}{data.frame}(x, ...)
}
\arguments{
\item{x}{data.frame}

\item{...}{passed to observations(), devalued(), widgets()}
}
\value{
class 'groupless', with output like \code{\link{widgets.devalued}}
}
\description{
Calculates widgets without groups.
Removes grouping variables and groups, then executes:
data.frame -> observations -> devalued -> widgets.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io_tablet.R
\name{io_tablet}
\alias{io_tablet}
\title{Import and Export Tablet}
\usage{
io_tablet(x, ...)
}
\arguments{
\item{x}{object}

\item{...}{passed arguments}
}
\value{
See methods.
}
\description{
Imports or exports tablet as comma-separated variable.
Generic, with methods that extend \code{\link[yamlet]{io_csv}}.
}
\examples{
library(yamlet)
library(boot)
library(dplyr)
library(magrittr)
library(tablet)
library(kableExtra)
melanoma \%<>\% redecorate('
time: [ Time, day ]
status: [ Patient Status, [ Died^1: 1, Alive^2: 2, Unrelated: 3 ]]
sex: [ Sex^3, [ Male: 1, Female^4: 0 ]]
age: [ Age, year ]
year: [ Year of Operation ]
ulcer: [ Ulceration, [ Present: 1, Absent: 0 ]]
thickness: [ Tumor Thickness, mm ]
')
tbl <- melanoma \%>\%
  select(-time, -year) \%>\%
  group_by(status, ulcer) \%>\%
  enscript \%>\%
  tablet(Missing~NULL)
path <- tbl \%>\% io_tablet(tempfile(fileext = '.csv'))
# alternatively:
path <- tbl \%>\% io_csv(tempfile(fileext = '.csv'))
tab <- path \%>\% io_tablet

tbl \%>\% 
kbl(booktabs = TRUE) \%>\%
kable_styling(latex_options = 'scale_down')

tab \%>\% 
kbl(booktabs = TRUE) \%>\%
kable_styling(latex_options = 'scale_down')
}
\seealso{
Other io: 
\code{\link{io_csv.tablet}()},
\code{\link{io_tablet.character}()},
\code{\link{io_tablet.data.frame}()},
\code{\link{io_tablet.tablet}()}
}
\concept{io}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gridtable.R
\name{as_gridtable}
\alias{as_gridtable}
\title{Coerce to Gridtable}
\usage{
as_gridtable(x, ...)
}
\arguments{
\item{x}{object}

\item{...}{passed arguments}
}
\value{
see methods
}
\description{
Coerces to "gridtable", i.e. the Pandoc construct "grid table" Generic,
with method \code{\link{as_gridtable.character}}.
}
\examples{
# as_gridtable() lets you create and render small data tables on-the-fly.

library(magrittr)
library(tablet)
library(yamlet)
library(kableExtra)
'+--+--+--+
|study | subjects | doses |
+==+==+==+
| 001 | 20 | 5 mg, 10 mg |
+--+--+--+
| 002 | 40 | 5 mg, 10 mg |
|     |    | 15 mg, 20 mg, 30 mg |
+--+--+--+' \%>\% 
as_gridtable

# We see that as_gridtable cleans up our Pandoc grid table markdown.
# We copy it here and render it with kableExtra::kbl().

'
+-------+----------+---------------------+
| study | subjects | doses               |
+=======+==========+=====================+
| 001   | 20       | 5 mg, 10 mg         |
+-------+----------+---------------------+
| 002   | 40       | 5 mg, 10 mg         |
|       |          | 15 mg, 20 mg, 30 mg |
+-------+----------+---------------------+
' \%>\%
as_gridtable \%>\%
as.data.frame(scale = 0.9) \%>\% # or: options(tablet.gridtable.scale = 0.9)
decorations

# We see that a gridtable when converted to data.frame
# has scalable column width attributes.
# Widths are controlled by longest line of text, including column name.
# kbl will try to honor widths.
# kbl() will call as.data.frame() implicitly if you pass a gridtable:

'
+-------+----------+---------------------+
| study | subjects | doses               |
+=======+==========+=====================+
| 001   | 20       | 5 mg, 10 mg         |
+-------+----------+---------------------+
| 002   | 40       | 5 mg, 10 mg         |
|       |          | 15 mg, 20 mg, 30 mg |
+-------+----------+---------------------+
' \%>\%
as_gridtable \%>\%
kbl(
  label = "studies",
  booktabs = TRUE,
  caption = "Study Characteristics" 
) \%>\%
  kable_styling(font_size = 9) \%>\%
  footnote(
    threeparttable = TRUE,
    general_title = "",
    general = "For more details, see study protocols."
  )
  
# We can start with a data.frame:

data.frame(
  study = c('001','002'), 
  subjects = c(20, 40)
) \%>\%
as_gridtable

# We can integrate decorations:

'
+-------+----------+
| study | subjects |
+=======+==========+
| 001   | 20       |
+-------+----------+
| 002   | 40       |
+-------+----------+
' \%>\%
as_gridtable \%>\%
as.data.frame \%>\%
decorate('
  study: [ Study, [ ST001: "001", ST002: "002" ]]
  subjects: [ N Subjects ]
') \%>\%
resolve \%>\%
kbl(
  label = "studies",
  booktabs = TRUE,
  caption = "Study Characteristics"
) \%>\%
kable_styling()
}
\seealso{
Other gridtable: 
\code{\link{as.data.frame.gridtable}()},
\code{\link{as_calibrated}()},
\code{\link{as_calibrated.data.frame}()},
\code{\link{as_gridtable.character}()},
\code{\link{as_gridtable.data.frame}()},
\code{\link{kbl}()},
\code{\link{kbl.calibrated}()},
\code{\link{kbl.default}()},
\code{\link{kbl.gridtable}()},
\code{\link{print.gridtable}()}
}
\concept{gridtable}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gridtable.R
\name{kbl}
\alias{kbl}
\title{Render with kbl}
\usage{
kbl(x, ...)
}
\arguments{
\item{x}{object}

\item{...}{passed arguments}
}
\value{
see methods
}
\description{
Generic wrapper for \code{\link[kableExtra]{kbl}}. The default method is the
existing \code{kableExtra::kbl} definition.
}
\seealso{
\code{\link[kableExtra]{kbl}}

Other gridtable: 
\code{\link{as.data.frame.gridtable}()},
\code{\link{as_calibrated}()},
\code{\link{as_calibrated.data.frame}()},
\code{\link{as_gridtable}()},
\code{\link{as_gridtable.character}()},
\code{\link{as_gridtable.data.frame}()},
\code{\link{kbl.calibrated}()},
\code{\link{kbl.default}()},
\code{\link{kbl.gridtable}()},
\code{\link{print.gridtable}()}
}
\concept{gridtable}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gridtable.R
\name{as_calibrated}
\alias{as_calibrated}
\title{Coerce to Calibrated}
\usage{
as_calibrated(x, ...)
}
\arguments{
\item{x}{object}

\item{...}{passed arguments}
}
\value{
see methods
}
\description{
Coerces to calibrated output class. Generic, with flagship method
\code{\link{as_calibrated.data.frame}}. The intent is to attach
calibrated column widths for downstream table rendering.
}
\seealso{
Other gridtable: 
\code{\link{as.data.frame.gridtable}()},
\code{\link{as_calibrated.data.frame}()},
\code{\link{as_gridtable}()},
\code{\link{as_gridtable.character}()},
\code{\link{as_gridtable.data.frame}()},
\code{\link{kbl}()},
\code{\link{kbl.calibrated}()},
\code{\link{kbl.default}()},
\code{\link{kbl.gridtable}()},
\code{\link{print.gridtable}()}
}
\concept{gridtable}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/recap.R
\name{recap}
\alias{recap}
\title{Recap Something}
\usage{
recap(x, ...)
}
\arguments{
\item{x}{object of dispatch}

\item{...}{passed arguments}
}
\description{
Recaps something.
Generic, with method \code{\link{recap.knitr_kable}}.
}
\seealso{
Other recap: 
\code{\link{recap.knitr_kable}()}
}
\concept{recap}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tablet.R
\name{numerics.data.frame}
\alias{numerics.data.frame}
\title{Identify Numerics for Data Frame}
\usage{
\method{numerics}{data.frame}(x, ..., na.rm_num = FALSE, all_levels = FALSE)
}
\arguments{
\item{x}{data.frame; names must not include 'name' or 'level'}

\item{...}{passed to \code{\link{classifiers}}}

\item{na.rm_num}{whether to drop NA observations; passed to \code{\link[tidyr]{gather}} as na.rm}

\item{all_levels}{whether to supply records for unobserved levels}
}
\value{
same class as x
}
\description{
Identifies numeric columns. Stacks values and supplies factor level 'numeric'.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tablet.R
\name{splice}
\alias{splice}
\title{Splice Some Things Together}
\usage{
splice(x, ...)
}
\arguments{
\item{x}{object}

\item{...}{passed arguments}
}
\value{
see methods
}
\description{
Spices some things together. Generic, with tablet-oriented method \code{\link{splice.data.frame}}.
}
\examples{
example(splice.data.frame)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tablette.R
\name{header_rows.tablet}
\alias{header_rows.tablet}
\title{Identify Header Rows for tablet}
\usage{
\method{header_rows}{tablet}(x, ...)
}
\arguments{
\item{x}{tablet}

\item{...}{ignored}
}
\value{
integer: indices for those rows representing headers
}
\description{
Identifies header rows for tablet.
}
\seealso{
Other tablet: 
\code{\link{as_tablet}()},
\code{\link{as_tablet.data.frame}()},
\code{\link{header_rows}()},
\code{\link{tablet.tablette}()},
\code{\link{tablette.tablet}()}
}
\concept{tablet}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tablet.R
\name{escape_latex.default}
\alias{escape_latex.default}
\title{Escape Special Characters for Latex by Default}
\usage{
escape_latex.default(x, secondary = TRUE, ...)
}
\arguments{
\item{x}{caracter}

\item{secondary}{logical: whether secondary backslashes should be pre-doubled}

\item{...}{ignored}
}
\value{
character
}
\description{
Escapes text characters that have special meaning in latex.
Adapted with gratitude from KableExtra internals.
I.e. inactivates material that otherwise looks like latex.
}
\details{
This function is used in \code{\link{as_kable.tablet}} in
preparation for a call to \code{\link[kableExtra]{kbl}}.
At kableExtra 1.3.4, sim_double_escape() only doubles
primary (leading) backslashes; the \code{secondary} argument
by default pre-doubles later backslashes to prevent
certain display errors.  Behavior may change if kableExtra changes.
}
\examples{
escape_latex('([#$\%&_{}])')
}
\seealso{
Other escape: 
\code{\link{escape_latex}()},
\code{\link{escape_latex.latex}()}
}
\concept{escape}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tablet.R
\name{tablet.data.frame}
\alias{tablet.data.frame}
\title{Generate a Tablet for Data Frame}
\usage{
\method{tablet}{data.frame}(
 x,
 ...,
 na.rm = FALSE,
 all = 'All',
 fun = list(
  mis ~ sum(is.na(x)),
  sum ~ sum(x,  na.rm = TRUE),
  pct ~ signif(digits = 3,     sum / n * 100        ),
  ave ~ signif(digits = 3,    mean(x,  na.rm = TRUE)),
  std ~ signif(digits = 3,      sd(x,  na.rm = TRUE)),
  med ~ signif(digits = 3,  median(x,  na.rm = TRUE)),
  min ~ signif(digits = 3,     min(x,  na.rm = TRUE)),
  max ~ signif(digits = 3,     max(x,  na.rm = TRUE))
 ),
 fac = list(
  ` ` ~ sum + ' (' + pct + '\%' + ')'
 ),
 num = list(
  `Mean (SD)` ~ ave + ' (' + std + ')',
  `Median (range)` ~ med + ' (' + min + ', ' + max + ')',
  `Missing` ~ mis
  ),
 lab = list(
  lab ~ name + '\\n(N = ' + n + ')'
 ),
 na.rm_fac = na.rm,
 na.rm_num = na.rm,
 exclude_fac = NULL,
 exclude_name = NULL,
 all_levels = FALSE
)
}
\arguments{
\item{x}{data.frame (possibly grouped)}

\item{...}{substitute formulas for elements of fun, fac, num, lab; if RHS is NULL, element is removed}

\item{na.rm}{whether to remove NA in general}

\item{all}{a column name for ungrouped statistics; can have length zero to suppress ungrouped column}

\item{fun}{default aggregate functions expressed as formulas}

\item{fac}{a list of formulas to generate widgets for factors}

\item{num}{a list of formulas to generate widgets for numerics}

\item{lab}{a list of formulas to generate label attributes for columns (see details)}

\item{na.rm_fac}{whether to drop NA 'factor' observations; passed to \code{\link[tidyr]{gather}} as na.rm, interacts with exclude_fac}

\item{na.rm_num}{whether to drop NA numeric observations; passed to \code{\link[tidyr]{gather}} as na.rm}

\item{exclude_fac}{which factor levels to exclude; see \code{\link{factor}} (exclude)}

\item{exclude_name}{whether to drop NA values of column name (for completeness); passed to \code{\link[tidyr]{gather}}}

\item{all_levels}{whether to supply records for unobserved levels}
}
\value{
'tablet'
A tablet is a special case of data.frame with grouped rows and columns.
\item{*}{There is always one level of row groups.}
\item{*}{There can be any number of column groups, including zero.}
\item{*}{All columns are character (as tested by \code{is.character()}).}
\item{*}{The first column has empty strings that represent
the last non-empty value. It can be class 'latex' or 'character'.}
\item{*}{Leading element(s) of first column are deliberately blank (one space character)
and correspond to header rows.  See \code{\link{header_rows}}.}
\item{*}{The second column represents group-specific property 
names. It is populated always and only where column 1 is not.}
\item{*}{All other columns represent group-specific property values; 
elements before the first non-empty group value represent nested headers.}
\item{*}{Header values may be repeated.}
\item{*}{Header values may be empty strings, representing the last non-empty value
to the left, or single spaces, which are deliberately blank.}
\item{*}{Internally, character NA is equivalent to an empty string.}
}
\description{
Generates a 'tablet': a summary table of
formatted statistics for factors (is.factor()) and
numerics (is.numeric()) in x, with and without grouping
variables (if present, see \code{\link[dplyr]{group_by}}).
Column names represent finest level of
grouping, distinguished by attribute 'nest' (the values of
higher other groups, if any) along with the 'all' column
for ungrouped statistics. Column attribute 'n' indicates
relevant corresponding observation count.
Input should not have column names beginning with '_tablet'.
}
\details{
Arguments 'fun', 'fac', 'num', and 'lab' are lists
of two-sided formulas that are evaluated in
an environment where '+' expresses concatenation
(for character elements).
The values of LHS should be unique across all four lists.
'fun' is a list of aggregate statistics that have access to
\code{N} (number of original records),
\code{n} (number of group members), and
\code{x} (the numeric observations, or 1 for each factor value).
Aggregate statistics generated by 'fun' are available
for use in 'fac' and 'num'
which create visualizations thereof ('widgets'). Column-specific
attributes are available to elements of 'lab', including
the special attribute \code{name} (the current column name).
For 'lab' only, if the RHS succeeds, it becomes the label
attribute of the corresponding output column. 'lab' is used
here principally to support annotation of *output*
columns; if *input* columns have attributes 'label' or 'title'
(highest priority) those will have been already substituted
for default column names at the appropriate positions in the
output.

Missingness of observations (and to a lesser extent, levels of
grouping variables) merits special consideration.
Be aware that na.rm_fac and na.rm_num take their defaults
from na.rm.  Furthermore, na.rm_fac may interact with
exclude_fac, which is passed to \code{\link{factor}} as \code{exclude}.
The goal is to support all possible ways of expressing or ignoring
missingness. That said, if aggregate functions are removing
NA, the values of arguments beginning with 'na.rm' or 'exclude'
may not matter.
}
\examples{
library(boot)
library(dplyr)
library(magrittr)
melanoma \%>\%
  select(-time, -year) \%>\%
  mutate(sex = factor(sex), ulcer = factor(ulcer)) \%>\%
  group_by(status) \%>\%
  tablet \%>\%
  kbl
  
## drop 'Missing', redefine 'range'
melanoma \%>\%
  select(-time, -year) \%>\%
  mutate(sex = factor(sex), ulcer = factor(ulcer)) \%>\%
  group_by(status) \%>\%
  tablet(
    Missing ~ NULL,
    `Median (range)` ~ med + ' (' + min + ' to ' + max + ')'
  ) \%>\%
  kbl
  
}
\seealso{
\code{\link{as_kable.tablet}} \code{\link{kbl.tablet}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gridtable.R
\name{as_gridtable.character}
\alias{as_gridtable.character}
\title{Validate and Normalize a Simple Pandoc Grid Table}
\usage{
\method{as_gridtable}{character}(x, ...)
}
\arguments{
\item{x}{Character scalar containing one simple Pandoc grid table.}

\item{...}{Currently unused.}
}
\value{
Character scalar with class \code{gridtable} containing a canonical Pandoc grid table.
}
\description{
Validate a character string containing a simple Pandoc grid table and
rebuild it by parsing structural `+` and `|` delimiters, discarding the
original spacing, and returning an aligned grid table.
}
\details{
The accepted grid-table form is intentionally restricted: every boundary row
must contain the same number of `+` delimiters; every text row is split on
every `|` and must therefore contain the same number of `|` delimiters; each
boundary span must contain at least one `-` or `=`; the first boundary row and
all body boundary rows must use `-`; the second boundary row must be the only
`=` boundary row; and exactly one header text row must appear between the
first and second boundary rows; header cells must be non-empty and
unique after trimming. Literal `|` characters inside cell text are interpreted
as delimiters, not cell content.
}
\seealso{
Other gridtable: 
\code{\link{as.data.frame.gridtable}()},
\code{\link{as_calibrated}()},
\code{\link{as_calibrated.data.frame}()},
\code{\link{as_gridtable}()},
\code{\link{as_gridtable.data.frame}()},
\code{\link{kbl}()},
\code{\link{kbl.calibrated}()},
\code{\link{kbl.default}()},
\code{\link{kbl.gridtable}()},
\code{\link{print.gridtable}()}
}
\concept{gridtable}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tablet.R
\name{headerlist}
\alias{headerlist}
\title{Create Header List.}
\usage{
headerlist(x, ...)
}
\arguments{
\item{x}{object}

\item{...}{passed}
}
\value{
see methods
}
\description{
Creates header list.  Generic, with method \code{\link{headerlist.tablette}}.
}
\examples{
example(classifiers)
headerlist(tablette(groupwise(x)))
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tablet.R
\name{widgets.devalued}
\alias{widgets.devalued}
\title{Calculate Widgets for Devalued}
\usage{
\method{widgets}{devalued}(
  x,
  fac = list(` ` ~ sum + " (" + pct + "\%" + ")"),
  num = list(`Mean (SD)` ~ ave + " (" + std + ")", `Median (range)` ~ med + " (" + min +
    ", " + max + ")", Missing ~ mis),
  ...
)
}
\arguments{
\item{x}{devalued}

\item{fac}{a list of formulas to generate widgets for factors}

\item{num}{a list of formulas to generate widgets for numerics}

\item{...}{formulas with matching LHS replace defaults; x ~ NULL removes x}
}
\value{
class 'widgets', arranged by groups:
\item{_tablet_name}{observation identifier}
\item{_tablet_level}{factor level (or special value 'numeric' for numerics)}
\item{_tablet_N}{number of records}
\item{_tablet_n}{number of records in group}
\item{_tablet_stat}{the LHS of formulas in 'fac' and 'num'}
\item{_tablet_widget}{the RHS of formulas in 'fac' and 'num' (evaluated)}
}
\description{
Calculates widgets for class 'devalued'.
In this context, a widget is a text fragment
that formats one or more aggregate statistics,
all of which must have been present as LHS of
some element of 'fun' in the call to devalued().
Furthermore, LHS must be unique across 'fac' and
'num' for expected behavior.
}
\details{
In the context of this call, "+" is redefined
to allow concatenation of text.
Evaluation proceeds left to right as usual.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tablet.R
\name{index.tablette}
\alias{index.tablette}
\title{Create Index for Tablet}
\usage{
\method{index}{tablette}(x, ...)
}
\arguments{
\item{x}{tablet}

\item{...}{ignored}
}
\value{
named integer giving row groupings with names as indicated, repeated integer times
}
\description{
Creates index for tablet.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/recap.R
\name{recap.knitr_kable}
\alias{recap.knitr_kable}
\title{Recap knitr_kable.}
\usage{
\method{recap}{knitr_kable}(x, cols = NULL, pos = "c", ...)
}
\arguments{
\item{x}{object of dispatch}

\item{cols}{number of columns to span; guesses \code{ncol(x)} by default}

\item{pos}{position of text: 'l','c' (default), or 'r'}

\item{...}{ignored}
}
\description{
Recaps knitr_kable.
Specifically, it replaces the first non-tabled caption
with multicolumn text.  The intent is to
prevent repeat bookmarks when generating pdf.
}
\seealso{
Other recap: 
\code{\link{recap}()}
}
\concept{recap}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tablet.R
\name{tablet}
\alias{tablet}
\title{Generate a Tablet}
\usage{
tablet(x, ...)
}
\arguments{
\item{x}{object}

\item{...}{passed arguments}
}
\value{
see methods
}
\description{
Generates a tablet.  Generic. See \code{\link{tablet.data.frame}} and \code{\link{as_kable.tablet}}.
}
\examples{
example(tablet.data.frame)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mesa.R
\name{mesa}
\alias{mesa}
\title{Drag-and-drop Descriptive Statistics}
\usage{
mesa(launch.browser = TRUE, display.mode = "normal", ...)
}
\arguments{
\item{launch.browser}{passed to \code{\link[shiny]{runApp}}}

\item{display.mode}{passed to \code{\link[shiny]{runApp}}}

\item{...}{passed to \code{\link[shiny]{runApp}}}
}
\value{
used for side effects: launches shiny application
shinyWidgets',
}
\description{
Generate a table of descriptive statistics
by selecting columns from a file.
Currently supported formats include *.xpt, *.sas7bdat, and *.csv.
Launch the application using \code{mesa()} and use the interface
to select a data file, such as 'mtcars.xpt' under 'examples/')
(or select configuration file 'mtcars.conf' under 'examples/').
Then classify the
columns of interest to generate the corresponding
displays.
}
\details{
Currently,

* xpt files are read using the defaults for \code{\link[foreign]{read.xport}},

* sas7bdat files are read using the defaults for \code{\link[haven]{read_sas}}, and

* csv files are read using the defaults for \code{\link[csv]{as.csv}}.

If a file in the same directory has a corresponding base name but a \code{.yaml}
extension, it is treated as metadata and an attempt is made to apply it
to the internal version of the data. This file will not be over-written,
but it WILL be constructed if missing.  You can hand-edit it to supply
metadata.  See \code{?yamlet} for format; see the Variables tab for an easy interface.

This is a metadata-driven application.  Columns in the data that are *not*
in the metadata will be ignored, and columns in the metadata that are *not*
in the data will be constructed (maybe *all* of them).

The \code{\link[datasets]{mtcars}} datasets in the 'examples' volume is from \pkg{datasets}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tablet.R
\name{categoricals}
\alias{categoricals}
\title{Identify Categories}
\usage{
categoricals(x, ...)
}
\arguments{
\item{x}{object}

\item{...}{passed}
}
\value{
see methods
}
\description{
Identifies categories.  Generic, with method \code{\link{categoricals.data.frame}}.
}
\examples{
example(classifiers)
categoricals(x)

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gridtable.R
\name{kbl.gridtable}
\alias{kbl.gridtable}
\title{Render a gridtable with kbl}
\usage{
\method{kbl}{gridtable}(x, ...)
}
\arguments{
\item{x}{Character scalar with class \code{gridtable}.}

\item{...}{Additional arguments passed to \code{kbl}.}
}
\value{
A kable object.
}
\description{
Coerce a gridtable to a calibrated data frame with \code{as.data.frame}
before redispatching to \code{kbl}. This lets gridtable-derived column
widths be consumed by \code{kbl.calibrated}.
}
\seealso{
\code{\link{kbl}}, \code{\link{kbl.calibrated}}, \code{\link{as.data.frame.gridtable}}

Other gridtable: 
\code{\link{as.data.frame.gridtable}()},
\code{\link{as_calibrated}()},
\code{\link{as_calibrated.data.frame}()},
\code{\link{as_gridtable}()},
\code{\link{as_gridtable.character}()},
\code{\link{as_gridtable.data.frame}()},
\code{\link{kbl}()},
\code{\link{kbl.calibrated}()},
\code{\link{kbl.default}()},
\code{\link{print.gridtable}()}
}
\concept{gridtable}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xtable.R
\name{as_xtable}
\alias{as_xtable}
\title{Coerce to Xtable}
\usage{
as_xtable(x, ...)
}
\arguments{
\item{x}{object}

\item{...}{passed arguments}
}
\value{
see methods
}
\description{
Coerces to xtable output class.  Generic,
with method \code{\link{as_xtable.tablet}}.
}
\examples{
example(classifiers)
as_xtable(tablet(tablette(groupwise(x))))
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tablette.R
\name{tablette.tablet}
\alias{tablette.tablet}
\title{Convert to tablette from tablet}
\usage{
\method{tablette}{tablet}(x, ...)
}
\arguments{
\item{x}{tablette}

\item{...}{passed arguments}
}
\value{
tablet data.frame
}
\description{
Converts to tablette from tablet.
Intends to be the inverse of \code{\link{tablet.tablette}}.
}
\examples{
example(tablet.tablette)
}
\seealso{
Other tablette: 
\code{\link{tablette}()}

Other tablet: 
\code{\link{as_tablet}()},
\code{\link{as_tablet.data.frame}()},
\code{\link{header_rows}()},
\code{\link{header_rows.tablet}()},
\code{\link{tablet.tablette}()}
}
\concept{tablet}
\concept{tablette}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gridtable.R
\name{kbl.default}
\alias{kbl.default}
\title{Default kbl Method}
\usage{
\method{kbl}{default}(
  x,
  format,
  digits = getOption("digits"),
  row.names = NA,
  col.names = NA,
  align,
  caption = NULL,
  label = NULL,
  format.args = list(),
  escape = TRUE,
  table.attr = getOption("knitr.table.html.attr", ""),
  booktabs = FALSE,
  longtable = FALSE,
  tabular = if (longtable) "longtable" else "tabular",
  valign = if (tabular \%in\% c("tabularx", "xltabular")) "{\\\\linewidth}" else "[t]",
  position = "",
  centering = TRUE,
  vline = getOption("knitr.table.vline", if (booktabs) "" else "|"),
  toprule = getOption("knitr.table.toprule", if (booktabs) "\\\\toprule" else
    "\\\\hline"),
  bottomrule = getOption("knitr.table.bottomrule", if (booktabs) "\\\\bottomrule" else
    "\\\\hline"),
  midrule = getOption("knitr.table.midrule", if (booktabs) "\\\\midrule" else
    "\\\\hline"),
  linesep = if (booktabs) c("", "", "", "", "\\\\addlinespace") else "\\\\hline",
  caption.short = "",
  table.envir = if (!is.null(caption)) "table",
  ...
)
}
\arguments{
\item{x}{Object to render.}

\item{format, digits, row.names, col.names, align, caption, label}{Arguments
passed to \code{\link[kableExtra]{kbl}}.}

\item{format.args, escape, table.attr, booktabs, longtable, tabular, valign}{Arguments passed to \code{\link[kableExtra]{kbl}}.}

\item{position, centering, vline, toprule, bottomrule, midrule, linesep}{Arguments passed to \code{\link[kableExtra]{kbl}}.}

\item{caption.short, table.envir}{Arguments passed to
\code{\link[kableExtra]{kbl}}.}

\item{...}{Additional arguments passed to \code{\link[kableExtra]{kbl}}.}
}
\value{
A kable object.
}
\description{
Calls the existing \code{\link[kableExtra]{kbl}} implementation.
}
\seealso{
\code{\link[kableExtra]{kbl}}

Other gridtable: 
\code{\link{as.data.frame.gridtable}()},
\code{\link{as_calibrated}()},
\code{\link{as_calibrated.data.frame}()},
\code{\link{as_gridtable}()},
\code{\link{as_gridtable.character}()},
\code{\link{as_gridtable.data.frame}()},
\code{\link{kbl}()},
\code{\link{kbl.calibrated}()},
\code{\link{kbl.gridtable}()},
\code{\link{print.gridtable}()}
}
\concept{gridtable}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tablet.R
\name{categoricals.data.frame}
\alias{categoricals.data.frame}
\title{Identify Categoricals for Data Frame}
\usage{
\method{categoricals}{data.frame}(x, ..., na.rm_fac = FALSE, exclude_fac = NULL, all_levels = FALSE)
}
\arguments{
\item{x}{data.frame; names must not start with '_tablet_'}

\item{...}{passed to \code{\link{classifiers}}}

\item{na.rm_fac}{whether to drop NA observations; passed to \code{\link[tidyr]{gather}} as na.rm}

\item{exclude_fac}{which factor levels to exclude; see \code{\link{factor}} (exclude)}

\item{all_levels}{whether to supply records for unobserved levels}
}
\value{
same class as x
}
\description{
Identifies categorical columns: literally, factors. Stacks factor levels and supplies value = 1.
}
\examples{
example(classifiers)
categoricals(x)
levels(categoricals(x)$`_tablet_level`)
}
\keyword{internal}
